/// Careful Triangles and Artful Graphics
/// Color map support
/// Copyright (c) 2018 by Howard I. Cannon - MIT License

"use strict";

/// Class to manage color maps
function ColorMapManager($scope, drawer, onmapchangefn, params) {
  this.$scope=$scope;
  this.drawer=drawer;
  this.onmapchangefn=onmapchangefn;
  this.ncolorsOptions=sliderOptions({from: 2, to: 128, step: 1});
  this.ncolors=parseInt(defarg(params.cmn_, "32"));
  this.fillOptions=[{label: "", type: "highcontrast", explicit: false},
                    {label: "High Contrast", type: "highcontrast", explicit: false},
                    {label: "Gray", type: "gray", explicit: false},
                    {label: "Random", type: "random", explicit: true}];
  var initfill=defarg(params.cmt_, "highcontrast");
  this.lastFill=this.fillOptions.find(function(fo) {return fo.type==initfill;});
  if (!this.lastFill) this.lastFill=this.fillOptions[0];
  this.colorMap=this.drawer.makeColorMap({length: this.ncolors, type: this.lastFill.type, map: params.cmm_});
  this.fill=this.fillOptions[0];;
  this.drawer.absMode=(defarg(params.cmabs_, "false")=="true");
  return;
}

jQuery.extend(ColorMapManager.prototype, {
  fillQueryParams: function(params) {
    params.cmn_=this.ncolors;
    params.cmt_=this.lastFill.type;
    params.cmabs_=this.drawer.absMode;
    if (this.lastFill.explicit) {
      // Need to save explicit color map here, e.g., for a random map
      params.cmm_=this.drawer.colorMapToString(this.colorMap);
    };
    return params;
  },

  fillChange: function() {
    this.colorMap=this.drawer.makeColorMap({length: this.ncolors, type: this.fill.type});
    this.onmapchangefn();
    this.lastFill=this.fill;
    this.fill=this.fillOptions[0];
    return;
  },

  ncolorsChange: function() {
    this.colorMap=this.drawer.makeColorMap({length: this.ncolors, type: this.lastFill.type});
    this.onmapchangefn();
    return;
  },

  absModeChange: function() {
    this.onmapchangefn();
    return;
  },

  end: ""
})
