﻿/// Careful Triangles and Artful Graphics
/// Angular controllers
/// Copyright (c) 2018 by Howard I. Cannon - MIT License

"use strict";

/// Support functions, classes
function sliderOptions(extras) {
  return jQuery.extend({realtime: true, limits: false,
                        css: {background: {"background-color": "silver"}, before: {"background-color": "purple"},
                              default: {"background-color": "white"}, after: {"background-color": "purple"},
                              pointer: {"background-color": "red"}}},
                       (extras || {}));
}

Array.prototype.find=function(fn) {
  for (var i=0; i<this.length; i++) {
    if (fn(this[i])) return this[i];
  };
  return null;
}


/// Support class to create and manage input controls
function InputControls($scope, ctlcontainerid, defvals) {
  this.$scope=$scope;
  this.$container=$("#"+ctlcontainerid);
  this.currentKey=null;
  if (!defvals) defvals={};
  this.defvals=defvals;
  $scope.controls={};
  return;
}

jQuery.extend(InputControls.prototype, {
  // Fills in query parameters to reproduce graphic
  fillQueryParams: function(params) {
      for (var ns=0; ns<this.specs.length; ns++) {
        var spec=this.specs[ns];
        var modelname=spec.name;
        params[modelname]=this.$scope[modelname];
      };
    return;
  },

  // specs is an array of control specifications, set up in order
  setupControls: function(key, animator, paramfn, clearfn, changefn, initfn, specs) {
    this.specs=specs;
    this.paramfn=paramfn;
    this.clearfn=clearfn;
    this.changefn=changefn;
    this.initfn=initfn;
    var $scope=this.$scope;
    var scopectls=$scope.controls;
    function makeMidiChangeFn(name, reset) {
      return function() {return $scope.controls.onctlchange(name, reset);}
    };
    if (this.currentKey) {
      $("#"+this.currentKey).hide(0);
    };
    this.currentKey="InputControls_"+key;
    var curctls=$("#"+this.currentKey);
    if (curctls.length>0) {
      curctls.show(0);
    } else {
      var html="<div id='"+this.currentKey+"'>\n";
      for (var ns=0; ns<specs.length; ns++) {
        var spec=specs[ns];
        var modelname=spec.name, name=spec.name+"_"+key, reset=spec.reset;
        if (!(modelname in $scope)) $scope[modelname]=spec.default;
        if (modelname in this.defvals) $scope[modelname]=parseFloat(this.defvals[modelname]);
        scopectls[modelname+"Old"]=spec.default;
        var cl=((ns&1) ? "odd" : "even")+"-slider-holder";
        html+="<div class='"+cl+"'>";
        if (spec.label) html+="<div class='slider-label'>"+spec.label+"</div>\n";

        switch (spec.type) {
        case "slider":
          // The slider itself
          var optname=name+"Options";
          scopectls[optname]=sliderOptions(spec.extras);
          html+="<div class='slider'>"
          html+="<input ng-model='"+modelname+"' type='number' slider options='controls."+optname+"' ng-change='controls.onctlchange(\""+modelname+"\", "+(reset ? "true" : "false")+")' />";
          html+="</div>\n";
          // Add the numeric display
          html+="<input class='slider-number' type='number' size='"+(defarg(spec.numbersize, 4))+"'";
          html+=" ng-model='"+modelname+"'";
          html+=" ng-change='controls.onctlchange(\""+modelname+"\", "+(reset ? "true" : "false")+")'";
          html+="/>";
        };
        html+="</div>";
        if (spec.midi) {
          // Also setup midi support
          $scope.midi.onFromSpec(spec.midi, modelname, makeMidiChangeFn(modelname, reset));
        };
      };
      html+="</div>";
      this.$container.append($scope.$compile(html)($scope));
    };
    scopectls.onctlchange=function(param, reset) {
      // Recalculate the parameters
      paramfn();
      reset=changefn(param, reset)
      // If need to do a reset, then call the initialization function
      if (reset) {
        animator.reset();
        initfn();
      };
      // Then, capture new parameter values as old values for next time around
      for (var ns=0; ns<specs.length; ns++) {
        var name=specs[ns].name;
        scopectls[name+"Old"]=$scope[name];
      };
      return;
    };

    animator.clearFn=function(reset) {
      return clearfn(reset);
    };

    this.defvals={};
    return;
  },

  reset: function(hard) {
    if (this.paramfn) {
      this.paramfn();
      if (hard) this.initfn();
    };
    return;
  },

  end: ""
});


/// Angular controllers
var thisControllers = angular.module('thisControllers', []);

var allAnimations=[
  // {label: "Test PI", key: "testpi", specfn: "RotorAnimations.getTestPIFreqSpec", page: "artcircle"},
  {label: "Circle", key: "circle", class: CircleAnim, page: "artcircle"},
  {label: "Besser", key: "besser", class: BesserAnim, page: "artcircle"},
  {label: "Gibbs Tri", key: "gibbstri", class: GibbsTriAnim, page: "artcircle"},
  {label: "Gibbs Square", key: "gibbssquare", class: GibbsSquareAnim, page: "artcircle"},
  {label: "Spiro", key: "spiro", class: SpiroAnim, page: "artcircle"},
  // Translated from Lisp code but seems totally wrong!
  // {label: "Dragon", key: "dragon", class: DragonAnim, page: "artcircle"},
  {label: "Hilbert", key: "hilbert", class: HilbertAnim, page: "artcircle"},
  {label: "Base2", key: "base2", class: Base2Anim, page: "artcircle"},
  {label: "InCirc", key: "incirc", specfn: "RotorAnimations.getInCircSpec", page: "artcircle"},
  {label: "D Drag", key: "ddrag", specfn: "RotorAnimations.getDDragSpec", page: "artcircle"},
  {label: "Big Doily", key: "bigdoily", specfn: "RotorAnimations.getBigDoilySpec", page: "artcircle"},
  {label: "Tri Flow Snake", key: "triflowsnake", specfn: "RotorAnimations.getTriFlowSnakeSpec", page: "artcircle"},
  {label: "Tri Drag", key: "tridrag", specfn: "RotorAnimations.getTDragSpec", page: "artcircle"},
  {label: "Tri Fill", key: "trifill", specfn: "RotorAnimations.getTriFillSpec", page: "artcircle"},
  {label: "Mandelflake", key: "mandelflake", specfn: "RotorAnimations.getMandelflakeSpec", page: "artcircle"},
  {label: "Ring No Ring", key: "ringnoring", specfn: "RotorAnimations.getRingNoRingSpec", page: "artcircle"},
  {label: "Ring No Ring 2", key: "ringnoring2", specfn: "RotorAnimations.getRingNoRing2Spec", page: "artcircle"},
  {label: "Dragrid", key: "dragrid", specfn: "RotorAnimations.getDragridSpec", page: "artcircle"},
  {label: "Sextuple Tri", key: "sexttri", specfn: "RotorAnimations.getSextuplePointTri", page: "artcircle"}
];

thisControllers.controller('IndexCtrl', ['$scope', '$location', function($scope, $location) {
  if (urlArgs && Object.keys(urlArgs).length>0) {
    // Dispatch to page based on specification from url
    var anim=allAnimations.find(function(a) {return a.key==urlArgs.a_;});
    if (anim) {
      var page=anim.page;
      if (!page) page="artcircle";
      if (!urlArgs.nohide) $("#navbar").hide();
      $location.path(page);
    };
  };
  return;
}]);

thisControllers.controller('TriangleCtrl', ['$scope', function($scope) {
  $scope.hidePixelValues=true;
  var canvas=$("#testcanvas").get(0);
  $scope.drawer=new ColorMappedCanvasDrawer(canvas, 16);
  var ranges={};
  $scope.range=function(n) {
    var range=ranges[n];
    if (!range) {
      range=new Array(n);
      for (var i=0; i<n; i++) range[i]=i;
      ranges[n]=range;
    };
    return range;
  };

  $scope.clear=function() {
    $scope.drawer.fill(0);
    $scope.drawer.update();
    return;
  };

  $scope.drawRectangle=function() {
    $scope.drawer.drawRectangle(parseFloat($scope.x0), parseFloat($scope.y0),
                                parseFloat($scope.x1), parseFloat($scope.y1),
                                ["add", 1]);
    $scope.drawer.update();
    return;
  };

  $scope.drawTriangle=function() {
    $scope.drawer.drawTriangle(parseFloat($scope.x0), parseFloat($scope.y0),
                               parseFloat($scope.x1), parseFloat($scope.y1),
                               parseFloat($scope.x2), parseFloat($scope.y2),
                               ["add", 1]);
    $scope.drawer.update();
    return;
  };

  $scope.drawRectWithTri=function() {
    var x0=parseFloat($scope.x0), y0=parseFloat($scope.y0),
        x1=parseFloat($scope.x1), y1=parseFloat($scope.y1);
/*
    // Randomly pick a point in the interior of the rectangle
    var ix=Math.round((x1-x0)*Math.random()),
        iy=Math.round((y1-y0)*Math.random());
*/
    // Randomly pick a point anywhere on the canvas
    var ix=Math.round(($scope.drawer.width-1)*Math.random()),
        iy=Math.round(($scope.drawer.height-1)*Math.random());
    // ix=9; iy=4;
    $scope.innerx=ix;
    $scope.innery=iy;
    $scope.drawer.drawTriangleAddDet(x0, y0, x1, y0, ix, iy, 1);
    $scope.drawer.drawTriangleAddDet(x1, y0, x1, y1, ix, iy, 1);
    $scope.drawer.drawTriangleAddDet(x1, y1, x0, y1, ix, iy, 1);
    $scope.drawer.drawTriangleAddDet(x0, y1, x0, y0, ix, iy, 1);
    $scope.drawer.update();
    return;
  };

  return;
}]);

thisControllers.controller('TriangleAnimateCtrl', ['$scope', function($scope) {
  var canvas=$("#animatecanvas").get(0);
  $scope.drawer=new ColorMappedCanvasDrawer(canvas, 16);

  var randomCenter=function() {
    return [Math.round(($scope.drawer.width-1)*Math.random()),
            Math.round(($scope.drawer.height-1)*Math.random())];
  };

  var startAnimatorFn=null;
  var animationStep=function(timestamp) {
    if ($scope.stopflag) return;
    // Do next animation step
    if ($scope.animator()) {
      startAnimatorFn();
    } else {
      requestAnimationFrame(animationStep);
    };
    $scope.drawer.update();
    return;
  };

  var animatePoly=function(poly, alu, randomcenter) {
    $scope.stopflag=false;
    var center=null;
    if (randomcenter===true) center=randomCenter();
    else if (Array.isArray(randomcenter)) center=randomcenter;
    $scope.animator=$scope.drawer.createAnimatePolyWithTriangles(poly, alu, center);
    requestAnimationFrame(animationStep);
    return;
  };

  $scope.stop=function() {
    $scope.stopflag=true;
    return;
  };

  $scope.clear=function() {
    $scope.drawer.fill(0);
    $scope.drawer.update();
    return;
  };

  $scope.drawRectangle=function(randomcenter) {
    var x0=parseFloat($scope.x0), y0=parseFloat($scope.y0);
    var x1=parseFloat($scope.x1), y1=parseFloat($scope.y1);
    startAnimatorFn=function() {animatePoly([[x0, y0], [x1, y0], [x1, y1], [x0, y1]], 1, randomcenter);};
    startAnimatorFn();
    return;
  };

  $scope.drawRectangleOneByOne=function(randomcenter) {
    var x0=parseFloat($scope.x0), y0=parseFloat($scope.y0);
    var x1=parseFloat($scope.x1), y1=parseFloat($scope.y1);
    var poly=[];
    var step=2;
    for (var x=x0; x<x1; x+=step) poly.push([x, y0]);
    for (var y=y0; y<y1; y+=step) poly.push([x1, y]);
    for (var x=x1; x>x0; x-=step) poly.push([x, y1]);
    for (var y=y1; y>y0; y-=step) poly.push([x0, y]);
    startAnimatorFn=function() {animatePoly(poly, 1, randomcenter);};
    startAnimatorFn();
    return;
  };

  return;
}]);

thisControllers.controller('CircleAnimateCtrl', ['$scope', function($scope) {
  var canvas=$("#animatecanvas").get(0);
  $scope.drawer=new ColorMappedCanvasDrawer(canvas, {type: "random", length: 32});
  $scope.animator=new Animator($scope, $scope.drawer, {});
  var setStepFn=function(sf) {$scope.animator.stepFn=sf;};

  var randomCenter=function() {
    return [Math.round(($scope.drawer.width-1)*Math.random()),
            Math.round(($scope.drawer.height-1)*Math.random())];
  };

  $scope.circleClass="STCircle";
  $scope.circleClasses=["Circle", "STCircle", "ZCircle"];
  $scope.setupDefaultCircleClass=function() {
    CircleUtils.defaultCircleClassName=$scope.circleClass;
    return;
  };
  $scope.setupDefaultCircleClass();

  $scope.useGRotor=false;
  $scope.setupDefaultRotor=function() {
    CircleUtils.defaultRotorFnPrefix=($scope.useGRotor ? "G" : "");
  };
  $scope.setupDefaultRotor();

  $scope.args="";
  var parseArgs=function(minargs) {
    var undef;
    var args=[];
    var splitargs=$scope.args.split(",");
    for (var na=0; na<splitargs.length; na++) {
      var a=splitargs[na].trim();
      if (a=="") args.push(undef);
      else args.push(eval(splitargs[na].trim()));
    };
    if (minargs) while (args.length<minargs) args.push(undef);
    return args;
  };

  $scope.circleOnce=function() {
    var args=parseArgs(5);
    var radius=defarg(args[0], 200), cx=defarg(args[1], $scope.drawer.width/2), cy=defarg(args[2], $scope.drawer.height/2);
    var period=args[3];
    var alpha=defarg(args[4], 255);
    $scope.drawer.drawFilledCircle(cx, cy, radius, ["add", 1, "set", alpha], period);
    $scope.drawer.update();
    return;
  };

  $scope.animateCircle=function(usecenter) {
    var args=parseArgs(3);
    var radius=defarg(args[0], 200), dt=defarg(args[1], 0.0123), phi=defarg(args[2], 0.0);
    var circle=CircleUtils.makeCircle({centerOf: $scope.drawer, radius: radius, dt: dt, phi: phi});
    var drawpt=[circle.cx, circle.cy];
    if (!usecenter) drawpt=randomCenter();
    var alu=["add", 1, "noop", 0];
    setStepFn(function(command) {
      // alu[1]=(command=="backward" ? -1 : 1);
      var p0=circle.point();
      var p1=(command=="backward" ? circle.back() : circle.step());
      // $scope.drawer.drawTriangle(p0[0], p0[1], p1[0], p1[1], drawpt[0], drawpt[1], alu);
      $scope.drawer.drawTriangleAddDet(p0[0], p0[1], p1[0], p1[1], drawpt[0], drawpt[1], alu[1]);
    });
    // Make the center transparent
    // $scope.drawer.drawFilledCircle(circle.cx, circle.cy, radius/2, ["set", 0, "set", 0]);
    return;
  };

  $scope.animateTwoCircles=function(usecenter) {
    var args=parseArgs(3);
    var radius=defarg(args[0], 200), dt=defarg(args[1], 0.0123), phi=defarg(args[2], 0.0);
    var circle0=new Circle({centerOf: $scope.drawer, radius: radius, dt: dt, phi: phi});
    circle0.setCenter(circle0.cx-Math.abs(radius), circle0.cy);
    var drawpt0=[circle0.cx, circle0.cy];
    if (!usecenter) drawpt0=randomCenter();
    var circle1=new ZCircle({centerOf: $scope.drawer, radius: radius, dt: dt, phi: phi});
    circle1.setCenter(circle1.cx+Math.abs(radius), circle1.cy);
    var drawpt1=[circle1.cx, circle1.cy];
    if (!usecenter) drawpt1=randomCenter();
    var alu=["add", 1];
    setStepFn(function(command) {
      alu[1]=(command=="backward" ? -1 : 1);
      var p0a=circle0.point();
      var p1a=(command=="backward" ? circle0.back() : circle0.step());
      $scope.drawer.drawTriangle(p0a[0], p0a[1], p1a[0], p1a[1], drawpt0[0], drawpt0[1], ["add", 1]);
      var p0b=circle1.point();
      var p1b=(command=="backward" ? circle1.back() : circle1.step());
      $scope.drawer.drawTriangle(p0b[0], p0b[1], p1b[0], p1b[1], drawpt1[0], drawpt1[1], ["add", 1]);
      var xdiff=Math.abs((p0a[0]-circle0.cx)-(p0b[0]-circle1.cx));
      var ydiff=Math.abs((p0a[1]-circle0.cy)-(p0b[1]-circle1.cy));
      if (xdiff>1.0 || ydiff>1.0) {
        // alert("xdiff="+xdiff+", ydiff="+ydiff);
        var foo=xdiff+ydiff;
      };
    });
    return;
  }

  var animate=function(fn) {
    setStepFn(fn($scope.drawer));
    return;
  }

  $scope.animateBogor=function() {
    var args=parseArgs(3);
    animate(function(d) {return RotorAnimations.getBogorFn(d, args[0], args[1], args[2]);});
    return;
  };

  $scope.animateBesser=function() {
    var args=parseArgs(5);
    animate(function(d) {return RotorAnimations.getBesserFn(d, args[0], args[1], args[2], args[3], args[4]);});
    return;
  };

  $scope.animateGibbsSquare=function() {
    var args=parseArgs(3);
    animate(function(d) {return RotorAnimations.getGibbsSquareFn(d, args[0], args[1], args[2]);});
    return;
  };

  $scope.animateGibbsTri=function() {
    var args=parseArgs(5);
    animate(function(d) {return RotorAnimations.getGibbsTriFn(d, args[0], args[1], args[2], args[3], args[4]);});
    return;
  };

  // Args are pairs of radius and period
  // 200,125,150,50,100,20
  $scope.linewidth=0;
  $scope.animateSpiro=function() {
    var args=parseArgs();
    if (args.length<2) args=[200,125, 150,51];
    var randds=[];
    for (var na=0; (na+1)<args.length; na+=2) {
      randds.push([args[na], (Math.PI*2)/args[na+1], 0]);
    };
    // Only color final circle
    randds[randds.length-1][2]=1;
    var linewidth=$scope.linewidth;
    animate(function(d) {return RotorAnimations.getSpiroFnFromList(d, randds, linewidth)});
    return;
  };

  $scope.stepOptions=sliderOptions({from: 0, to: 100, step: 1});
  $scope.delayOptions=sliderOptions({from: 0, to: 2000, step: 10});

  return;
}]);

thisControllers.controller('ArtCircleCtrl', ['$scope', '$parse', '$compile', '$location', function($scope, $parse, $compile, $location) {
  var navbar=$("#navbar");
  navbar.hide();

  var canvas=$("#animatecanvas");
  $scope.$compile=$compile;
  $scope.drawer=new ColorMappedCanvasDrawer(canvas.get(0), {type: "random", length: 32});
  $scope.animator=new Animator($scope, $scope.drawer, urlArgs);
  $scope.colorMap=new ColorMapManager($scope, $scope.drawer, function() {$scope.colorMapChange();}, urlArgs);
  $scope.midi=new MidiSupport($scope, $parse);
  $scope.ics=new InputControls($scope, "controlsHolder", urlArgs);

  $scope.animator.setupMidi($scope.midi);

  // Animation sliders
  $scope.stepOptions=sliderOptions({from: 1, to: 100, step: 1});
  $scope.midi.onBendInRangeInScope(230, "animator.stepsPerCycle", 1, 100);
  
  $scope.delayOptions=sliderOptions({from: 0, to: 2000, step: 10});
  $scope.midi.onBendInRangeInScope(231, "animator.delayMS", 0, 2000);

  $scope.animations=allAnimations.filter(function(a) {return a.page=="artcircle";});
  $scope.selectedAnimation=$scope.animations[0];
  $scope.animationChange=function() {
    var cls=$scope.selectedAnimation.class;
    var initarg=$scope.selectedAnimation.specfn;
    if (initarg) cls=CoeffsAnim;
    else initarg=$scope.selectedAnimation.initarg;
    $scope.animator.pause();
    $scope.animation=new (cls)($scope, $scope.ics, $scope.animator, initarg);
    $scope.ics.reset(true);
    $scope.animator.clear(true);
    return;
  }
  var anim=allAnimations.find(function(a) {return a.key==urlArgs.a_;});
  var autostart=false;
  if (anim) {
    $scope.selectedAnimation=anim;
    autostart=true;
  };
  $scope.animationChange();
  urlArgs={};

  $scope.colorMapChange=function() {
    $scope.drawer.setColorMap($scope.colorMap.colorMap);
    $scope.drawer.update(true);
    return;
  };
  // Setup initial color map and don't bother remapping
  $scope.drawer.setColorMap($scope.colorMap.colorMap, true);

  var resizeCanvas=function() {
    // Size canvas to fit within container by width and window by height
    var w=Math.floor($("#canvascontainer").innerWidth());
    var h=Math.floor($(window).height()-canvas.offset().top);
    if ($scope.drawer.width!=w || $scope.drawer.height!=h) {
      $scope.drawer.resizeCanvas(w, h, false);
      // Always do a hard reset
      $scope.ics.reset(true);
      $scope.drawer.update();
    };
    return;
  };

  $(window).resize(resizeCanvas);
  resizeCanvas();

  // Expand/contract controls
  $scope.showControls=true;
  $scope.setControls=function(state) {
    if (state==="toggle") state=!$scope.showControls;
    $scope.showControls=state;
    $scope.controlsFlag=state ? "^" : "v";
    var cp=$("#controlPanel");
    if (state) cp.fadeIn(500);
    else cp.fadeOut(500);
    return;
  };
  $scope.selectPanel=function(panel) {
    function inOrOut(id) {
      var p=$("#"+id+"Controls");
      if (id==panel) p.show(0);
      else p.hide(0);
      return;
    }
    if (panel=="url") $scope.setUrl();
    inOrOut("animation");
    inOrOut("color");
    inOrOut("url");
    return;
  };
  $scope.setControls(true);

  $scope.setUrl=function() {
    // Create a dictionary of parameters
    var params={ };
    $scope.ics.fillQueryParams(params);
    $scope.colorMap.fillQueryParams(params);
    $scope.animator.fillQueryParams(params);
    var url=g_urlBase+"?a_="+$scope.selectedAnimation.key;
    for (var key in params) {
      url+="&"+key+"="+params[key];
    };
    // Special case the mapping of internal to external URL
    url=url.replace(/hicannon\.dyndns\.org:82/i, "hichacks.is-a-geek.net");
    $scope.url=url;
    /*
    // +++ This might be nice, but it causes Chrome to never display the URL tab
    if (window.history.replaceState) {
      window.history.replaceState(null, document.title, url);
    };
    */
    return;
  },

  $scope.midi.onCommand(58, "noteon", function(data) {
    $scope.$apply(function() {if (data==127) $scope.setControls(!$scope.showControls);});
    return;
  });

  // Some good numbers:
  // Period: 381, incr: 125
  // 362, 177

  if (autostart) $scope.animator.setRunState(true);

  return;
}]);
