/// Careful Triangles and Artful Graphics
/// Midi support
/// Copyright (c) 2018 by Howard I. Cannon - MIT License

"use strict";

function MidiSupport($scope, $parse) {
  this.$scope=$scope;
  this.$parse=$parse;
  this.cmdFns=[];
  this.cmdTypes=[];
  // Handle sub-dispatch for continuous controllers
  this.continuousFns=[];
  this.cmdFns[0xB0]="continuous";
  this.cmdTypes[0xB0]="continuous";

  this.noteOffFns=[];
  this.cmdFns[0x80]="noteoff";
  this.cmdTypes[0x80]="noteoff";
  this.noteOffFns=[];
  this.noteOnFns=[];
  this.cmdFns[0x90]="noteon";
  this.cmdTypes[0x90]="noteon";
  this.noteOnFns=[];

  if (typeof(navigator.requestMIDIAccess)==="function") {
    console.log("midi");
    this.midi=navigator.requestMIDIAccess({sysex: false});
    this.midi.then(this.onAccessOK.bind(this), this.onAccessFail.bind(this));
  };
}

jQuery.extend(MidiSupport.prototype, {

  onAccessOK: function(midiaccess) {
    this.midiAccess=midiaccess;
    var inputs=midiaccess.inputs.values();
    for (var input=inputs.next(); input && !input.done; input=inputs.next()) {
      var inp=input.value;
      console.log(inp.type+", "+inp.id+", "+inp.manufacturer+", "+inp.name);
      inp.onmidimessage=this.onMIDIMessage.bind(this);
    };
    return;
  },

  onAccessFail: function(e) {
    this.midi=null;
    return;
  },

  onMIDIMessage: function(msg) {
    var data=msg.data;
    console.log('Midi data', data);
    var cmd=data[0];
    var type=this.cmdTypes[cmd];
    var fn=this.cmdFns[cmd];
    var subfns;
    if (fn && type) {
      switch (type) {
      case "bend": {
        fn((data[2]<<7)|data[1]);
        break;
      };
      case "continuous": subfns=this.continuousFns; break;
      case "noteoff": subfns=this.noteOffFns; break;
      case "noteon": subfns=this.noteOnFns; break;
      };
    };
    if (subfns) {
      fn=subfns[data[1]];
      if (fn) fn(data[2]);
    };
    return;
  },

  // Type is: noteoff, noteon, aftertouch, continuous, patch, pressure, bend, other
  onCommand: function(cmd, type, fn) {
    switch (type) {
    case "continuous": this.continuousFns[cmd]=fn; break;
    case "noteoff": this.noteOffFns[cmd]=fn; break;
    case "noteon": this.noteOnFns[cmd]=fn; break;
    default:
      this.cmdFns[cmd]=fn;
      this.cmdTypes[cmd]=type;
    };
    return;
  },

  // args is a list where the first element is the type and the remaining elements
  // are the arguments to the type-specific handler
  onFromSpec: function(args, varpath, changefn) {
    switch(args[0]) {
    case "bend":
      return this.onBendInRangeInScope(args[1], varpath, args[2], args[3], changefn);
    };
    return;
  },

  onBendInRangeInScope: function(cmd, varpath, low, high, changefn) {
    var accessor=this.$parse(varpath);
    var $scope=this.$scope;
    var round=(Number.isInteger(low) && Number.isInteger(high));
    this.onCommand(cmd, "bend", function(data) {
      var ndata=low+(high-low)*(data/((1<<14)-1));
      if (round) ndata=Math.round(ndata);
      $scope.$apply(function() {
        accessor.assign($scope, ndata);
        if (changefn) changefn();
      });
      return;
    });
    return;
  },

  end: ""
});
