;;; -*- Base: 10; Mode: LISP; Syntax: Common-lisp; Package: USER -*-

(eval-when (compile load eval)
 (defmacro second-value (exp)
  `(multiple-value-bind (ignore val) ,exp val))

 (DEFUN number-length (NUMBER)
  (TYPECASE NUMBER
    (FLOAT (second-value (decode-float number)))
    (integer (integer-length number))
    (RATIONAL (let* ((num (abs (numerator number)))
		      (denom (DENOMINATOR NUMBER))
		      (ans (- (integer-length num) (integer-length denom))))
		 (if (< ans 0) (if (> (ash num (- ans)) denom) (+ 1 ans) ans)
			       (if (> num (ash denom ans)) (+ 1 ans) ans))))		    
    (COMPLEX (max (number-length (REALPART NUMBER))
		   (number-length (IMAGPART NUMBER))))))
)
(DEFMACRO ASET (STUFF &REST ARRAY&INDICES)
  `(ZL:ASET ,STUFF .,ARRAY&INDICES))

(defmacro funcall1 (obj . body)
  (if (or (nlistp obj) (eq (car obj) 'quote)) `(progn (funcall ,obj . ,body) ,obj)
    `(let ((.funcall1-internal-var. ,obj))
       (funcall .funcall1-internal-var. . ,body)
       .funcall1-internal-var.)))

(defflavor aostri-fixin ()
	   ())

(defwhopper (:aos-triangle aostri-fIXin) (x1 y1 x2 y2 x3 y3 &rest dcolor)
  (or (zerop (- (* (- (setq x1 (floor x1)) (setq x2 (floor x2)))
		   (- (setq y2 (floor y2)) (setq y3 (floor y3))))
		(* (- (setq y1 (floor y1)) y2) (- x2 (setq x3 (floor x3))))))
      (LEXPR-CONTINUE-WHOPPER x1 y1 x2 y2 x3 y3 dcolor)))

(defflavor reposition-mixin (min-x max-x min-y max-y
			     dx dy largest-x largest-y)
	   ()
  (:required-flavors tv:window))

(defflavor graphics-window
	()
	(aostri-fixin #||replay-triangles-mixin||# reposition-mixin tv:window)
  :gettable-instance-variables)

(defflavor color-graphics-window
	((sos-alu (color:sc-fill-alu (LSH -1 0) (LSH -1 0)))	;16
	 (aos-alu (color:sc-fill-alu  (LSH 1 0) (LSH -1 0)))	;#O377 16
	 (add-alu (color:sc-fill-alu  0 (LSH -1 0))))	;16
	(graphics-window)
  :gettable-instance-variables)

(defmethod (:init color-graphics-window :after) (ignore)
  (send sos-alu :set-alu color:alu-add)
  (send aos-alu :set-alu color:alu-add)
  (send add-alu :set-alu color:alu-add))

(defmethod (:aspect graphics-window) (x y &aux (width (funcall self ':inside-size)) height)
  (setq width (+ width (floor x 2) (- (mod (+ width (floor x 2)) x)))
	height (floor (* width y) x))
  (multiple-value-bind (wid hei) (funcall tv:superior ':inside-size)
    (let ((lap (max 0
		    (ceilING (+ width tv:left-margin-size tv:right-margin-size (- wid)) x)
		    (ceilING (+ height tv:top-margin-size tv:bottom-margin-size (- hei)) y))))
    (funcall self ':set-inside-size (- width (* x lap)) (- height (* y lap))))))

(defmethod (:center graphics-window) ()
  (multiple-value-bind (width height) (funcall self ':inside-size)
    (values (floor width 2) (floor height 2))))

(defvar color-HAX-window)

(DEFMETHOD (:EXPOSE COLOR-GRAPHICS-WINDOW :AFTER) (&REST IGNORE)
  (SEND TV:SUPERIOR :SET-MAP-MODE :8-BIT 1 0))	;16

(defun make-color-window ()
  (setq color-HAX-window
	(MULTIPLE-VALUE-BIND (WIDTH HEIGHT) (SEND COLOR:COLOR-SCREEN :VISIBLE-SIZE)
	  (tv:make-window 'color-graphics-window ':superior color:color-screen
			  :WIDTH WIDTH :HEIGHT HEIGHT ':expose-p 't
			  ':blinker-p 'nil ':borders 'nil
			  ':label '(:font fonts:tr8 :string "So draw something")))))

(defvar bw-window)

(defun make-bw-window ()
  (setq bw-window
	(tv:make-window 'graphics-window ':superior tv:main-screen
			':edges-from ':mouse ':expose-p 't
			':blinker-p 'nil ':borders 'nil
			':label '(:font FONTS:5x5 :string "So draw something")
			':save-bits 'T)))

(defvar graphics-window)

(defun get-window ()
  (SEND (setq graphics-window (cond ((and (boundp 'color:color-screen)
					  color:color-screen
					  (tv:sheet-exposed-p color:color-screen))
				     (IF (AND (boundp 'color-HAX-window)
					      color-HAX-window
					      (EQL (SEND color-HAX-window :SUPERIOR)
						   color:color-screen))
					 color-HAX-window
				      (make-color-window)))
				    ((get-bw-window))))
	:EXPOSE)
  graphics-window)

(defun get-bw-window ()
  (send (or (and (boundp 'bw-window) bw-window) (make-bw-window)) :expose)
  bw-window)

(defun-in-flavor (calculate-new-size reposition-mixin) ()
  (multiple-value-bind (width height)
      (funcall self ':inside-size)
    (setq largest-x (1- width)
	  largest-y (1- height))))

(defmethod (:init reposition-mixin :after) (ignore)
  (calculate-new-size)
  (SEND SELF :CLEAR-WINDOW))

(defmethod (:change-of-size-or-margins reposition-mixin :after) (&rest ignore)
  (calculate-new-size))

(defmethod (:clear-WINDOW reposition-mixin :after) ()
  (setq min-x largest-x  max-x 0 min-y largest-y max-y 0 dx 0 dy 0))

(defwhopper (:DRAW-FILLED-IN-CIRCLE reposition-mixin) (X Y R &REST REST)
  (SETF MIN-X (MIN MIN-X (- X R))  MAX-X (MAX MAX-X (+ X R))
	MIN-Y (MIN MIN-Y (- Y R))  MAX-Y (MAX MAX-Y (+ Y R)))
  (LEXPR-CONTINUE-WHOPPER (+ x DX) (+ y DY) R REST))

(defwhopper (:DRAW-STRING reposition-mixin) (STRING X Y &optional (TO-X (+ x 1)) (TO-Y y)
						    &REST REST)
  (LEXPR-CONTINUE-WHOPPER STRING (+ x DX) (+ y DY) (+ TO-X DX) (+ TO-Y DY) REST))

(defwhopper (:aos-triangle reposition-mixin) (x1 y1 x2 y2 x3 y3 &rest dcolor &aux lap)
  (setq x1 (+ x1 dx) y1 (+ y1 dy) x2 (+ x2 dx) y2 (+ y2 dy) x3 (+ x3 dx) y3 (+ y3 dy)
	min-x (min min-x x1 x2 x3) max-x (max max-x x1 x2 x3))
  (cond ((and (minusp min-x) (plusp (setq lap (- largest-x max-x))))
	 (let ((gunch (min lap
			   (max (floor 32. (tv:screen-bits-per-pixel (tv:sheet-get-screen self)))
				(- min-x)))))
	   (cond ((> max-y min-y)
		  (funcall self ':bitblt-within-sheet tv:alu-seta
				(- -1 max-x) (- max-y min-y -1)
				0 min-y gunch min-y)
		  (funcall self ':draw-rectangle gunch (- max-y min-y -1)
						 0 min-y tv:erase-aluf)))
	   (setq min-x (+ min-x gunch) dx (+ dx gunch) max-x (+ max-x gunch)
		 x1 (+ x1 gunch) x2 (+ x2 gunch) x3 (+ x3 gunch))))
	(t (setq min-x (max min-x 0))))
  (cond ((and (plusp (setq lap (- max-x largest-x))) (plusp min-x))
	 (let ((gunch (min min-x
			   (max lap
				(floor 32.
				    (tv:screen-bits-per-pixel (tv:sheet-get-screen self)))))))
	   (cond ((> max-y min-y)
		  (funcall self ':bitblt-within-sheet tv:alu-seta
				(- largest-x min-x -1) (- max-y min-y -1)
				min-x min-y (- min-x gunch) min-y)
		  (funcall self ':draw-rectangle gunch (- max-y min-y -1)
						 (- largest-x gunch -1) min-y
						 tv:erase-aluf)))
	   (setq min-x (- min-x gunch) dx (- dx gunch) max-x (- max-x gunch)
		 x1 (- x1 gunch) x2 (- x2 gunch) x3 (- x3 gunch))))
	(t (setq max-x (min max-x largest-x))))
  (setq min-y (min min-y y1 y2 y3) max-y (max max-y y1 y2 y3))
  (cond ((and (minusp min-y) (plusp (setq lap (- largest-y max-y))))
	 (let ((gunch (min (- min-y) lap)))
	   (funcall self ':bitblt-within-sheet tv:alu-seta (- max-x min-x -1) (- -1 max-y)
							   min-x 0
							   min-x gunch)
	   (funcall self ':draw-rectangle (- max-x min-x) gunch min-x 0 tv:erase-aluf)
	   (setq min-y (+ min-y gunch) dy (+ dy gunch) max-y (+ max-y gunch)
		 y1 (+ y1 gunch) y2 (+ y2 gunch) y3 (+ y3 gunch))))
	(t (setq min-y (max min-y 0))))
  (cond ((and (plusp (setq lap (- max-y largest-y))) (plusp min-y))
	 (let ((gunch (min min-y lap)))
	   (funcall self ':bitblt-within-sheet tv:alu-seta
			 (- max-x min-x -1) (- largest-y min-y -1)
			 min-x min-y
			 min-x (- min-y gunch))
	   (funcall self ':draw-rectangle (- max-x min-x -1) gunch
					  min-x (- largest-y gunch -1)
					  tv:erase-aluf)
	   (setq min-y (- min-y gunch) dy (- dy gunch) max-y (- max-y gunch)
		 y1 (- y1 gunch) y2 (- y2 gunch) y3 (- y3 gunch))))
	(t (setq max-y (min max-y largest-y))))
  (LEXPR-CONTINUE-WHOPPER x1 y1 x2 y2 x3 y3 dcolor))

(defmethod (:aos-triangle graphics-window) (x1 y1 x2 y2 x3 y3 &optional (increment 1))
  (if (oddp increment) (SEND self ':draw-triangle x1 y1 x2 y2 x3 y3 tv:alu-xor)))

(defmethod (:aos-triangle color-graphics-window) (x1 y1 x2 y2 x3 y3 &optional (increment 1))
  (or (zerop (logand increment #O377))
      (let* ((area (- (* (- x2 x1) (- y3 y1)) (* (- x3 x1) (- y2 y1))))
	     (inc (* (signum area) increment)))
	(or (zerop inc)
	    ;(tv:prepare-sheet (self)		;??
	      (send self :draw-triangle x1 y1 x2 y2 x3 y3
		    (selectOR inc =
		      (1 aos-alu)
		      (-1 sos-alu)
		      (otherwise (send add-alu :set-fill-data (LSH INC 0))	;16
				 add-alu)))))));)

(defun one (ignore) 1)

(defun compilablep (f) (and (listp f) (member (car f) '(lambda named-lambda subst))))

(defun funcallablep (f)
  (or (compilablep f) (arrayp f) (ZL:typep f :instance) (FUNCTIONp f)))

(defun evaluablep (f) (or (symbolp f) (numberp f) (and (listp f) (not (compilablep f)))))

(defun funkify (f) (if (evaluablep f) `(lambda (n) ,f) f))

(defun random-color-map (&optional (n #o200) (MAP-SEGMENT 1))
  (loop for i below n
	for r = 0 THEN (random #o2000)
	for g = 0 THEN (random #o2000)
	for b = 0 THEN (random #o2000) do
	(send color:color-screen :write-color-map-image i r g b MAP-SEGMENT)
	(send color:color-screen :write-color-map-image
	      (- #o400 i) (- #o1777 r) (- #o1777 g) (- #o1777 b) MAP-SEGMENT))
  (send color:color-screen :new-color-map 'NIL MAP-SEGMENT))

(DEFFLAVOR RANDOM-SEQUENCE ((ARRAY (ZL:MAKE-ARRAY 71 :TYPE 'ZL:ART-FIXNUM))
			    (DEFAULT-MASK #O17777777777)
			    POINTER)
	   ()
  (:INITABLE-INSTANCE-VARIABLES DEFAULT-MASK))

(DEFMETHOD (:INIT RANDOM-SEQUENCE) (&REST IGNORE)
  (SETF POINTER 0)
  (LOOP FOR I BELOW 71
	FOR LO = 696969 THEN (SI:%LSHC-BIGNUM-STEP LO HI 1)
	FOR HI = 105105 THEN (LOGXOR LO HI) DO
    (ASET HI ARRAY I)))

(DEFMETHOD (:NEXT RANDOM-SEQUENCE) (&OPTIONAL (MASK DEFAULT-MASK))
  (LOGAND (COND ((= POINTER 70)
		 (SETF POINTER 0 (AREF ARRAY 0) (SI:%32-BIT-PLUS (AREF ARRAY 0) (AREF ARRAY 36))))
		((< POINTER 34)
		 (ASET (SI:%32-BIT-PLUS (AREF ARRAY (INCF POINTER)) (AREF ARRAY (+ POINTER 36)))
		       ARRAY POINTER))
		('T (ASET (SI:%32-BIT-PLUS (AREF ARRAY (INCF POINTER)) (AREF ARRAY (- POINTER 35)))
			  ARRAY POINTER)))
	  MASK))

(DEFMETHOD (:PREVIOUS RANDOM-SEQUENCE) (&OPTIONAL (MASK DEFAULT-MASK) &AUX (OLD (AREF ARRAY POINTER)))
  (COND ((= POINTER 0)
	 (SETF POINTER 70 (AREF ARRAY 0) (SI:%32-BIT-DIFFERENCE OLD (AREF ARRAY 36))))
	((< POINTER 35)
	 (ASET (SI:%32-BIT-DIFFERENCE OLD (AREF ARRAY (+ POINTER 36))) ARRAY POINTER)
	 (DECF POINTER))
	('T (ASET (SI:%32-BIT-DIFFERENCE OLD (AREF ARRAY (- POINTER 35))) ARRAY POINTER)
	 (DECF POINTER)))
  (LOGAND OLD MASK))

(defun step-color-map (&optional (seed 0) (ONCE 'NIL) (SLEEP 'NIL) (say 't)
		       &AUX (RANDOMS (GET-RANDOM-SEQUENCE))
			    (COLOR-START 1) (COLOR-END 128)
			    (COLOR-STEP 1) (COLOR-UNSTEP -1)
			    (RANDOM-STEP :NEXT) (RANDOM-UNSTEP :PREVIOUS))
  (SEND RANDOMS :INIT)
  (LOOP REPEAT SEED DO (SEND RANDOMS :NEXT)
		       (ROTATEF COLOR-START COLOR-END) (ROTATEF COLOR-STEP COLOR-UNSTEP)
		       (ROTATEF RANDOM-STEP RANDOM-UNSTEP))
  (loop with i = seed  with dir = :next  with rid = :previous do
     (LOOP FOR J FROM COLOR-START BY COLOR-STEP REPEAT 128
	   FOR RANDOM = (SEND RANDOMS RANDOM-STEP)
	   FOR RED = (LDB (BYTE 10 0) RANDOM)
	   FOR GREEN = (LDB (BYTE 10 10) RANDOM)
	   FOR BLUE = (LDB (BYTE 10 20) RANDOM) DO
       (SEND COLOR:COLOR-SCREEN :WRITE-COLOR-MAP-IMAGE J RED GREEN BLUE 1)
       (SEND COLOR:COLOR-SCREEN :WRITE-COLOR-MAP-IMAGE (- 256 J)
	     (- 1023 RED) (- 1023 GREEN) (- 1023 BLUE) 1))
     (SEND COLOR:COLOR-SCREEN :NEW-COLOR-MAP 'NIL 1)
     (WHEN ONCE (RETURN 'T))
     (UNLESS SLEEP
       (and say (= I SEED) (format 't "~&Space goes forward;  End ends; others backward. "))
       (when say (format't"~d "i)))
     (ROTATEF COLOR-START COLOR-END) (ROTATEF COLOR-STEP COLOR-UNSTEP)
     (ROTATEF RANDOM-STEP RANDOM-UNSTEP)
     (IF SLEEP (PROGN (incf i) (when (send *standard-input* :tyi-no-hang) (rotatef dir rid))
		      (SEND RANDOMS dir) (when (> sleep 0) (SLEEP SLEEP)))
      (SELECTOR (READ-CHAR) EQL
	(#\space (incf i) (SEND RANDOMS :NEXT))
	(#\END (RETURN-FROM step-color-map 'T))
	(OTHERWISE (decf i) (SEND RANDOMS :PREVIOUS))))))

(DEFVAR RANDOM-SEQUENCE 'NIL)

(DEFUN GET-RANDOM-SEQUENCE ()
  (OR RANDOM-SEQUENCE (SETF RANDOM-SEQUENCE (MAKE-INSTANCE 'RANDOM-SEQUENCE))))

(defflavor circ (x y eps) ()
  :settable-instance-variables)

(defmethod (:next circ) ()
  (values (setq x (- x y)) (setq y (+ y (* eps x)))))

(defmethod (:back circ) ()
  (values (setq y (- y (* eps x)) x (+ x y)) y))

(LT:UNDEFINE-VARIABLE '%)
(DEFconstant pi (FLOAT PI .0))

(DEFUN MAKE-SIN-WAVE (&KEY (AMPLITUDE 1) (PERIOD 'NIL) (PHASE 0)
			   (DTHETA (IF PERIOD (/ pi PERIOD 1/2) (/ pi 2)))
		      &AUX (SIN (SIN DTHETA)) (SIN2 (* 2 (EXPT (SIN (/ DTHETA 2)) 2)))
			   (X (* AMPLITUDE (COS PHASE)))
			   (Y (* AMPLITUDE (SIN PHASE))))
  (MAKE-INSTANCE 'CIRC :X Y :Y (+ (* Y SIN2) (* X SIN)) :EPS (* -2 SIN2)))

(DEFUN MAKE-COS-WAVE (&KEY (AMPLITUDE 1) (DTHETA (/ pi 2)) (PHASE 0)
		      &AUX (SIN (SIN DTHETA)) (SIN2 (* 2 (EXPT (SIN (/ DTHETA 2)) 2)))
			   (X (* AMPLITUDE (COS PHASE)))
			   (Y (* AMPLITUDE (SIN PHASE))))
  (MAKE-INSTANCE 'CIRC :X X :Y (- (* X SIN2) (* Y SIN)) :EPS (* 2 SIN2)))
  
(defflavor circle (c1 c2 (cx 0) (cy 0) (step ':next) (PERIOD 'NIL))
	   ()
  (:init-keywords :x :y :dt :center-of :r :phi)
  (:initable-instance-variables)
  (:GETtable-instance-variables PERIOD))

(defmethod (:reverse circle) ()
  (setq step (selectOR step EQL (:next ':back) (:back ':next))))

(DEFMETHOD (:SET-CENTER circle) (X Y)
  (SETF CX X CY Y))

(defmethod (make-instance circle :after) (&rest plist &aux (x (getf plist ':x))
				      (y (getf plist ':y))
				      (phi (getf plist ':phi))
				      (pi% (float pi (cond ((floatp x) x)
							 ((floatp y) y)
							 ((floatp phi) phi)
							 (pi))))
				      (fl0at (float 0 pi%))
				      (dt (OR (getf plist ':dt)
					      (/ pi% (getf plist :PERIOD) 1/2)))
				      (sin (sin dt))
				      (sin2 (* 2 (EXPT (sin (/ dt 2)) 2)))
				      (dawk (getf plist ':center-of))
				      (r (OR (getf plist ':r) 1)))
  (cond (x (OR Y (setq y (sqrt (- (* r r) (* x x))))))
	(y (setq x (sqrt (- (* r r) (* y y)))))
	(t (setq x (* r (cos (setq phi (if phi (- phi) 0))))
		 y (* r (sin phi)))))
  (if dawk (funcall self ':center-of dawk))
  (setF c1 (make-instance 'circ ':x (+ fl0at (setq x (if x x 0)))
				':y (+ fl0at (- (* x sin2) (* (setq y (if y y 0)) sin)))
				':eps (+ fl0at (* 2 sin2)))
	c2 (make-instance 'circ ':x (+ fl0at y)
				':y (+ fl0at (+ (* y sin2) (* x sin)))
				':eps (+ fl0at (* 2 sin2)))
	cx (+ fl0at cx)
	cy (+ fl0at cy)
	PERIOD (OR PERIOD (AND (!= 0 DT) (ABS (ROUND (* 2 pi%) DT))))))

(defmethod (:next circle) ()
  (values  (+ cx (funcall c1 ':next))
	   (+ cy (funcall c2 ':next))))

(defmethod (:point circle) ()
  (values (+ cx (funcall c1 ':x))
	  (+ cy (funcall c2 ':x))))

(defmethod (:back circle) ()
  (values (+ cx (funcall c1 ':back))
	  (+ cy (funcall c2 ':back))))

(defmethod (:step circle) ()
  (values (+ cx (funcall c1 step))
	  (+ cy (funcall c2 step))))

(defmethod (:x circle) ()
  (+ cx (funcall c1 ':x)))

(defmethod (:y circle) ()
  (+ cy (funcall c2 ':x)))

(defmethod (:center circle) ()
  (VALUES cx cy))

(defmethod (:center-of circle) (window)
  (multiple-value-SETQ (cx cy) (funcall window ':center)))

(defmethod (:set-step circle) (newd)
  (setf step newd))

(defun GROTOR (n rf dtf &optional (phif 0) (colorf #'one)ignore (w (get-window))
                 (tyi 't) &AUX LISTENER)
  #||(SEND W :EXPOSE)||# (SEND W :CLEAR-WINDOW)
                         (SETF LISTENER 
                               (IF (SEND *STANDARD-INPUT* :EXPOSED-P)
                                   *STANDARD-INPUT*
                                   (SEND W :SELECT) W))
                         (SEND-if-handles (send W :superior) :SET-PAN-AND-ZOOM 0 0 0 0)
                         (SETF colorf (funkify colorf)  dtf (funkify dtf)  rf (funkify rf) phif (funkify phif))
                         (si:with-stack-array (colors n :fill-pointer 0)
                                              (si:with-stack-array (CIRCLEs n :fill-pointer 0)
                                                                   (FLET ((SEND-CIRCLES (&REST ARGS) (LOOP FOR circle BEING THE ARRAY-ELEMENTS OF circles DO (APPLY circle ARGS))))
                                                                         (loop with cy
                                                                               with cx = (multiple-value-SETQ (nil cy) (send w :center))
                                                                               with stepping = 'nil
                                                                               with period = 'NIL
                                                                               WITH FIRST-MOVER
                                                                               WITH CHAR AND SUBPERIOD
                                                                               with count = 0
                                                                               WITH CONNECTED = 0
                                                                               WITH inc = 1
                                                                               INITIALLY (loop with x = CX and y = CY
                                                                                               for n below n
                                                                                               FOR DT = (funcall dtf n)
                                                                                               FOR R = (funcall RF n)
                                                                                               FOR CIRCLE = (make-instance 'circle
                                                                                                                           ':r R
                                                                                                                           ':dt (* pi DT)
                                                                                                                           ':phi (* pi (funcall phif n))
                                                                                                                           #|| (* pi (-  (funcall phif n)
                                                                                                                           (IF PERIOD
                                                                                                                           (* (- (FILL-POINTER cIRCLEs)
                                                                                                                           -1
                                                                                                                           FIRST-MOVER)
                                                                                                                           1/2 DT PERIOD)
                                                                                                                           0))) ||#
                                                                                                                           :CX X :CY Y)
                                                                                               WHEN (!= 0 R) DO
                                                                                               (ZL:array-push colors (funcall colorf n))
                                                                                               (ZL:array-push cIRCLEs CIRCLE)
                                                                                               (multiple-value-SETQ (X Y) (funcall circle ':point))
                                                                                               DO (OR period (SETF period (SEND CIRCLE :PERIOD)
                                                                                                                   FIRST-MOVER (FILL-POINTER cIRCLEs))))
                                                                               (SETF N (LENGTH cIRCLEs))
                                                                               do (WHEN (= 0 (mod COUNT period))
                                                                                        (IF (= CONNECTED N) (SETF STEPPING 'T)
                                                                                            (SEND (AREF CIRCLES CONNECTED) :SET-CENTER 0 0)
                                                                                            (ASET (FUNCALL COLORF CONNECTED) COLORS CONNECTED)
                                                                                            (INCF CONNECTED)))
                                                                               (AND (< CONNECTED N)
                                                                                    (SETF SUBPERIOD (SEND (AREF CIRCLES CONNECTED) :PERIOD))
                                                                                    (= 0 (mod (+ SUBPERIOD COUNT) period))
                                                                                    (ASET (- (* (ROUND COUNT SUBPERIOD)
                                                                                                (AREF COLORS CONNECTED)))
                                                                                          COLORS CONNECTED))
                                                                               (case (setq char (send LISTENER (if stepping (if tyi :tyi
                                                                                                                              (return-from grotor))
                                                                                                                 :tyi-no-hang)))
                                                                                     (#\triangle (setF stepping 'nil  INC -1) (SEND-CIRCLES :SET-STEP :BACK))
                                                                                     (#\square (setq stepping 'nil  INC 1) (SEND-CIRCLES :SET-STEP :NEXT))
                                                                                     (#\Circle (SEND-CIRCLES :REVERSE) (SETF INC (- INC)))
                                                                                     (#\space (setF stepping 't))
                                                                                     ((#\Clear-input #\Refresh) (send w :clear-window) (go si:next-loop))
                                                                                     (#\End (RETURN-FROM GROTOR 'T)))
                                                                               (loop FOR I BELOW N
                                                                                     for circle = (AREF circles I)
                                                                                     with x = cx and ox = cx and y = cy and oy = cy and ty
                                                                                     for color = (AREF colorS I) do
                                                                                     (IF (< I CONNECTED)
                                                                                         (if (= color 0)
                                                                                             (setq ox (+ ox (multiple-value-SETQ (nil ty) (funcall circle ':point)))
                                                                                                   oy (+ oy ty)
                                                                                                   x (+ x (multiple-value-SETQ (nil ty) (send circle :step)))
                                                                                                   y (+ y ty))
                                                                                           (funcall w ':aos-triangle x y  ox oy
                                                                                                    (INCF ox (multiple-value-SETQ (nil ty) (funcall circle ':point)))
                                                                                                    (INCF oy ty)
                                                                                                    color)
                                                                                           (funcall w ':aos-triangle x y  ox oy
                                                                                                    (INCF x (multiple-value-SETQ (nil ty) (send circle :step)))
                                                                                                    (INCF y ty)
                                                                                                    color))
                                                                                         (MULTIPLE-VALUE-CALL W ':aos-triangle
                                                                                                              (SEND CIRCLE :CENTER) (SEND CIRCLE :POINT) (send CIRCLE :STEP)
                                                                                                              color)))
                                                                               (INCF COUNT INC))))))

(defun rotor-XY (n Xf YF dtf &optional (colorf #'one) (w (get-window)))
  (si:with-stack-array (colors n :fill-pointer 0)
    (si:with-stack-array (CIRCLEs n :fill-pointer 0)
      (loop WITH non-0-dt = 'NIL
	  ;  with cy with cx = (multiple-value-SETQ (nil cy) (send w :center))
	    for n below n
	    FOR DT = (funcall (funkify dtf) n) DO
	(OR non-0-dt (= 0 DT) (SETF non-0-dt DT))
	(ZL:array-push colors (funcall (funkify colorf) n))
	(ZL:array-push cIRCLEs (make-instance 'circle
					      ':X (+ #||CX||# (funcall (funkify Xf) n))
					      ':Y (+ #||CY||# (funcall (funkify Yf) n))
					      ':dt (* pi DT)))
	FINALLY (ROTOR% cIRCLEs colors colors (round 2 non-0-dt) W)))))

(defun rotor (n rf dtf &optional (phif 0) (colorf1 #'one) (COLORF2 COLORF1) (w (get-window))
	      (tyi 't))
  (si:with-stack-array (colors1 n :fill-pointer 0)
    (si:with-stack-array (colors2 n :fill-pointer 0)
      (si:with-stack-array (CIRCLEs n :fill-pointer 0)
	(SETF DTF (FUNKIFY DTF)  COLORF1 (FUNKIFY COLORF1)  COLORF2 (FUNKIFY COLORF2)
	      RF (funkify rf)  PHIF (funkify phif))
	(loop WITH non-0-dt = 'NIL
	      for n below n
	      FOR DT = (funcall DTF n) DO
	  (OR non-0-dt (= 0 DT) (SETF non-0-dt DT))
	  (ZL:array-push colors1 (funcall COLORF1 n))
	  (ZL:array-push colors2 (funcall COLORF2 n))
	  (ZL:array-push cIRCLEs (make-instance 'circle
						':r (funcall rf n)
						':dt (* pi DT)
						':phi (* pi (funcall phif n))))
	      FINALLY (ROTOR% cIRCLEs colors1 COLORS2 (round 2 non-0-dt) W tyi))))))

(defun rotor% (CIRCLES COLORS1 COLORS2 PERIOD W &optional (tyi 't))
 (when tyi (SEND W :CLEAR-WINDOW))
  (loop with cy with cx = (multiple-value-SETQ (nil cy) (send w :center))
	  with stepping = 'nil
	  with count = 0
	  for inc = 1 then (- inc)
	  for back = '(#\triangle #\circle) then forth
	  and forth = '(#\square #\circle) then back do
      (loop with char do
	(loop ;WITH ZOOM-FUDGE = (1+ (SEND COLOR:COLOR-SCREEN :ZOOM))
	      with x = cx and ox = cx and y = cy and oy = cy and ty
	      for circle BEING THE ARRAY-ELEMENTS OF circles
	      for color1 BEING THE ARRAY-ELEMENTS OF colorS1
	      for color2 BEING THE ARRAY-ELEMENTS OF colorS2
	      do (if (= coloR1 0 color2)
		     (setq ox (+ ox (multiple-value-SETQ (nil ty) (funcall circle ':point)))
			   oy (+ oy ty)
			   x (+ x (multiple-value-SETQ (nil ty) (send circle :step)))
			   y (+ y ty))
		  (IF (!= 0 coloR1)
		      (funcall w ':aos-triangle x y  ox oy
			  (INCF ox (multiple-value-SETQ (nil ty) (funcall circle ':point)))
			  (INCF oy ty) color1)
		   (INCF ox (multiple-value-SETQ (nil ty) (funcall circle ':point)))
		   (INCF oy ty))
		  (IF (!= 0 coloR2)
		      (funcall w ':aos-triangle x y  ox oy
			   (INCF x (multiple-value-SETQ (nil ty) (send circle :step)))
			   (INCF y ty) color2)
		   (INCF x (multiple-value-SETQ (nil ty) (send circle :step)))
		   (INCF y ty))))
	(case (setq char (send *STANDARD-INPUT* (if (or (= 0 (mod (incf count inc) period))
							    stepping)
                                                    (if tyi :tyi (RETURN-FROM ROTOR% :tyi))
                                                  :tyi-no-hang)))
	  ((#\triangle #\square) (setq stepping 'nil))
	  ((#\Clear-input #\refresh) (send w :clear-window) (go si:next-loop))
	  (#\space (setq stepping 't))
	  (#\End (RETURN-FROM ROTOR% 'T)))
	until (member char back))
      (LOOP FOR circle BEING THE ARRAY-ELEMENTS OF circles DO (send circle :reverse))))

;(defun lose () (send *STANDARD-INPUT* (RETURN-FROM lose :tyi)))

(defun bogor (&optional (r 222)) (grotor 3 r '(+ 1/222 (* .000 n)) '(/ n 2)))

(defun spiral ()
  (rotor 5 '(if (= 0 n) 0 105) '(if (= 0 n) 1/2113 (/ (+ 1 (* 1/64 (expt -1 n))) 128))
	 '(/ (expt -1 n) 64)))

(DEFUN GIBBS-SQUARE (&OPTIONAL (R 239) (N 9) (DT .004) (colorf `(ash 1 (MOD n 8)))
			       (FUN #'GROTOR))
  (FUNCALL FUN N `(/ ,R (expt -1 n) (+ n n 1))
	  `(* (expt -1 n) ,DT(+ n n 1)) 0 colorf))

(DEFUN pool-table (&key (R 239) (N 9) (DT 1/256) (colorf `(ash 1 (MOD n 8))) (colorf1 colorf)
			       (FUN #'ROTOR) (window (get-window)))
  (FUNCALL FUN N `(/ ,R (expt -1 n) (+ n n 1))
	  `(* (ppmm n) ,DT(+ n n 1)) 1/2 colorf colorf1 window))

(defCONSTANT totem '(rotor 15. '(if (= n 0) 0 (/ 333 (EXPT -1 n) (+ n n 1)))
			    '(if (= n 0) 1/128 (* (EXPT -1 n) 1/256 (+ n n 1)))
			     0 #'1-))

(defun frac1 (r n dt
	      &optional (pow .5) (colorf `(ash 1 (MOD n 4))) (window (get-window)))
  (rotor n `(let ((n/2 (1+ (ash n -1))))
	      (/ ,r n/2 n))
	   `(* ,dt ,pow n n (EXPT -1 n) n)
	    0
	    colorf COLORF
	    window))

(defun dragon (r n dt &optional (pow .5) (colorf `(ash 1 (MOD n 4))))
  (rotor n `(let ((n/2 (1+ (ash n -1))))
	      (/ ,r n/2 n/2))
	   `(let ((n/2 (1+ (ash n -1))))
	      (* ,dt (EXPT -1 n) ,pow n/2 n/2 n/2))
	    '(+ .5 (/ n 3.0))
	    colorf)) 

(defun louse (r n dt &optional (colorf '(ash 1 (MOD n 4))))
  (rotor n `(/ ,r (float n) n) `(* ,dt (EXPT -1 n) n n n) 0 colorf))

(defun homog (r n dt z &optional (colorf '(ash 1 (MOD n 4))))
  (rotor n `(* ,r (EXPT ,z n)) `(* ,dt (1+ n)) 0 colorf))

(defun kapteyn (r n dt amp &optional (colorf '(ash 1 (MOD n 4))))
  (rotor n `(let ((n/2 (ash (1+ n) -1)))
	      (* ,r (climax::$bessel_j n (* n ,amp))))
         `(* ,dt n)
         0
         colorf))

(defun besser (&key (r 199) (n 11) (dt 1/256) (amp 4)
	       (colorf '(ash 1 (MOD n 8))) (FUN #'GROTOR) (window (get-window))
	       (tyi 't))
  (FUNCALL FUN n `(let ((n/2 (ash (1+ n) -1)))
                    (* ,r (float (climax::$bessel_j (* (EXPT -1 n) n/2) (float ,amp .0d0)) .0)))
           `(* ,dt (ash (1+ n) -1) (EXPT -1 n))
           0
           colorf colorf window tyi))

(defun bessemer () (besser :r 444 :fun #'rotor :n 19 :amp 6.5 :colorf #'one :dt (/ 512.0)))
 
(DEFUN SMOOTH-TRI (&OPTIONAL (R 190) (N 14) (DT 1/128) (COLORF (QUOTE (ASH 1 (MOD N 8))))
			     (FUN (FUNCTION GROTOR)))
  (FUNCALL FUN N `(* ,R (EXPT (/ (SELECTOR (MOD N 3) = (2 0) (T 1.0))
				 (1+ N))
			      2))
           `(* ,DT (EXPT -1.0 (MOD N 3)) (1+ N))
           0
           COLORF))

(defun gibbs-tri (&KEY (r 180) (n 14) (dt 1/128) (colorf '(ash 1 (MOD n 8))) (FUN #'GROTOR)
		  (WINDOW (GET-WINDOW)))
  (FUNCALL FUN (* 3 n) `(/ (* ,r (selectOR (mod n 6) = (0 1.0)
					   (3 -1.0)
					   (t 0)))
			   (1+ (ash n -1)))
	   `(* ,dt (EXPT -1.0 n) (1+ (ash n -1)))
	   0
	   colorf colorf
	   WINDOW))

(defun gubbs-tri (&KEY (r 180) (n 14) (dt 1/128) (colorf '(ash 1 (MOD n 8))) (FUN #'GROTOR)
		       (WINDOW (GET-WINDOW)))
  (FUNCALL FUN (* 3 n) `(/ (* ,r (selectOR (mod (+ 1 n) 6) = (0 1.0)
						       (3 -1.0)
						       (t 0)))
		     (1+ (ash (+ 1 n) -1)))
	   `(* ,dt (EXPT -1.0 (+ 1 n)) (1+ (ash (+ 1 n) -1)))
	   0
	   colorf colorf
	   WINDOW))

;; Dumb:  could have used symbolic fns for xs and ys, and called phif and grotor
(DEFUN SALAMIN-TRI (&KEY (r 72) (n 3) (dt 1/512) (colorf '(1+ N))
		       (WINDOW (GET-WINDOW)))
  (LOOP WITH ARMS = (* 2 N)
	WITH K = -1
	REPEAT N
	FOR 16K2 = (* 16 K K)
	FOR SCALE = (/ (* R 243) K (- 16K2 9) (- 16K2 81))
  COLLECT (* SCALE (- (* 675/32 (- (/ 24 (- 16K2 36)) (/ 1134 (- (* 20 K K) 405))))
		      (* 16 #.(SQRT 3) K)
		      54)) INTO XS
  COLLECT (* SCALE (+ (/ #.(* 2835/8 (SQRT 3)) (- (* K K) 9))
		      (* 16 K)
		      #.(* 54 (SQRT 3)))) INTO YS
  DO (SETF K (- 1 K)
	   16K2 (* 16 K K)
	   SCALE (/ (* R 243) K (- 16K2 9) (- 16K2 81)))
  COLLECT (* SCALE (- (* 675/32 (- (/ 24 (- 16K2 36)) (/ 1134 (- (* 20 K K) 405))))
		      (* 16 #.(SQRT 3) K)
		      54)) INTO XS
  COLLECT (* SCALE (+ (/ #.(* 2835/8 (SQRT 3)) (- (* K K) 9))
		      (* 16 K)
		      #.(* 54 (SQRT 3)))) INTO YS
  DO (SETF K (- -2 K))
    FINALLY (ROTOR-XY ARMS `(NTH N ',XS) `(NTH N ',YS)
		      `(* (- (* (EXPT -1 N) -3/2 (+ N 1/2)) 1/4) ,DT) COLORF WINDOW)))

(DEFUN SEPTIC-TRI (&KEY (r 69) (n 6) (dt 1/256) (colorf '(ASH 1 (MOD N 8)))
			(WINDOW (GET-WINDOW)) (FUN #'GROTOR))
  (FUNCALL FUN
	   N `(LET ((2N (- (* (EXPT -1 N) (- (* -3 N) 3/2)) 1/2)))
	      (/ (* 984150 ,R) (- 2N 15) (- 2N 9) (- 2N 3) 2N (+ 2N 3) (+ 2N 9) (+ 2N 15)))
	 `(* ,DT (- (* (EXPT -1 N) (- (* -3/2 N) 3/4)) 1/4))
	 '(LSH (+ N 1) -1) COLORF COLORF WINDOW))

(defun pots-tri (&KEY (r 239) (n 8) (dt 1/256) (colorf '(if (> n 3) 1 -1)) (FUN #'ROTOR)
		       (WINDOW (GET-WINDOW)))
  (FUNCALL FUN (1+ (* 2 n)) `(selector n =
			       (0 0)
			       (1 (/ ,r pi 10/27 (sqrt 1)))
			       (2 (/ ,r pi 10/27 (sqrt 1)))
			       (t (let* (rem (quo (multiple-value-setq (nil rem)
						    (ceiling (- n 3) 2)))
					 (n (* quo (+ rem rem 1))))
				    (/ ,r (- n 1/6) (- n -1/3) (- n -5/6) 8 pi))))
	   `(selector n =
	      (0 (/ ,dt 2))
	      (1 (* 3/2 ,dt)) (2 (* -3/2 ,dt))
	      (t (let* (rem (quo (multiple-value-setq (nil rem)
				   (ceiling (- n 3) 2)))
			(n (* quo (+ rem rem 1))))
		   (* ,dt (1+ (* 3 n))))))
	   '(selector n =
	      (2 2/2) (1 2/2) (t 0))
	   colorf colorf
	   WINDOW))

(defun stop-tri (&KEY (r 377) (n 7) (dt 1/256) (colorf '(ash 1 (mod n 8))) (FUN #'gROTOR)
		       (WINDOW (GET-WINDOW)))
  (FUNCALL FUN n `(let* (rem (quo (multiple-value-setq (nil rem)
					       (ceiling n 2)))
				    (n (* quo (+ rem rem 1))))
			       (/ ,r (- n 1/6) (- n -1/3) (- n -5/6) 8 pi))
	   `(let* (rem (quo (multiple-value-setq (nil rem)
					       (ceiling n 2)))
		   (n (* quo (+ rem rem 1))))
	      (* ,dt (1+ (* 3 n))))
	   0
	   colorf colorf WINDOW))

(defun sham (&optional (p 3) &KEY (r 177) (n 7) (dt 1/256) (WINDOW (GET-bw-WINDOW))
	      (FUN #'ROTOR) (colorf (if (eql window bw-window) 1 '(ash 1 (mod n 8)))))
  (funcall fun n `(let ((n (0-11-2 n)) (pi% (float pi .0)))
		    (/ (* ,r (expt -1 n))
		       (+ (* ,p n) 1) (+ n  (/ 1 ,p) -1/2) pi% pi%))
	  `(* ,dt (+ 1 (* ,p (0-11-2 n)))) 0 colorf colorf WINDOW))

(defun foo-tri (r n dt)
  (rotor n `(/ (* ,r (selectOR (MOD n 6) = (0 #.(sqrt 27))
						 (1 -2)
						 (2 2)
						 (3 #.(- (sqrt 27)))
						 (t 0)))
	       (1+ (ash n -1)))
	   `(* ,dt (EXPT -1.0 n) (1+ (ash n -1)))
	   `(selectOR (MOD n 6) = ((0 3) .5) (t 0))))

(ZL:COMMENT (DEFUN SNOWFLAKE (&KEY (R 222) (DT 1/1024) (COLORF #'1+) (N 20))
  (ROTOR N `(* ,R (NTH N '(+9.329330529694819E-1 8.252786998457396E-2 +9.194509591124072E-2
			   -3.128256314362122E-2 -6.991123664861624E-3
			   3.079807072400421E-2 -4.369723935459711E-3
			   9.22974442096716E-3 +1.955205944894159E-2 -3.489660395964418E-4
			    1.877137327476354E-2 2.7648E-4 -5.580896654169479E-3
			    -.00452 .00308 -.00887 -.00293 .00470 -6.397E-4 -.00459)))
	 `(* ,DT (NTH N '(-1 5 -7 11 -13 17 -19 23 -25 29 -31 35 -37 41 -43 47 -49 53 -55 59)))
	0 COLORF))

(ZL:COMMENT "
[[2.857492335182093B-6*i-1.652027287210555B-6,-3.299547937849943B-6*i-5.580896654169479B-3]
 [4.654501120886363B-6-8.052617939648641B-6*i,9.298362270274791B-6*i+1.877137327476354B-2],
 [3.908875128555903B-6-6.76412866294438B-6*i,7.810543008770132B-6*i+1.955205944894159B-2],
 [1.148912445458179B-6*i-6.637900715261396B-7,-1.326649819323737B-6*i-4.369723935459711B-3],
 [1.257678573097028B-6*i-7.264694421846252B-7,-1.452242125462298B-6*i-6.991123664861624B-3],
 [5.14348503156436B-6-8.906476596269674B-6*i,1.028431332078094B-5*i+9.194509591124072B-2],
 [7.453928075653419B-6-1.291010573191186B-5*i,1.490730603913998B-5*i+9.329330529694819B-1],
 [5.710182129979018B-6*i-3.29616693252652B-6,8.252786998457396B-2-6.593550379727087B-6*i],
 [2.748130830332119B-6-4.761834953187866B-6*i,5.498493384122668B-6*i-3.128256314362122B-2],
 [7.245223099046855B-6*i-4.180407771307389B-6,3.079807072400421B-2-8.366063013148621B-6*i],
 [2.93762834380999B-6*i-1.694601318929232B-6,9.22974442096716B-3-3.39208103016127B-6*i],
 [8.076719478371897B-8-1.400427120655889B-7*i,1.617073950194236B-7*i-3.489660395964418B-4]]")
)
(DEFUN SNOWFLAKE (&KEY (R 222) (DT 1/1024) (COLORF #'INTEGER-LENGTH) (N 128) (WINDOW (GET-WINDOW)) (fun #'rotor))
  (funcall fun N `(* ,R (NTH N '(9.3294434957E-1 8.2522874136E-2 9.1952890302E-2 -3.1278397568E-2 -6.9922244544E-3 3.0791733573E-2 -4.3707296643E-3 9.2271753237E-3 1.9557981385E-2 -3.4884358184E-4 
 1.8778424221E-2 2.7628736072E-4 -5.5833990386E-3 -4.5186824834E-3 3.0838369088E-3 -8.866805153E-3 -2.9354729374E-3 4.7018067326E-3 -6.4018895551E-4 -4.5914864745E-3 
 9.3043021351E-6 8.4997062158E-3 8.2821166286E-6 4.8103867919E-3 -4.5074460558E-4 -5.5632810376E-4 -1.6260500259E-3 -4.1390078054E-4 1.3310822489E-3 1.9637575361E-3 
 -1.8497788926E-3 1.8516651033E-3 4.6673580673E-3 -5.0627288966E-4 3.5214041584E-3 2.3933447468E-4 -5.3861762647E-4 -1.7649723259E-4 -5.3204311219E-4 -2.479326298E-5 
 3.4052801503E-3 1.2799277624E-7 4.4661867805E-3 -1.2076329225E-7 -1.7886212972E-3 2.0819208402E-5 1.3508829931E-3 1.3181860505E-4 -1.8611994842E-3 -1.5880922306E-4 
 -6.8290364216E-4 2.9799004314E-4 2.9427602626E-5 -9.6474716959E-4 -2.5386030891E-5 -9.0324514628E-4 4.3793899186E-4 1.6750265483E-4 8.8438987192E-4 1.9727086585E-4 
 -4.7296563681E-4 -1.4869841172E-3 4.5763461543E-4 -2.2761801259E-3 -8.2571350858E-4 1.0570002473E-3 -4.4815989253E-4 -9.2154511079E-4 4.8873635993E-5 1.4616441865E-3 
 3.365298313E-5 6.1664885494E-4 -1.4457818035E-4 -3.0561433647E-5 -1.2012519826E-4 3.0372858711E-5 2.7906843982E-5 -6.0554973815E-4 -1.0630983413E-5 -1.4200102581E-3 
 5.7902896838E-6 8.8763684872E-4 5.0377103959E-7 -1.0126113816E-3 -8.9435612037E-10 2.178419402E-3 -8.6955158642E-10 1.4302108228E-3 4.6167618523E-7 -1.9221461801E-4 
 5.0060879372E-6 -1.6712559749E-4 -8.6698107335E-6 9.3648336357E-4 2.1463598937E-5 1.0610262478E-3 -8.7111427369E-5 -3.5827686278E-4 -9.8823754263E-5 2.1878096221E-4 
 2.1674152745E-5 -2.2570790862E-4 2.9646169197E-5
 -5.25295975E-5 -2.55926916E-4 8.04759935E-7 -4.43641302E-4 7.47455313E-7 2.31188777E-4 -4.20851438E-5 -2.24499025E-4 -1.55907867E-4 3.94112891E-4 1.30214829E-4 1.83062151E-4 
 -1.83563001E-4 -9.94408034E-6 4.67375529E-4 1.07901381E-5 3.54110875E-4 -2.34109757E-4 -5.41457916E-5 -5.95740244E-4 -5.32382901E-5 4.03127365E-4 3.37770948E-4 
 -4.96809896E-4 4.37356517E-4)))
	 `(* ,DT (+ (* (EXPT -1 N) -3 (+ N 1/2)) 1/2))
	 0 COLORF colorf WINDOW))

(defun berzerkle (&key (r 105) (scale .03125) (window (get-window)) (fudge .7072))
  (loop with cx = (/ (send window :width) 2)
	with cy = (/ (send window :height) 1.1)
;	with scale = (/ scale r)
	for x = r then new-x
	for y = 0 then new-y
	for new-x = (- x (* scale (identity y) (exp (/ y r fudge))))
	for new-y = (+ y (* scale (identity new-x) (exp  (* 0 (/ new-x fudge r))))) do
    (when (> (max (abs new-x) (abs new-y)) (* 99 r)) (break))
    (send window :aos-triangle cx cy (+ cx x) (+ cy y) (+ cx new-x) (+ cy new-y))))

(defun zerkle (&key (r 222) (scale .03125) (window (get-window)))
  (loop with cx = (/ (send window :width) 2)
	with cy = (/ (send window :height) 2)
	with scale = (/ scale 1 r r)
	for x = r then new-x
	for y = 0 then new-y
	for new-x = (- x (* scale y y y))
	for new-y = (+ y (* scale new-x new-x new-x)) do
    (when (> (max (abs new-x) (abs new-y)) (* 9 r)) (si:break))
    (send window :aos-triangle cx cy (+ cx x) (+ cy y) (+ cx new-x) (+ cy new-y))))

(defun zerkle (&key (r 222) (scale .03125) (window (get-window)))
  (loop with cx = (/ (send window :width) 2)
	with cy = (/ (send window :height) 2)
	with scale = (/ scale 1); r r)
	with x = r  with y = 0 do
    (when (> (max (abs x) (abs y)) (* 29 r)) (si:break))
    (send window :aos-triangle cx cy (+ cx x) (+ cy y)
	  (+ cx (decf x (* scale y))); y y)))
	  (+ cy (incf y (* scale x)))))); x x))))))

(DEFUN CLIMAX::FUNNYZET (N)
  (LOOP FOR K FROM 0 TO 99999
	FOR 3K+2 = (+ (* K 3) 2)
	FOR -3K+1 = (- 1 3K+2)
	SUM (+ (/ (EXPT -3.0 (SI:2-POWERS-IN 3K+2)) (EXPT (FLOAT 3K+2) N))
	       (/ (EXPT -3.0 (SI:2-POWERS-IN -3K+1)) (EXPT (FLOAT -3K+1) N)))))

(ZL:COMMENT(DEFUN FLAKIER (K S)
  (LOOP FOR N FROM 1
	FOR PROD = 1 THEN (* PROD TERM)
	FOR TERM = (- 1 (* S (TAN (* PI (/ (+ K 1/3) (EXPT -2.0D0 N))))))
	WHEN (= TERM 1) RETURN (/ PROD (* (SQRT (/ 4.0D0 3)) (EXPT (* PI (+ K 1/3)) 2)))))


 (DEFUN CLIMAX::FLAKIER (Z &OPTIONAL (S #.(-(SQRT 3.0D0))))
  (LOOP FOR PROD = 1 THEN (* PROD TERM)
	FOR ARG = (* pi Z -1/2) THEN (/ ARG -2)
	FOR TERM = (+ 1 (* S (TAN ARG)))
	WHEN (= TERM 1) RETURN PROD))

 (DEFUN CLIMAX::DFLAKIER (Z &OPTIONAL (S #.(-(SQRT 3.0))))
  (LOOP FOR ARG = (* pi Z -1/2) THEN (/ ARG -2)
	FOR TAN = (TAN ARG)
	FOR DEN = (+ 1 (* S TAN))
	UNTIL (< (ABS ARG) .00000001) SUM (/ TAN DEN)))

 (DEFUN CLIMAX::WHAT (Z)
  (* Z (LOOP FOR 3K-1 FROM -169 BY 3 TO 170
	 SUM (/ (EXPT -2 (- (SI:2-POWERS-IN 3K-1) 1)) (- (/ 3K-1 3) Z) 3K-1))))

 (DEFUN CLIMAX::GUESS (Z &OPTIONAL (AMP (/ pi 2)))
  (LOOP FOR PROD = AMP THEN (/ PROD GUNCH)
	FOR TERM = (* 3 Z) THEN (/ TERM -2)
	FOR GUNCH = (+ 1 TERM)
	WHEN (= GUNCH 1.0) RETURN PROD))

 (DEFUN CLIMAX::GUESS (Z &OPTIONAL (AMP 1/2))
  (* AMP (- (LOOP WITH 3Z = (* 3 Z)
		  FOR TERM = 1 THEN (* TERM -2)
		  SUM (/ 3 (+ 3Z TERM))
		  UNTIL (> (ABS TERM) 9999999.9))
	    0)))

 (DEFUN CLIMAX::GUESS2 (Z &OPTIONAL (AMP 1/10))
  (* AMP (- (LOOP WITH 3Z = (* 3 Z)
		  FOR TERM = 5 THEN (* TERM -2)
		  SUM (/ 3 (- 3Z TERM))
		  UNTIL (> (ABS TERM) 9999999.9))
	    0)))

 (DEFUN CLIMAX::GUESS3 (Z &OPTIONAL (AMP 1/14))
  (* AMP (- (LOOP WITH 3Z = (* 3 Z)
		  FOR TERM = 7 THEN (* TERM -2)
		  SUM (/ 3 (+ 3Z TERM))
		  UNTIL (> (ABS TERM) 9999999.9))
	    0)))

 (DEFUN CLIMAX::GUESS4 (Z &OPTIONAL (AMP 1/22))
  (* AMP (- (LOOP WITH 3Z = (* 3 Z)
		  FOR TERM = 11 THEN (* TERM -2)
		  SUM (/ 3 (- 3Z TERM))
		  UNTIL (> (ABS TERM) 9999999.9))
	    0)))

 (DEFUN CLIMAX::GUESS5 (Z &OPTIONAL (AMP 1/26))
  (* AMP (- (LOOP WITH 3Z = (* 3 Z)
		  FOR TERM = 13 THEN (* TERM -2)
		  SUM (/ 3 (+ 3Z TERM))
		  UNTIL (> (ABS TERM) 9999999.9))
	    0)))

 (DEFUN CLIMAX::GUESS6 (Z &OPTIONAL (AMP 1/34))
  (* AMP (- (LOOP WITH 3Z = (* 3 Z)
		  FOR TERM = 17 THEN (* TERM -2)
		  SUM (/ 3 (- 3Z TERM))
		  UNTIL (> (ABS TERM) 9999999.9))
	    0)))
)

(DEFUN FLAKIER (K S)
  (LOOP FOR N FROM 1
	FOR PROD = 1 THEN (* PROD TERM)
	FOR TERM = (- 1 (* S (TAN (* pi (/ (+ K 1/3) (EXPT -2 N))))))
	WHEN (= TERM 1) RETURN (/ PROD (* #.(SQRT 4/3) (EXPT (* pi (+ K 1/3)) 2)))))

(DEFUN NOSEFLAKE (&KEY (R 444) (DIMENSION (LOG 4 3)) (DT 1/1024) (COLORF1 #'INTEGER-LENGTH)
		       (COLORF2 COLORF1) (N 128) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (S (SQRT (- (EXPT 4 (- 1 (/ DIMENSION))) 1))))
  (ROTOR (+ N 1) `(* ,R (if (= 0 n) 0
			 (FLAKIER (- (* (EXPT -1 N) (- 1/2 N) 1/2) 1/4) ,S)))
	 `(* ,DT (if (= n 0) ,(/ 1 how-far)
		  (+ (* (EXPT -1 N) (- 3/4 (* 3/2 N))) 1/4)))
	 -1/3 COLORF1 COLORF2 WINDOW))

(DEFUN frillIER (K S)
  (LOOP FOR N FROM 2
	FOR PROD = (EXPT (* pi (+ K 1/2)) -2) THEN (* PROD TERM)
	FOR TERM = (+ 1 (* S (TAN (* pi (/ (+ k K 1) (EXPT -2 N))))))
	WHEN (= TERM 1) RETURN PROD))

(DEFUN NOSEfrill (&KEY (R 384) (DIMENSION 2) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 128) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (S (SQRT (- (EXPT 4 (- 1 (/ DIMENSION))) 1)))
		       (fun #'rotor))
  (funcall fun (+ N 1) `(* ,R (if (= 0 n) 0 (frillIER (0-11-2 (- n 1)) ,S)))
	 `(* ,DT (if (= n 0) ,(/ 1 how-far)
		  (+ (* (EXPT -1 N) (- 1/2 n)) 1/2))) 0 COLORF1 COLORF2 WINDOW))

(DEFUN frillem (K m S)
  (LOOP with k+1/m = (+ k (/ 1 m))
	FOR N FROM 1
	FOR PROD = (EXPT k+1/m -2) THEN (* PROD TERM)
	for gunch = (/ k+1/m (EXPT -2 N))
	FOR TERM = (- 1 (* S (TAN (* pi gunch))))
	WHEN (< (min (abs prod) (abs gunch)) single-float-epsilon)
	  RETURN (* (sin (/ pi m)) (expt pi -2) PROD)))

(zl:comment(DEFUN mOSEfrill (&KEY (R 256) (DIMENSION 2) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 33) (how-far 1) (m 2)
		       (WINDOW (GET-BW-WINDOW)) (S (SQRT (- (EXPT 4 (- 1 (/ DIMENSION))) 1)))
		       (fun #'rotor) (phase 0))
  (funcall fun (+ N 1)
	   `(* ,R (if (= 0 n) 0 (frillem (0-11-2 (- n 1)) ,m ,S)))
	   `(* 2 ,DT (if (= n 0) ,(/ 1 m how-far) (+ (/ 1 ,m) (0-11-2 (- n 1)))))
	 (+ phase (/ -1 m)) COLORF1 COLORF2 WINDOW))
)
(DEFUN mOSEfrill (&KEY (R 256) (DIMENSION 2) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 33) (how-far 1) (m 5)
		       (WINDOW (GET-BW-WINDOW)) (S (SQRT (- (EXPT 4 (- 1 (/ DIMENSION))) 1)))
		       (fun #'rotor) (phase 0) (erase 't))
  (funcall fun (+ N 1)
	   `(* ,R (if (= 0 n) 0 (abs (frillem (0-11-2 (- n 1)) ,m ,S))))
	   ;`(* 2 ,DT (if (= n 0) ,(/ 1 m how-far) (+ (/ 1 ,m) (0-11-2 (- n 1)))))
	   `(* 2 ,DT (if (= n 0) ,(/ 1 how-far) (+ 1 (* ,m (0-11-2 (- n 1))))))
	 `(+ ,phase ,(/ -1 m) (/ (phase (frillem (0-11-2 (- n 1)) ,m ,S)) ,pi))
	 COLORF1 COLORF2 WINDOW erase))

(DEFUN mOSEfrills (x y &KEY (R 256) (DIMENSION 2) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 33) (m 5)
		       (WINDOW (GET-BW-WINDOW)) (S (SQRT (- (EXPT 4 (- 1 (/ DIMENSION))) 1)))
		       (fun #'rotor) (clear&tyi 'nil) (how-far 1))
  (funcall fun (+ N 2)
	   `(* ,R (if (= 0 n) 0
		   (if (= n 1) ,(sqrt (+ (* x x) (* y y))) (frillem (0-11-2 (- n 2)) ,m ,S))))
	   `(* 2 ,DT (if (= n 0) ,(/ 1 m how-far)
		      (if (= n 1) 0 (+ ,(/ 1 m) (0-11-2 (- n 2))))))
	   `(if (< n 2) ,(/ (atan y x) pi) ,(/ -1 m)) COLORF1 COLORF2 WINDOW clear&tyi))

;-(SIN(%PI/M)*(-1)^N*S+COS(%PI/M)*(-1)^N+1)/(2*%PI*(N+1/M-1)*(N+1/M)*(N+1/M+1))
;:m -5/2 :n 4 :phase -1/2 :colorf 1 :window (get-window ) :fun #'grotor :s (tan (/ pi 5)))
(defun sawteeth (&key (m 5) (r 222) (dt 1/128) (colorf 1) (n 33) (s (sqrt 3)) ;(how-far 1)
		 (fun #'rotor) (WINDOW (GET-BW-WINDOW)) (phase 0) (x 0) (y 0)
		 &aux (d (sqrt (+ (* x x) (* y y)))) (sc (+ (* s (sin (/ pi m))) (cos (/ pi m)))))
  (funcall fun (+ n 2)
	   `(case n
	      (0 0)
	      (1 ,d)
	      (t (let ((k (0-11-2 (- n 2))))
		   (* ,(/ r -2 pi) (/ (+ 1 (* (expt -1 k) ,sc))
				     (+ k ,(/ m)) (+ k ,(/ m) -1) (+ k ,(/ m) 1))))))
	   `(case n
	      (0 (/ ,dt 1/2 ,m))
	      (1 0)
	      (t (* 2 ,dt (+ (/ 1 ,m) (0-11-2 (- n 2))))))
	   `(if (or (= ,d 0) (!= n 1)) ,phase ,(* pi (atan y x)))
	   colorf colorf WINDOW))

(defun oldsaw (&key (m 5) (r 222) (dt 1/128) (colorf 1) (n 33) (s (sqrt 3)) ;(how-far 1)
		 (fun #'rotor) (WINDOW (GET-BW-WINDOW)) (phase 0) ;(x 0) (y 0)
		 &aux  (sc (+ (* s (sin (/ pi m))) (cos (/ pi m)))))
  (funcall fun n
	   `(let ((k (0-11-2 n)))
	      (* ,(/ r -2 pi) (/ (+ 1 (* (expt -1 k) ,sc))
				(+ k ,(/ m)) (+ k ,(/ m) -1) (+ k ,(/ m) 1))))
	   `(* 2 ,dt (+ ,(/ m) (0-11-2 n)))
	   phase colorf colorf WINDOW))

(defun mander (&KEY (R 300) (DT 1/2048) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 99); (how-far 1)
		       (WINDOW (GET-bw-WINDOW)) ;(S 1)
		       (fun #'rotor))
  (funcall fun n `(* ,r (if (= n 0) 1 ;climax::(macsyma-to-lisp ($mand (- user::n 1)))
			  (mandb (- n 1))))
	   `(* ,dt (- 1 n))
	   0 COLORF1 COLORF2 WINDOW))

(defun mandb (m) (/ (mndb 0 (+ 1 m)) (expt 2 (+ m m 1))))

; B[N,M]:=IF N >= 1 AND 1 <= M AND M <= 2^(N+1)-2 THEN 0
;         ELSE 2^(2^(N+2)-1)*B[N+1,M]
;             -2^(2^(N+2)-4)*SUM(B[N,K]*B[N,M-K],K,2^(N+1)-1,-2^(N+1)+M+1)
;             -2*B[0,-2^(N+1)+M+1]

(defvar mandarays (make-array 0 :fill-pointer 0
			      :initial-element
			      (make-array 0 :fill-pointer 0 :initial-element 0)))
(defun mndb (n m)
  (if (= 0 m) 1/2
   (let ((2^n+1-1 (- (expt 2 (+ n 1)) 1)))
     (if (and (> n 0) (< m 2^n+1-1)) 0
      (loop repeat (- n (fill-pointer mandarays) -1) do
	(vector-push-extend (make-array 0 :fill-pointer 0 :initial-element nil) mandarays))
      (let* ((nvec (aref mandarays n))
	     (m- (- m 2^n+1-1)))
	(loop repeat (- m- (fill-pointer nvec) -1) do (vector-push-extend nil nvec))
	(or (aref nvec m-)
	    (setf (aref nvec m-)
		  (let* ((2^2^n+2-4 (expt 2 (+ 2^n+1-1 2^n+1-1 -2))))
		    (- (* 8 2^2^n+2-4 (mndb (+ 1 n) m))
		       (* 2^2^n+2-4 (loop for k from 2^n+1-1 to m-
					  sum (* (mndb n k) (mndb n (- m k)))))
		       (* 2 (mndb 0 m-)))))))))))

(defun mosearray (&key (window (get-bw-window)) (across 5) (down 4) (m 4) (s -1) (dt 1/256))
  (send window :clear-window)
  (loop with n = 0
	with r = (min (/ (send window :inside-width) across 2)
		      (/ (send window :inside-height) down 2))
	for y from (- down 1) by -2 repeat down do
    (loop for x from (- 1 across) by 2 repeat across do
      (mosefrills x y :n (incf n) :r r :m m :s s :clear&tyi 'nil :dt dt :window window))))

(zl:comment(defun drag1 (k &optional (s 1) (bits 22))
  (loop with m0 = (make-array '(2 2) :initial-contents
			      `((,(/ (complex 1 (- s)) 2) ,(/ (complex 1 s) 2)) (0 1)))
	with m1 = (make-array '(2 2) :initial-contents
			      `((,(/ (complex -1 (- s)) 2) ,(/ (complex -1 s) 2)) (0 1)))
	with m1m0 = (math:multiply-matrices m1 m0)
	for n downfrom bits to 1
	for om = (cis (* pi k -1/2 (expt 1/2 bits))) then om^2
	for om^2 = (* om om)
	for old-an = #2a((1 0) (0 1)) then an
	for an = old-an then
	 (math:combine-matrices
	   (math:multiply-matrices
	     (math:combine-matrices m0 m1 'nil #'(lambda (x y) (/ (+ x (* om y)) 2)))
	     an)
	   (math:combine-matrices an
	     (math:multiply-matrices m1m0 old-an) 'nil
	     #'(lambda (ignore y) (* y (* om (- om^2 1) 1/4)))))
	finally (return (/ (aref an 0 1) 1))))

 (defun drag2 (k &optional (s 1) (bits 22))
  (loop with m0 = (make-array '(2 2) :initial-contents
			      `((,(/ (complex 1 (- s)) 2) ,(/ (complex 1 s) 2)) (0 1)))
	with m1 = (make-array '(2 2) :initial-contents
			      `((,(/ (complex -1 (- s)) 2) ,(/ (complex -1 s) 2)) (0 1)))
	repeat bits
	for om = (cis (* pi k -1 (expt 1/2 bits))) then (* om om)
	for an = #(1 1) then
	 (math:combine-matrices
	   (math:multiply-matrices m0 an)
	   (math:multiply-matrices m1 bn)
	   'nil #'(lambda (x y) (+ x (* om y))))
	and bn = #(-1 1) then
	 (math:combine-matrices
	   (math:multiply-matrices m1 an)
	   (math:multiply-matrices m0 bn)
	   'nil #'(lambda (x y) (+ x (* om y))))
	finally (return (/ (aref an 0) (expt 2 (- bits 1))))))
)
(defvar *drag1-hashes* (make-hash-table))

(defun drag1 (k s &optional (bits 22))
  (modify-hash *drag1-hashes* s
   #'(lambda (ignore val &rest ignore)
       (if val
	   (return-from drag1
	     (modify-hash val k
	      #'(lambda (ignore val &rest ignore)
		  (if val (return-from drag1 val)
		   (loop with 1+is = (/ (complex 1 s) 4)  with 1-is = (/ (complex -1 s) -4)
			 repeat (+ 1 bits)
			 for pow = (* -1 pi k) then (/ pow 2)
			 for om = (cis pow)
			 for an = #2a((1 0 0)) then
			     (math:multiply-matrices an
			      (make-array '(3 3) :initial-contents
					  `((,1-is ,(* -1 om 1+is) ,(- 1+is (* om 1-is)))
					    (,(- 1+is) ,(* om 1-is) ,(- (* om 1+is) 1-is))
					    (0 0 ,(/ (+ 1 om) 2)))))
			 finally (return (aref an 0 2)))))))
	(make-hash-table))))
  (drag1 k s bits))

(defsubst 11-13-3 (n) (+ (* (EXPT -1 N) (- 1/2 n)) 1/2))
 ; = (+ 1 (* 2 (0-11-2 (- n 1))))

(defun furd1 (x y &KEY (R 256) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 8)
		       (WINDOW (GET-BW-WINDOW)) (S 1)
		       (fun #'rotor) (clear&tyi 'nil))
  (funcall fun (+ N 1) `(* ,R (if (= 0 n) ,(sqrt (+ (* x x) (* y y)))
			 (abs (drag1 (11-13-3 n) ,S))))
	 `(* ,DT (if (= n 0) 0 (11-13-3 n)))
	 `(if (= 0 n) ,(/ (atan y x) pi) (/ (phase (drag1 (11-13-3 n) ,S)) pi))
	 COLORF1 COLORF2 WINDOW clear&tyi))

(defun tri-in-coef (n) climax::(macsyma-to-lisp ($tricircopt user::(0-11-2 n))))

(defun tri-out-coef (n)
	   (+ (tri-in-coef n) climax::(macsyma-to-lisp ($fuj user::(0-11-2 n)))))

(defun tricirc (&KEY (r 180) (n 14) (dt 1/256) (colorf '(ash 1 (MOD n 8))) (FUN #'GROTOR)
		  (WINDOW (GET-WINDOW)) (clear&tyi 't) (in 't))
  (flet ()
  (FUNCALL FUN N `(* ,r (abs (,(if in `tri-in-coef 'tri-out-coef) n)))
	   `(* ,dt (0-11-2 n))
	   `(/ (phase (,(if in `tri-in-coef 'tri-out-coef) n)) pi)
	   colorf colorf WINDOW clear&tyi)))

(DEFUN furdragon (&KEY (R 256) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 8) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (S 1)
		       (fun #'rotor))
  (funcall fun (+ N 1) `(* ,R (if (= 0 n) 0
			 (abs (drag1 (11-13-3 n) ,S))))
	 `(* ,DT (if (= n 0) ,(/ 1 how-far) (11-13-3 n)))
	 `(if (= 0 n) 0 (/ (phase (drag1 (11-13-3 n) ,S)) pi))
	 COLORF1 COLORF2 WINDOW))

(defun tdrag1 (k &optional (z #c(#.(sqrt 1/3) 1)) (m 3) (trits 13))
  (loop with z/2 = (/ z 2)  with c = (/ 1 (tan (/ pi m)))
	repeat (+ 1 trits)
	for pow = (* -1 pi (+ k (/ 2 m))) then (/ pow 3)
	for om = (cis pow)
	for an = #2a((1 0)) then
	 (math:multiply-matrices an (make-array '(2 2) :initial-contents
			 `((,(- om (* #c(0 1) z/2 (expt (- om 1) 2)))
			    ,(+ c #c(0 -1) (* om om (+ c #c(0 1)))
				(* 1/2 (- om 1) (- (* c #c(0 1) (- om 1)) om 1) z)))
			   (0 ,(+ 1 (* om (+ 1 om)))))))
	finally (return (/ (aref an 0 1) (expt 3 trits)))))
;(W-1)*(i*(W-1)*C-W-1)*Z/2+W^2*(C+i)+C-i$

(defun tdrag2 (k &optional (z #c(#.(sqrt 1/3) 1)) (m 3))
  (loop with zi/2 = (/ z #c(0 -2))  with c = (/ 1 (tan (/ pi m)))  with prd = 1
	with k+2/m = (+ k (/ 2 m))
	for pow = (* -1/3 pi k+2/m) then (/ pow 3)
	for om = (cis pow)
	when (= prd (setf prd (* prd (- om (* zi/2 (expt (- 1 om) 2))))))
	  return (/ (+ c #c(0 -1) (* (cis (* pi k+2/m -1)) -1 (+ c #c(0 1))) (* #c(0 2) prd))
		    #c(0 1) pi k+2/m)))

(defun tdrag3 (k &optional (z #c(#.(sqrt 1/3) 1)) (m 3))
  (loop with 2iz = (* z #c(0 2))  with prd = 1
	with pik/2 = (* pi (+ k (/ 2 m)) 1/2)
	for pow = (/ pik/2 3) then (/ pow 3)
	when (= prd (setf prd (* prd  (+ 1 (* 2iz (expt (sin pow) 2))))))
	  return (/ (+ (/ (sin pik/2) (tan (/ pi m))) (- (cos pik/2)) prd) pik/2 (cis pik/2))))

(DEFUN tdragon (&KEY (R 256) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 16) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (z #c(#.(sqrt 1/3) 1)) (m 3)
		       (fun #'rotor) (tyi 't))
  (funcall fun (+ N 1) `(* ,(* R (sin (/ pi m))) (if (= 0 n) 0
			 (abs (tdrag3 (00-22 n) ,z ,m))))
	 `(* ,DT (if (= n 0) ,(/ 2 m how-far) (+ (00-22 n) ,(/ 2 m))))
	 `(if (= 0 n) 0 (/ (phase (tdrag3 (00-22 n) ,z ,m)) pi))
	 COLORF1 COLORF2 WINDOW tyi))

(defun rdrag2 (k &optional (z #c(#.(sqrt 1/3) 1)) (m 3))
  (loop with zi/2 = (/ z #c(0 -2))  with c = (/ 1 (tan (/ pi m)))  with prd = 1
	with k+2/m = (+ k (/ 2 m))
	for pow = (* -1/3 pi k+2/m) then (/ pow 3)
	for om = (cis pow)
	when (= prd (setf prd (* prd (- om (* zi/2 (expt (- 1 om) 2))))))
	  return (/ (+ c #c(0 -1) (* (cis (* pi k -1)) -1 (+ c #c(0 1))) (* #c(0 2) prd))
		    #c(0 1) pi k+2/m)))

(defsubst 00-22 (n) (+ (* (EXPT -1 N) (- 1/2 n)) -1/2))

(zl:comment(DEFUN conef (&KEY (x 0) (y 0) (phi 0) (R 256) (DT 1/256) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 6) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)); (z #c(#.(sqrt 1/3) 1))
		       (fun #'rotor) (clear&tyi 't) &allow-other-keys)
  (funcall fun (+ N 2) `(if (= 0 n) 0 (* ,r (if (= n 1) ,(sqrt (+ (* x x) (* y y)))
						(abs (conef1 (0-11-2 (- n 1)))))))
	 `(* ,DT (if (= n 0) ,(/ 1 how-far) (0-11-2 (- n 1))))
	 `(if (< n 2) ,(if (= 0 x y) 0 (/ (atan y x) pi))
		   (/ (+ ,phi #|(/ ,m)|# (phase (conef1 (0-11-2 (- n 1))))) pi))
	 COLORF1 COLORF2 WINDOW clear&tyi))

 (defun conef1 (n)
 (/ (climax::macsyma-to-lisp
      (if (and (< -3 n 5) (evenp n))
	  (funcall (aref ;climax::#(#.#'$coneh1 #.#'$coneh2 #.#'$coneh3 #.#'$coneh4)
			 (+ 1 (ash n -1))))
	  (multiple-value-bind (n k) (floor n 8)
	    (funcall (aref ;climax::#(#.#'$coneg0 #.#'$coneg1 #.#'$coneg2 #.#'$coneg3
			 ;	     #.#'$coneg4 #.#'$coneg5 #.#'$coneg6 #.#'$coneg7)
			   k) n))))
    2 2.15))
)
(DEFUN rdrag1 (&KEY (x 0) (y 0) (phi 0) (R 256) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 16) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (z #c(#.(sqrt 1/3) 1)) (m 6)
		       (fun #'rotor) (clear&tyi 't))
  (funcall fun (+ N 2) `(* ,(* R (sin (/ pi m)))
			   (if (= 0 n) 0 (if (= n 1) ,(sqrt (+ (* x x) (* y y)))
						(abs (tdrag3 (00-22 (- n 1)) ,z ,m)))))
	 `(* ,DT (if (= n 0) ,(/ 2 m how-far) (if (= n 1) 0 (+ (00-22 (- n 1)) ,(/ 2 m)))))
	 `(if (< n 2) ,(if (= 0 x y) 0 (/ (atan y x) pi))
		   (/ (+ ,phi #|(/ ,m)|#  (phase (tdrag3 (00-22 (- n 1)) ,z ,m))) pi))
	 COLORF1 COLORF2 WINDOW clear&tyi))

(DEFUN trifl1-sideways (k &optional (z #c(1/2 1/2)) (m 2))
  (loop with ct = (if (typep (realpart z) 'single-float) 24 56)	;need to be even
	with pik/2 = (* pi (+ k (/ 1 m)) 1/1)
	for w = (cis (/ pik/2 (expt 2 ct))) then (* w w)
	for z0 = z then z1 and z1 = (conjugate z) then z0
	for num = 0 then (* 1/2 (+ (* num (+ z0 (* (- 1 z0) w))) (* den z0 w)))
	for den = 1 then (* den 1/2 (+ 1 w))
		repeat ct 
	finally (return num)))			;(/ num den)

(DEFUN trifl1 (k &optional (z #c(1/2 1/2)) (m 2))
  (loop with ct = (if (typep (realpart z) 'single-float) 24 56)	;need to be even
	with pik/2 = (* pi (+ k (/ 1 m)) 1/1)
	for w = (cis (/ pik/2 (expt 2 ct))) then (* w w)
	for z0 = (/ z 2) then z1 and z1 = (conjugate (/ z 2)) then z0
	for iz0 = (* #c(0 1) z) then iz1 and iz1 = (conjugate (* #c(0 1) z)) then iz0
	for c0 = (/ (- z #c(0 1)) 2) then c1 and c1 = (conjugate (/ (- z #c(0 1)) 2)) then c0
	for num = 0 then (* 1/2 (+ (* num (+ iz0 (* (- -1 iz0) w)))
				   (* den (+ c0 (* z0 w)))))
	for den = 1 then (* den 1/2 (+ 1 w))
		repeat ct
	finally (return (* m (- num (/ (* (cos (/ pi m)) (cis (/ pi m))) 2 pik/2))))))

(DEFUN trifle (&KEY (x 0) (y 0) (phi 0) (R 1111) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 16) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (z (* (cos (* pi 2/9)) (cis (* pi 2/9)))) (m 9)
		       (fun #'rotor) (clear&tyi 't))
  (funcall fun (+ N 2) `(* ,(* R (sin (/ pi m)))
			   (if (= 0 n) 0 (if (= n 1) ,(sqrt (+ (* x x) (* y y)))
						(abs (trifl1 (0-11-2 (- n 2)) ,z ,m)))))
	 `(* ,DT (if (= n 0) ,(/ 1 m how-far) (if (= n 1) 0 (+ (0-11-2 (- n 2)) ,(/ 1 m)))))
	 `(if (< n 2) ,(if (= 0 x y) 0 (/ (atan y x) pi))
		   (+ ,phi #|(/ ,m)|# (/ (phase (trifl1 (0-11-2 (- n 2)) ,z ,m)) pi)))
	 COLORF1 COLORF2 WINDOW clear&tyi))

(DEFUN zzerpf (&KEY (x 0) (y 0) (phi 0) (R 256) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 16) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (z #c(#.(sqrt 1/3) 1)) (m 6)
		       (fun #'rotor) (clear&tyi 't) (theta 1/2))
  (funcall fun (+ N 2) `(* ,(* R (sin (/ pi m)))
			   (if (= 0 n) 0 (if (= n 1) ,(sqrt (+ (* x x) (* y y)))
						(abs (tdrag3 (00-22 (- n 1)) ,z ,m)))))
	 `(* ,DT (if (= n 0) ,(/ 2 m how-far) (if (= n 1) 0 (+ (00-22 (- n 1)) ,(/ 2 m)))))
	 `(if (< n 2) ,(if (= 0 x y) 0 (/ (atan y x) pi))
	   (+ ,phi (* (+ (* ,m (00-22 (- n 1))) 2) ,theta)
	      (/ (phase (tdrag3 (00-22 (- n 1)) ,z ,m)) pi)))
	 COLORF1 COLORF2 WINDOW clear&tyi))

(defun zurpfle (&key (across 5) (window (get-bw-window)) (phi 1/2)
		     (fun #'(lambda (n) (climax::$fib (+ 1 n)))) (scale 1)
		&aux (width (send window :inside-width)) (height (send window :inside-height))
		     (dx (floor height across))
		     (r (* scale dx)) (ds (/ dx r)))
  (loop with y = (/ (- (* 4/3 dx) height) 2 r)
	for x from (/ (- (* (+ across 2/3) dx) (/ width 2)) r) by (- ds)
	for n from 1 to across  for fn = (funcall fun n) do
    (zzerpf :r r :phi phi :n fn :dt (expt 1/2 (+ 6 (floor (log fn)))) :m 2 :how-far 1/6
	    :x x :y y :theta 1/4 :clear&tyi 'nil :window window)
    finally
  (loop for y from y by ds
	for n from (+ 1 across) to (+ across across -1)  for fn = (funcall fun n) do
    (zzerpf :r r :phi phi :n fn :dt (expt 1/2 (+ 6 (floor (log fn)))) :m 2 :how-far 1/6
	    :x x :y y :theta 1/4 :clear&tyi 'nil :window window)
    finally
      (loop for x from x by ds
	    for n from (* 2 across) to (* 3 across)  for fn = (funcall fun n) do
	(zzerpf :r r :phi phi :n fn :dt (expt 1/2 (+ 6 (floor (log fn)))) :m 2 :how-far 1/6
	    :x x :y y :theta 1/4 :clear&tyi 'nil :window window)))))

(defun tdragons (&key (across 6) (window (get-bw-window)) (dt 1/256) (fun #'rdrag1)
		 (how-far 1/2) (phi 0) (m 2)
		 &aux (width (send window :inside-width))
		      (height (send window :inside-height))
		      (r (/ width (+ across across 1)))
		      (down (ceiling (- (/ height r (sqrt 3)) 1/3))))
  (send window :clear-window)
  (loop with n = 1
	for x0 = (- 2/3 across) then (- 7/3 (* 2 across) x0)
	for y from (* 1/2 (sqrt 3) (- down 1)) by (- (sqrt 3))
	repeat down do
    (loop for x from x0 by 2
	  repeat across do
      (funcall fun :phi phi :x x :y y :r r :dt dt :n n :window window :how-far how-far
	      :colorf1 (if (eql (send window :superior) tv:main-screen) 1 n) :m m
	      :clear&tyi 'nil)
      (incf n))))

(DEFUN rdragon (&KEY (R 256) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 16) (how-far 1)
		       (WINDOW (GET-BW-WINDOW)) (z #c(#.(sqrt 1/3) 1)) (m 6)
		       (fun #'rotor))
  (funcall fun (+ N 1) `(* ,(* R (sin (/ pi m))) (if (= 0 n) 0
			 (abs (tdrag2 (00-22 n) ,z ,m))))
	 `(* ,DT (if (= n 0) ,(/ 2 m how-far)
		  (+ (00-22 n) ,(/ 2 m))))
	 `(if (= 0 n) 0 (/ (+ (/ ,m) (phase (tdrag2 (00-22 n) ,z ,m)))
			   pi))
	 COLORF1 COLORF2 WINDOW))

#||(defun equifou (&KEY (R 256) (DT 1/256) (COLORF1 #'one) (COLORF2 COLORF1) (N 16) (how-far 1)
		(WINDOW (GET-BW-WINDOW)) (fun #'rotor))
  (funcall fun (+ N 1) `(* ,r (if (= 0 n) 0
			       (let ((r (climax::macsyma-to-lisp (nth (ceiling (+ 2 n) 2)
								      climax::$d28))))
				 (if (oddp n) (- (realpart r) (imagpart r))
				  (+ (realpart r) (imagpart r))))))
	   `(* ,DT (if (= n 0) ,(/ 1 how-far) (0-11-2 n)))
	   0 #||`(if (= 0 n) 0 (+ (/ (phase (climax::macsyma-to-lisp (nth (ceiling (+ 2 n) 2)
								     climax::$d26)))
				pi)
			     n))||#
	   COLORF1 COLORF2 WINDOW))||#

;[	   2   1			     ]
;[	  w  - -			 2   ]
;[   1	       2     w	     w	    1 - w    ]
;[   -	  ------   - -	  - ---	    ------   ]
;[   6	    s	     6	    2 s	      6	     ]
;[					     ]
;[				   2	     ]
;[  2   1			  w  + 1     ]
;[ w  - -			  ------ + w ]
;[      2     1	    w	     w	    2	     ]
;[ ------   - -	   ---	   - -	  ---------- ]
;[   s	     6	   2 s	     6	      s	     ]
;[					     ]
;[			       2	     ]
;[			      w		     ]
;[ 		     2	  1 - --     2	     ]
;[    w	     w	    w	      2	    w  - 1   ]
;[  - -	  - ---	    --	  ------    ------   ]
;[    6	    2 s	    6	    s	      6	     ]
;[					     ]
;[		       2	   2	     ]
;[		      w		  w  + 1     ]
;[		  1 - --      2	  ------ + w ]
;[   w	     w	      2	     w	    2	     ]
;[  ---	   - -	  ------   - --	  ---------- ]
;[  2 s	     6	    s	     6	      s	     ]
;[					     ]
;[				   2	     ]
;[				  w  + w + 1 ]
;[   0	    0	    0	    0	  ---------- ]
;[				      3	     ]
;
; s := sqrt(27), w := cis(- pi j/3^n)

(defun equifoo (&KEY (R 256) (DT 1/256) (COLORF1 #'one) (COLORF2 COLORF1) (N 16) (how-far 1)
		(WINDOW (GET-BW-WINDOW)) (fun #'rotor)(clear&tyi 't))
  (funcall fun (+ N 1) `(* ,r (equi5 (0-11-2 n) 11))
	   `(* ,DT (if (= n 0) ,(/ 1 how-far) (0-11-2 n)))
	   0
	   COLORF1 COLORF2 WINDOW clear&tyi))

(defun equi5 (j terms)
  (loop with s = (sqrt 27)
	for n from terms downto 1
	for w = (cis (- (/ (* pi j) (expt 3 n)))) then (expt w 3)
	for w^2 = (expt w 2)
	for -w/6 = (/ w -6)  for w/2s = (/ w 2 s)  for -w/2s = (- w/2s)
	for v1 = 0 then (+ (/ v1 6) (/ (* v2 (- (* w w) 1/2)) s) (* v3 -w/6) (* v4 -w/2s)
			   (* v5 (- 1 w^2) 1/6))
	and v2 = 0 then (+ (/ (* v1 (- (* w w) 1/2)) s) (/ v2 -6) (* v3 w/2s) (* v4 -w/6)
			   (* v5 (/ (+ w^2 1 w w) 2 s)))
	and v3 = 0 then (+ (* v1 -w/6) (* v2 -w/2s) (* v3 w^2 1/6) (* v4 (/ (- 2 w^2) 2 s))
			   (* v5 (- w^2 1) 1/6))
	and v4 = 0 then (+ (* v1 w/2s) (* v2 -w/6) (* v3 (/ (- 2 w^2) 2 s)) (* v4 w -w/6)
			   (* v5 (/ (+ w^2 1 w w) 2 s)))
	for v5 = 1 then (* v5 (+ 1 w w^2) 1/3)
	finally (return (+ (realpart v1) (imagpart v2)))))

;0,-1,1,-2,2,...

;(defsubst 0-11-2 (n) n)

(defsubst 0-11-2 (n) (- (* (EXPT -1 N) (+ 1/2 n) 1/2) 1/4))

(defsubst ppmm (n) (expt -1 (ash n -1)))

(defun astri (&KEY (R 288) (DT 1/512) (WINDOW (GET-BW-WINDOW)))
  (rotor 2 `(/ ,r (+ 1 n n)) `(* ,dt (- 1 (* 4 n))) 0 1 1 window 't)
  ;(rotor 2 `(* ,r n 1/3) `(* ,dt (- 6/5 n)) 0 1 1 window 'nil)
  (rotor 3 `(* ,r (nth n '(0 1/3 1))) `(* ,dt (nth n '(6 1 -1/3))) 0 1 1 window 'nil)
  #|(rotor 3 `(* ,r (nth n '(0 1/3 1))) `(* ,dt (nth n '(6 1 -1/3))) '(nth n '(0 0 2/3))
	 1 1 window 'nil)|#
  #|(rotor 3 `(* ,r (nth n '(0 1/3 1))) `(* ,dt (nth n '(6 1 -1/3))) '(nth n '(0 0 4/3))
	 1 1 window 'nil)|#
  (rotor 3 `(* ,r (nth n '(0 1/3 1))) `(* ,dt (nth n '(3 0 1))) '(nth n '(0 1/3 -1/9))
	 1 1 window 'nil)
  (rotor 3 `(* ,r (nth n '(0 1/3 1))) `(* ,dt (nth n '(3 0 -1))) '(nth n '(0 1/3 -1/9))
	 1 1 window 'nil))

(defun tdr (s phi m)
  (loop for k below m
	for j = (+ (0-11-2 k) phi)
	sum (/ (- 1 (loop with prd = 1
			  for threep = 3 then (* 3 threep)
			  until (= prd (setf prd (* prd (+ 1 (* s (expt (sin (/ (* pi j)
										threep))
									2))))))
			  finally (return prd)))
	       j)))

(defsubst v-p-e-value (val vec) (vector-push-extend val vec) val)

(defvar footab (make-array 22 :adjustable 't :fill-pointer 0 :initial-element 'nil))

(defun reset-footab ()
  (fill footab 'nil) (setf (fill-pointer footab) 0) (vector-push-extend 0 footab))

; That's a bad sine,
(defvar Sinbad 0)

(defun foozero (s phi m)
  (or (and (> (fill-pointer footab) m) (aref footab m))
      (flet ((fook (z)
	       (loop with rez = (realpart z)
		     with iters = (ceiling (+ (max (number-length s)
						   #.(number-length 1/6))
					      (float-precision rez)
					      9 (number-length rez))
					   (log 8 2))	;9
		     for pro1 = 0 then (+ pro1 (* (+ 1 pro1) s sin sin))
		     repeat iters
		     for sin = (/ z (expt 3 iters)) then (* sin (- 3 (* 4 sin sin)))
		     finally (si:comment(when (> (setf sin (abs (- sin (sin (/ z 3))))) sinbad)
			       (format 't "~% z=~a sin(z)=~a iters=~a error=~a"
				       (/ z 3) (sin (/ z 3)) iters (setf sinbad sin))))
			     (return pro1))))
	(setf sinbad 0)
	(v-p-e-value (+ (foozero s phi (- m 1))
			       (print(loop with pi% = (float pi (realpart phi))
				   ;  with max = 0
				     repeat (* 4/3 (expt 3 m))
				     for k from (- (* 2/3 (expt 3 m)) 2)
				     for k+phi = (+ phi (0-11-2 k))
				     for fook = (fook (* pi% k+phi))
				     sum (/ fook k+phi)
			   ;when (> (abs fook) (abs max)) do (format 't " ~a" (setf max fook))
					 )))
			    footab))))

(defun cscsum (t0 dt &optional (term-fun #'identity))
  (loop with tand = (tan (/ dt 2))  with sinn = (sin t0)
	with cosn = (- (cos t0) (* tand sinn))
	with sind = (sin dt)  with 2tan = (* 2 tand)
	with sum = 0 do
    (when (= sum (incf sum (funcall term-fun (expt sinn -2)))) (return sum))
    (decf cosn (* 2tan (incf sinn (* sind cosn))))))

(defvar accels (make-array '(11 33)))

(defun foocer (s a phi m &optional (x 1) &aux (res (realpart s)))
  (setf (fill-pointer footab) 0) (clear-accels)
  (stack-let* ((pi% (float pi res))
	       (n-circles (ceiling (+ (* m (log 3 2))
				      (max (number-length s) #.(number-length 1/6))
				      (float-precision res))
				   (log 9 2)))
	       (sin+n (make-array n-circles :fill-pointer 0))
	       (sin-n (make-array n-circles :fill-pointer 0))
	       (cos+n (make-array n-circles :fill-pointer 0))
	       (cos-n (make-array n-circles :fill-pointer 0))
	       (2tand (make-array n-circles :fill-pointer 0))
	       (sind (make-array n-circles :fill-pointer 0)))
    (loop repeat n-circles
	  for pi/3^n = (/ pi (expt 3 n-circles)) then (* pi/3^n 3)
	  for dt = (* a pi/3^n)  for phin = (* phi pi/3^n)
	  for tan2 = (tan (/ dt 2)) 
	  for sin = (sin phin)  for cos = (- (cos phin) (* tan2 sin)) do
      (vector-push sin sin+n) (vector-push sin sin-n)
      (vector-push cos cos+n) (vector-push cos cos-n)
      (vector-push (* 2 tan2) 2tand) (vector-push (sin dt) sind))
    (flet ((fook (sinner)		 ; 1-Prod(1+s (sin (k+phi)pi/3^n)^2,n1)
	     (loop for pro1 = 0 then (+ pro1 (* (- pro1 1) s sin sin))
		   for i below n-circles
		   for sin = (funcall sinner i)
		   finally (return pro1)))
	   (sin-adv (i &aux (sin (aref sin+n i)))
	     (decf (aref cos+n i)
		   (* (aref 2tand i)
		      (setf (aref sin+n i) (+ sin (* (aref sind i) (aref cos+n i))))))
	     sin)
	   (ret-sin (i)
	     (decf (aref sin-n i)
		   (* (aref sind i)
		      (incf (aref cos-n i) (* (aref 2tand i) (aref sin-n i)))))))
      (loop with sum = 0  for n from 1 to m
	    for 3^n-1 = 1 then (* 3 3^n-1) do
	(vector-push-extend (print (loop repeat (* 2 3^n-1)
					 with y = (expt x (- 3^n-1 1))
					 for k from (- 3^n-1 1) do
				     (incf sum (+ (/ (* y (fook #'sin-adv)) (+ (* a k) phi))
						  (/ 0 (- phi (* a (+ k 1))) (setf y (* y x)))
						  ))
				     finally (return sum)))
			    footab)))))

(defvar foolist '())

(defun fooloop (n acc &optional (skip 0) (lo 2) (hi 3) (like (+ .0 lo hi))
		&aux (phi (/ (+ 1 (sqrt (float 5 like))) 2)) (wid (- hi lo)))
  (loop repeat n
	for x = (+ 1/2 (* skip phi)) then (+ x phi)
	for y = (+ lo (* wid (setf x (mod x 1)))) do
	(push (list y (progn (format 't "~%~d" y)
			     (foocero (float 239/1000 like)
				      y (float 288/1000 like) acc)
			     (aref footab (- acc 1)))
		    acc)
	      foolist)))

(defun foocero (s a phi m &optional (x 1) &aux (res (realpart s)))
  (setf (fill-pointer footab) 0) (clear-accels)
  (stack-let* ((pi% (float pi res))
	       (n-circles (ceiling (+ (* m (log 3 2))
				      (max (number-length s) #.(number-length 1/6))
				      (float-precision res))
				   (log 9 2)))
	       (sin+n (make-array n-circles :fill-pointer 0))
	       (sin-n (make-array n-circles :fill-pointer 0))
	       (cos+n (make-array n-circles :fill-pointer 0))
	       (cos-n (make-array n-circles :fill-pointer 0))
	       (2tand (make-array n-circles :fill-pointer 0))
	       (sind (make-array n-circles :fill-pointer 0)))
    (loop repeat n-circles
	  for pi/3^n = (/ pi% (expt 3 n-circles)) then (* pi/3^n 3)
	  for dt = (* a pi/3^n)  for phin = (* phi pi/3^n)
	  for tan2 = (tan (/ dt 2)) 
	  for sin = (sin phin)  for cos = (- (cos phin) (* tan2 sin)) do
      (vector-push sin sin+n) (vector-push sin sin-n)
      (vector-push cos cos+n) (vector-push cos cos-n)
      (vector-push (* 2 tan2) 2tand) (vector-push (sin dt) sind))
    (flet ((fook (sinner)		 ; 1-Prod(1+s (sin (k+phi)pi/3^n)^2,n1)
	     (loop for pro1 = 0 then (+ pro1 (* (- pro1 1) s sin sin))
		   for i below n-circles
		   for sin = (funcall sinner i)
		   finally (return pro1)))
	   (sin-adv (i &aux (sin (aref sin+n i)))
	     (decf (aref cos+n i)
		   (* (aref 2tand i)
		      (setf (aref sin+n i) (+ sin (* (aref sind i) (aref cos+n i))))))
	     sin)
	   (ret-sin (i)
	     (decf (aref sin-n i)
		   (* (aref sind i)
		      (incf (aref cos-n i) (* (aref 2tand i) (aref sin-n i)))))))
      (loop with sum = 0  for n from 1 to m
	    for 3^n-1 = 1 then (* 3 3^n-1) do
	(loop repeat (* 2 3^n-1)
	      with y = (expt x (- 3^n-1 1))
	      for k from (- 3^n-1 1) do
	  (incf sum (+ (/ (* y (fook #'sin-adv)) (+ (* a k) phi))
		       (/ (fook #'ret-sin) (- phi (* a (+ k 1))) (setf y (* y x))))))
	(format 't " ~d" sum)
	(vector-push-extend sum footab)))))

(defun ofoocero (s a phi m &aux (res (realpart s)))
  (setf (fill-pointer footab) 0) (clear-accels)
  (flet ((fook (circles)		 ; 1-Prod(1+s (sin (k+phi)pi/3^n)^2,n1)
	   (loop for pro1 = 0 then (+ pro1 (* (- pro1 1) s sin sin))
		 for circle being the array-elements of circles
		 for sin = (second-value (send circle :step))
		 finally (return pro1))))
    (stack-let* ((pi% (float pi res))
		 (n-circles (ceiling (+ (* m (log 3 2))
					     (max (number-length s)
						  #.(number-length 1/6))
					     (float-precision res))
				     (log 9 2)))
		 (circles (make-array n-circles :fill-pointer 0))
		 (selcric (make-array n-circles :fill-pointer 0)))
      (loop repeat n-circles
	    for 3^n = (expt 3 n-circles) then (/ 3^n 3)
	    for period = (/ 3^n a 1/2) do
	(vector-push (make-instance 'circle :period period :phi (/ (* pi% (- phi a)) 3^n))
		     circles)
	(vector-push (make-instance 'circle :period period :phi (/ (* pi% phi) 3^n)
				    :step :back)
		     selcric))
      (loop with sum = 0  for n from 1 to m
	    for 3^n-1 = 1 then (* 3 3^n-1) do
	(vector-push-extend (print (loop repeat (* 2 3^n-1)
					 for k from (- 3^n-1 1) do
				     (incf sum (+ (/ (fook circles) (+ (* a k) phi))
						  (/ (fook selcric) (- phi (* a (+ k 1))))))
				     finally (return sum)))
			    footab)))))

(defun shanks (k n &aux (kn (array-dimensions accels)))
  (if (< k 0) (aref footab n)
   (when (or (> k (first kn)) (> n (second kn)))
     (adjust-array accels (list (max k (first kn)) (max n (second kn)))
		   :initial-element 'nil))
   (or (aref accels k n)
       (setf (aref accels k n)
	     (let* ((a (shanks (- k 1) (- n 1)))
		    (b (shanks (- k 1) n))
		    (c (shanks (- k 1) (+ n 1))))
	       (/ (- (* a c) (* b b))
		  (+ a (* -2 b) c)))))))

(defun clear-accels ()
  (stack-let ((wiper (make-array (array-total-size accels) :displaced-to accels)))
    (fill wiper nil)))

(DEFUN NEAR-RT3 (&KEY (R 444) (DIMENSION (LOG 4 3)) (DT 1/1024) (COLORF1 #'INTEGER-LENGTH)
		       (COLORF2 COLORF1) (N 128)
		       (WINDOW (GET-BW-WINDOW)) (S (- (SQRT (- (EXPT 4 (- 1 (/ DIMENSION))) 1))))
		 &AUX (RHO (/ (SQRT (- 1 (* #.(SQRT 3) S))) 2))
		      (LOGRHO (LOG RHO))
		      (M (- 1 (FLOOR (LOG (+ RHO (* N (- 1 RHO)))) LOGRHO)))
		      (B (/ (* N (- 1 RHO)) (- 1 (EXPT RHO M))))
		      FREQ)
  (FLET ((3K+1 (NU &AUX (N (PRINT (+ 1 (FLOOR (LOG (- 1 (/ (* NU (- 1 RHO)) B))) LOGRHO))))
			(L (PRIN1 (- NU (CEILING (* B (- 1 (EXPT RHO (- N 1)))) (- 1 RHO))))))
	   (SETF FREQ (* (EXPT 2 (- N 1)) (+ (* 6 (- 1/4 (* (EXPT -1 L) 1/2 (+ 1/2 L))))
				  3 (* 2 (EXPT -1 N))))))
	 (FREAK () FREQ))
    (ROTOR N `(* ,R (FLAKIER (/ (- (,#'FREAK) 1) 3) ,S))
	   `(* ,DT (,#'3K+1 N)) -1/3 COLORF1 COLORF2 WINDOW)))

(DEFUN sierp (&KEY (R 256) (DT 1/1024) (COLORF1 #'one)
		       (COLORF2 COLORF1) (N 33) (how-far 1) (m 3)
		       (WINDOW (GET-BW-WINDOW)) 
		       (fun #'rotor) (phase 0) (erase 't))
  (funcall fun (+ N 1)
	   `(* ,R (if (= 0 n) 0 (realpart (sierp1 (0-11-2 (- n 1)) ,m))))
	   ;`(* 2 ,DT (if (= n 0) ,(/ 1 m how-far) (+ (/ 1 ,m) (0-11-2 (- n 1)))))
	   `(* 2 ,DT (if (= n 0) ,(/ 1 how-far) (+ 1 (* ,m (0-11-2 (- n 1))))))
	 ;`(+ ,phase ,(/ -1 m) (/ (phase (sierp1 (0-11-2 (- n 1)) ,m)) ,pi))
	   phase
	 COLORF1 COLORF2 WINDOW erase))

(defun sierp1 (k m)
 (si:without-floating-underflow-traps
   (loop with pi/3 = (/ pi 3)  with x0 = (* 2 pi (+ k (/ m)))
	 for x = x0 then (/ x 3)
	 for a11 = 1 then (- (/ a11 2) (* a12 (cos (+ x pi/3))))
	 and a12 = 0 then (- (/ a12 2) (* a11 (cos (- x pi/3))))
	 and a21 = 0 then (- (/ a21 2) (* a22 (cos (+ x pi/3))))
	 and a22 = 1 then (- (/ a22 2) (* a21 (cos (- x pi/3))))
	 when (< (abs (+ a11 a12)) .1d-9) return (/ a11 x0 1/4))))

;       [        1               x    %pi  ]
;  inf  [        -         - cos(-- - ---) ]
; /===\ [        2                n    3   ]
;  | |  [                        3         ]
;  | |  [                                  ]
;  | |  [       x    %pi          1        ]
; n = 1 [ - cos(-- + ---)         -        ]
;       [        n    3           2        ]
;       [       3                          ]

(zl:comment(defun sierp1 (k m &aux (x (* 1(+ k (/ m)))))
  (si:without-floating-underflow-traps
    (/ (* 2 climax::(macsyma-to-lisp ($ze (lisp-to-macsyma (* 2 pi user::x)))))
       pi x)))

 (defun sierp1 (k m &aux (x (* 1(+ k (/ m))))) (si:without-floating-underflow-traps
  (+ climax::(macsyma-to-lisp ($sierp (lisp-to-macsyma (* user::x 1))))
     (/ (cos (* (float pi .0) x)) (* 1 (float pi .0) x)))))
)
(DEFUN RECFLAKE (&KEY (S 1.5) (R 444) (STEPS 1024) (WINDOW (GET-BW-WINDOW)) (PHASE 'NIL)
		 &AUX CY (CX (MULTIPLE-VALUE-SETQ (NIL CY) (SEND WINDOW :CENTER)))
	 	      (A (- -1/2 (* S #C(0 1/2))))
		      (P (* (+ 1 (/ S (SQRT (FLOAT 3 (INCF S .00))))) (CIS (/ pi 6))))
		      (ABAR (CONJUGATE A)) (PBAR (CONJUGATE P)))
  (LABELS ((REC1 (U)
	     (IF (= U 0) (/ (+ PBAR (* P ABAR)) (- 3 (* S S)) 1/4)
	      (IF (> U 1/2) (+ (* A (REC1 (- 2 U U))) P)
			    (+ (* ABAR (REC1 (- 1 U U))) PBAR))))
	   (TRIDRAW (Z1 Z2)
	     (SEND WINDOW :AOS-TRIANGLE CY CX
		   (- CY (ROUND (IMAGPART Z1))) (- CX (ROUND (REALPART Z1)))
		   (- CY (ROUND (IMAGPART Z2))) (- CX (ROUND (REALPART Z2))) 1)))
    (SEND WINDOW :CLEAR-WINDOW)
    (LOOP WITH DT = (/ STEPS)
	  with phase = (mod (OR PHASE DT) dt)
	  WITH Z = (* R (REC1 (float phase s)))
	  FOR U FROM phase BY DT UPTO 1 DO
      (TRIDRAW Z (SETF Z (* R (REC1 (float U s))))))))

(DEFUN wRECFLAKE (&KEY (S 1.0) (STEPS 1024) (WINDOW (GET-BW-WINDOW)) (PHASE 'NIL)
		       (R (/ 1 (+ 1 (sqrt 1/3)))) (scale 1) (dz #c(0 0)) (stepping 'nil)
		 &AUX height (width (MULTIPLE-VALUE-SETQ (NIL height)
				      (SEND WINDOW :INSIDE-SIZE)))
	 	      (A (- -1/2 (* S #C(0 1/2))))
		      (P (* (+ 1 (/ S (SQRT (FLOAT 3 (INCF S .00))))) (CIS (/ pi 6))))
		      (ABAR (CONJUGATE A)) (PBAR (CONJUGATE P))
		      cx cy)
  (setf dz (* dz height)  r (* r scale height #c(0 1))
	dz (complex (/ (* (REALPART dz) width) height) (IMAGPART dz))
	cx (ROUND (REALPART dZ))
	cy (- height (ROUND (IMAGPART dz))))
  (when (send window :exposed-p) (send window :select))
  (LABELS ((REC1 (U)
	     (IF (= U 0) (/ (+ PBAR (* P ABAR)) (- 3 (* S S)) 1/4)
	      (IF (> U 1/2) (+ (* A (REC1 (- 2 U U))) P)
			    (+ (* ABAR (REC1 (- 1 U U))) PBAR))))
	   (TRIDRAW (Z1 Z2)
	     (incf z1 dz) (incf z2 dz)
	     (SEND WINDOW :draw-TRIANGLE CX CY
		  (ROUND (REALPART Z1)) (- height (ROUND (IMAGPART Z1)))
		  (ROUND (REALPART Z2)) (- height (ROUND (IMAGPART Z2))) tv:alu-xor)))
    (SEND WINDOW :CLEAR-WINDOW)
    (LOOP WITH DT = (/ STEPS)
	  with phase = (mod (OR PHASE DT) dt)
	  WITH Z = (* R (REC1 0))
	  FOR U FROM phase BY DT below 1 DO
      (TRIDRAW Z (SETF Z (* R (REC1 (float U s)))))
      (and stepping (send window :exposed-p)
       (when (eql #\square (send window :tyi))
	 (setf stepping 'nil)))
      FINALLY (TRIDRAW Z (* r (REC1 1)))))
  (when (send window :exposed-p) (send window :tyi) (send window :deselect)))

(defun dimension (s) (/ 1 (- 1 (log (+ 1 (* s s)) 4))))

(DEFUN triflake (&KEY (dimension 2) (STEPS 1024) (WINDOW (GET-BW-WINDOW)) (PHASE 'NIL)
			       (R (/ 1 (+ 1 (sqrt 3)))) (scale 1)
			       (S (sqrt (- (expt 4 (- 1 (/ 1 dimension))) 1)))
			       (dz (complex .5 (/ 2 (+ 3 (sqrt 3)))))
		 &AUX height (width (MULTIPLE-VALUE-SETQ (NIL height)
				      (SEND WINDOW :INSIDE-SIZE)))
	 	      (A (- -1/2 (* S #C(0 1/2))))
		      (P (* (+ 1 (/ S (SQRT (FLOAT 3 s)))) (CIS (/ pi 6))))
		      (ABAR (CONJUGATE A)) (PBAR (CONJUGATE P))
		      cx cy (dt (/ 1 STEPS)))
  (setf dz (* dz height)  r (* r scale height #c(0 1))
	dz (complex (/ (* (REALPART dz) width) height) (IMAGPART dz))
	cx (ROUND (REALPART dZ))
	cy (- height (ROUND (IMAGPART dz)))
	phase (mod (OR PHASE DT) dt))
  (when (send window :exposed-p) (send window :select))
  (LABELS ((REC1 (U)
	     (IF (= U 0) (/ (+ PBAR (* P ABAR)) (- 3 (* S S)) 1/4)	;1/sqrt3 - i
	      (IF (> U 1/2) (+ (* A (REC1 (- 2 U U))) P)
			    (+ (* ABAR (REC1 (- 1 U U))) PBAR))))
	   (TRIDRAW (Z1 Z2)
	     (incf z1 dz) (incf z2 dz)
	     (SEND WINDOW :draw-TRIANGLE CX CY
		  (ROUND (REALPART Z1)) (- height (ROUND (IMAGPART Z1)))
		  (ROUND (REALPART Z2)) (- height (ROUND (IMAGPART Z2))) tv:alu-xor)))
    (SEND WINDOW :CLEAR-WINDOW)
    (LOOP WITH Z = (* R (REC1 0))
	  FOR U FROM phase BY DT below 1 DO
      (TRIDRAW Z (SETF Z (* R (REC1 (float U s)))))
      FINALLY (TRIDRAW Z (* r (REC1 1))))
    (LOOP with rom = (* r (cis (* pi 2/3)))
	  WITH Z = (* R (REC1 1))
	  FOR U FROM phase BY DT below 1 DO
      (TRIDRAW Z (SETF Z (* Rom (REC1 (float U s)))))
      FINALLY (TRIDRAW Z (* r (- (sqrt 4/3)))))
    (LOOP with rom = (* r (cis (* pi -2/3)))
	  WITH Z = (* R (- (sqrt 4/3)))
	  FOR U FROM phase BY DT below 1 DO
      (TRIDRAW Z (SETF Z (* Rom (REC1 (float U s)))))
      FINALLY (TRIDRAW Z (* R (REC1 0)))))
  (when (send window :exposed-p) (send window :tyi) (send window :deselect)))

(DEFUN wRECPEANO (&KEY (STEPS 1024) (WINDOW (GET-BW-WINDOW)) (pause 't)
		 (scale 1) (dz #c(0 0)) (R 1) (close 't)
		 (PHASE 'NIL) (fun #'hash-peano)	;or 30-60-90-fill
		 &AUX CY (CX (FLOOR (MULTIPLE-VALUE-SETQ (NIL CY) (SEND WINDOW :INSIDE-SIZE))
				    2)))
  (setf dz (* dz cy)  dz (complex (/ (* (REALPART dz) 2 cx) cy) (IMAGPART dz))
	r (* r scale cy))
  (when (send window :exposed-p) (send window :select))
  (LABELS ((TRIDRAW (Z1 Z2)
	     (incf z1 dz) (incf z2 dz)
	     (SEND WINDOW :draw-TRIANGLE CX CY
		   (ROUND (REALPART Z1)) (- CY (ROUND (IMAGPART Z1)))
		   (ROUND (REALPART Z2)) (- CY (ROUND (IMAGPART Z2))) tv:alu-xor)))
    (SEND WINDOW :CLEAR-WINDOW)
    (LOOP WITH Z = (* R (funcall fun 0))
	  WITH DT = (/ STEPS)
	  FOR U FROM (OR PHASE DT) BY DT UPTO 1 DO
      (TRIDRAW Z (SETF Z (* R (funcall fun U))))
      FINALLY (when close (TRIDRAW Z R))))
  (when (send window :exposed-p) (when pause (send window :tyi)) (send window :deselect)))

(defmacro eip (r) `(cis (* (float pi .0) ,r)))

(defun 30-60-90-fill (s &optional (dep 13) &aux (%i #c(0 1)) (w12 (eip -1/6)) (o12 (eip 1/6)))
  (cond ((< dep 0) 0)				;compiler blows out optimizing 
	((< s 1/3)
	 (/ (+ (* w12 (conjugate (30-60-90-fill (* 3 s) (- dep 1)))) o12)
	    (sqrt 3)))
	((< s 2/3) (/ (+ (* (eip 5/6) (30-60-90-fill (- 2 (* 3 s)) (- dep 1))) %i)
		      (sqrt 3)))
	(t (/ (- (* %i (conjugate (30-60-90-fill (- (* 3 s) 2) (- dep 1)))) w12)
	      (sqrt 3)))))

(DEFUN equifill (&KEY (STEPS (* 15/2 (+ 729 1/15))) (WINDOW (GET-BW-WINDOW)) (pause 't)
		 (scale #c(0 1)) (dz #c(.5 .33)) (R .64)
		 (PHASE (/ -0 15 729)) (fun #'30-60-90-fill)
		 &AUX CY (CX (FLOOR (MULTIPLE-VALUE-SETQ (NIL CY) (SEND WINDOW :INSIDE-SIZE))
				    2))
		      (lo-res (send window :exposed-p)))
  (send window :set-label (format 'nil "~d steps, phase = ~d" STEPS PHASE))
  (setf dz (* dz cy)  dz (complex (/ (* (REALPART dz) 2 cx) cy) (IMAGPART dz))
	r (* r scale cy))
  (when (send window :exposed-p) (send window :select))
  (flet ((TRIDRAW (Z1 Z2)
	   (incf z1 dz) (incf z2 dz)
	   (SEND WINDOW :draw-TRIANGLE CX CY
		 (ROUND (REALPART Z1)) (- CY (ROUND (IMAGPART Z1)))
		 (ROUND (REALPART Z2)) (- CY (ROUND (IMAGPART Z2))) tv:alu-xor))
	 (xflip (z 2r) (complex (- 2r (realpart z)) (imagpart z))))
    (SEND WINDOW :CLEAR-WINDOW)
    (LOOP WITH DT = (/ STEPS)
	  WITH Z = (* R (funcall fun (setf phase (OR PHASE 0))))
	  WITH r2Z1 = (* 2 (realpart (* R (funcall fun 1))))
	  FOR U FROM (+ phase DT) BY DT UPTO 1 DO
      (TRIDRAW Z (SETF Z (* R (funcall fun U))))	;(read-char window)
      FINALLY (TRIDRAW Z (xflip (SETF Z (* R (funcall fun (setf u (- 2 u))))) r2Z1))
	      (loop do (TRIDRAW (xflip Z r2Z1)
				(xflip (SETF Z (* R (funcall fun (setf u (- u dt))))) r2Z1))
		       ;(read-char window)
		    until ( u phase))
	      (TRIDRAW (xflip z r2Z1) (* R (funcall fun phase)))
	      ))
  (when lo-res (when pause (send window :tyi)) (send window :deselect)))

(defvar moby-bw-window)

(defun get-moby-window ()
  (or (and (boundp 'moby-bw-window) moby-bw-window) (make-moby-bw-window)))

(defun make-moby-bw-window (&optional (width 3200) (height 2058))
  (setq moby-bw-window
	(tv:make-window 'graphics-window ':superior tv:main-screen
			':edges `(0 0 ,width ,height)
			':expose-p 'nil :deexposed-typeout-action :permit
			':blinker-p 'nil ':borders 'nil
			':label nil ':save-bits 't)))

(defun peano-poster (&key (steps 8192/5) (phase 1/16384) (panels-high 7)
			  (window (get-moby-window)) (skip 0)
		     &aux height (width (MULTIPLE-VALUE-SETQ (NIL height)
					  (SEND WINDOW :INSIDE-SIZE)))
			  (panels-wide (ceiling (* panels-high height) width)))
  (loop with r = (/ 1 1.025)
	with scale = (* panels-high 1.02)
	for x from 0 below panels-wide do
    (loop for y from 0 below panels-high
	  for dz = (complex (- x) (- y)) do
      (when (< (decf skip) 0)
	(wrecpeano :r r :steps steps :phase phase :scale scale :dz dz)
	(wrecpeano :r r :steps steps :phase phase :scale scale :dz dz :window window)
	(hci::hardcopy-window window :margins 'nil :zoom 1))))) 

(defun flake-poster (&key (steps 8192/7) (phase 0) (panels-high 7)
			  (window (get-moby-window)) (skip 0)
		     &aux height (width (MULTIPLE-VALUE-SETQ (NIL height)
					  (SEND WINDOW :INSIDE-SIZE)))
			  (panels-wide (ceiling (* panels-high height) width)))
  (loop with r = (/ 1 1.025)
	with scale = (* panels-high 1.02)
	for x from 0 below panels-wide do
    (loop for y from 0 below panels-high
	  for dz = (complex (- x) (- y)) do
      (when (< (decf skip) 0)
	(wrecflake :r r :steps steps :phase phase :scale scale :dz dz)
	(wrecflake :r r :steps steps :phase phase :scale scale :dz dz :window window)
	(hci::hardcopy-window window :margins 'nil :zoom 1)))))

(DEFUN RECPEANO (&KEY (STEPS 1024) (WINDOW (GET-BW-WINDOW))  (R (send window :inside-height))
		 (PHASE 'NIL)
		 &AUX CY (CX (FLOOR (MULTIPLE-VALUE-SETQ (NIL CY) (SEND WINDOW :INSIDE-SIZE))
				    2)))
  (send window :select)
  (LABELS ((TRIDRAW (Z1 Z2)
	     (SEND WINDOW :AOS-TRIANGLE CX CY
		   (ROUND (REALPART Z1)) (- CY (ROUND (IMAGPART Z1)))
		   (ROUND (REALPART Z2)) (- CY (ROUND (IMAGPART Z2))) 1)))
    (SEND WINDOW :CLEAR-WINDOW)
    (LOOP WITH Z = (* R (HASH-PEANO 0))
	  WITH DT = (/ STEPS)
	  FOR U FROM (OR PHASE DT) BY DT UPTO 1 DO
      (TRIDRAW Z (SETF Z (* R (HASH-PEANO U))))
      FINALLY (TRIDRAW Z R)))
  (send window :tyi) (send window :deselect))

(DEFUN PEANO-FRAME (FRAME)
  (RECPEANO :R 774 :STEPS (/ (EXPT 4 (+ 6 (/ FRAME 144))) 5) :PHASE (/ FRAME 16384 144)))

;; i(f(1,1-2t)-1/sqrt3)
(DEFUN FILL-MAP (SLOW-T &AUX (FAST-T SLOW-T) (SLOW-A 1) (SLOW-P 0) (FAST-A 1) (FAST-P 0))
  (MACROLET ((FILL-STEP (TI A P) `(IF (> ,TI 0) (SETF ,TI (- 1 ,TI ,TI)
						      ,P (+ ,P (* ,A #C(1/2 1/2)))
						      ,A (* ,A #C(-1/2 1/2)))
				   (SETF ,TI (- -1 ,TI ,TI)
					 ,P (+ ,P (* ,A #C(-1/2 1/2)))
					 ,A (* ,A #C(-1/2 -1/2))))))
    (LOOP DO (FILL-STEP FAST-T FAST-A FAST-P)
	  WHEN (= FAST-T SLOW-T)
	    RETURN (/ (- (* FAST-A SLOW-P) (* SLOW-A FAST-P))
		      (- FAST-A SLOW-A))
	  DO (FILL-STEP SLOW-T SLOW-A SLOW-P)
	     (FILL-STEP FAST-T FAST-A FAST-P))))

(DEFUN UNFILL-MAP (Z)
  (LABELS ((MERGE (A B) (IF (EQUAL A B) A (APPEND A B)))
	   (+STEP (Z A P)
	     (VALUES (CONJUGATE (- (* #C(1 1) Z) #C(0 1))) (/ A 2) (+ P (/ A 2))))
	   (-STEP (Z A P)
	     (VALUES (CONJUGATE (- (* #C(1 -1) Z) #C(0 1))) (/ A 2) (+ P (* A -1/2))))
	   (UNFILL-REC1 (SLOW-Z SLOW-A SLOW-P FAST-Z FAST-A FAST-P
			 &AUX (RZ (REALPART FAST-Z)))
	     (IF (= RZ 0)
		 (MERGE (MULTIPLE-VALUE-CALL #'UNFILL-REC1
						    (+STEP FAST-Z FAST-A FAST-P)
						    (+STEP FAST-Z FAST-A FAST-P))
			 (MULTIPLE-VALUE-CALL #'UNFILL-REC1
						    (-STEP FAST-Z FAST-A FAST-P)
						    (-STEP FAST-Z FAST-A FAST-P)))
	      (MULTIPLE-VALUE-CALL #'UNFILL-REC2
					 SLOW-Z SLOW-A SLOW-P
					 (IF (> RZ 0) (+STEP FAST-Z FAST-A FAST-P)
					  (-STEP FAST-Z FAST-A FAST-P)))))
	   (UNFILL-REC2 (SLOW-Z SLOW-A SLOW-P FAST-Z FAST-A FAST-P)
	     (IF (= SLOW-Z FAST-Z) (LIST (/ (- (* FAST-A SLOW-P) (* SLOW-A FAST-P))
					    (- FAST-A SLOW-A)))
	      (LET ((RZ (REALPART SLOW-Z)))
		(IF (= RZ 0) (ERROR "Slow 0")
		 (MULTIPLE-VALUE-CALL #'UNFILL-REC3
					    (IF (> RZ 0) (+STEP SLOW-Z SLOW-A SLOW-P)
					     (-STEP SLOW-Z SLOW-A SLOW-P))
					    FAST-Z FAST-A FAST-P)))))
	   (UNFILL-REC3 (SLOW-Z SLOW-A SLOW-P FAST-Z FAST-A FAST-P
			 &AUX (RZ (REALPART FAST-Z)))
	     (IF (= RZ 0)
		 (MERGE (MULTIPLE-VALUE-CALL #'UNFILL-REC1
						    (+STEP FAST-Z FAST-A FAST-P)
						    (+STEP FAST-Z FAST-A FAST-P))
			 (MULTIPLE-VALUE-CALL #'UNFILL-REC1
						    (-STEP FAST-Z FAST-A FAST-P)
						    (-STEP FAST-Z FAST-A FAST-P)))
	      (MULTIPLE-VALUE-CALL #'UNFILL-REC1
					 SLOW-Z SLOW-A SLOW-P
					 (IF (> RZ 0) (+STEP FAST-Z FAST-A FAST-P)
					  (-STEP FAST-Z FAST-A FAST-P))))))
    (UNFILL-REC1 Z 1 0 Z 1 0)))

;(peano :order 10 :blue-period (/(expt 2 21)104):red-period (/ (expt 2 21) 109) :green-period (/ (expt 2 21) 115))
;(peano :order 10 :red-period (/(expt 2 20) 64):green-period (/ (expt 2 20) 67) :blue-period (/ (expt 2 20) 61))
;(MAKE-INSTANCE 'CIRCLE :DT (/ 2pi BLUE-PERIOD) :PHI BLUE-PHASE :R 64 :CX 65)
(DEFUN PEANO (&KEY (ORDER 10) (X 0) (Y 0)
		   (RED-PERIOD 75025) (GREEN-PERIOD 121393) (BLUE-PERIOD 196418)
		   (RED-PHASE (* pi 3/2)) (GREEN-PHASE (* pi 1/2)) (BLUE-PHASE (* pi 1/2))
		   (CONTRAST 64) (BRIGHTNESS 65)	;for bow-mode t
	      &AUX (ALU (COLOR:SC-FILL-ALU #O77777777 -1)) (WINDOW COLOR-HAX-WINDOW)
		   (2pi (* 2 pi))
		   (RED-CIRC (MAKE-COS-WAVE :AMPLITUDE CONTRAST :DTHETA (/ 2pi RED-PERIOD)
					    :PHASE RED-PHASE))
		   (GREEN-CIRC (MAKE-COS-WAVE :AMPLITUDE CONTRAST :DTHETA (/ 2pi GREEN-PERIOD)
					     :PHASE GREEN-PHASE))
		   (BLUE-CIRC (MAKE-COS-WAVE :AMPLITUDE CONTRAST :DTHETA (/ 2pi BLUE-PERIOD)
					     :PHASE BLUE-PHASE)))
  (SEND WINDOW :DRAW-RECTANGLE (SEND WINDOW :INSIDE-WIDTH) (SEND WINDOW :INSIDE-HEIGHT) 0 0
	ALU)
  (LABELS ((PEAN (ORDER DX1 DY1 DX2 DY2)
	     (COND ((= ORDER 0)
		    (SEND ALU :SET-FILL-DATA
			  (DPB (FLOOR (+ BRIGHTNESS (SEND BLUE-CIRC :NEXT))) (BYTE 8 16)
			       (DPB (FLOOR (+ BRIGHTNESS (SEND GREEN-CIRC :NEXT))) (BYTE 8 8)
				    (FLOOR (+ BRIGHTNESS (SEND RED-CIRC :NEXT))))))
		    (SEND WINDOW :DRAW-POINT X Y ALU))
		   ('T (PEAN (1- ORDER) DX2 DY2 DX1 DY1)
		       (INCF X DX2) (INCF Y DY2)
		       (PEAN (1- ORDER) DX1 DY1 DX2 DY2)
		       (INCF X DX1) (INCF Y DY1)
		       (PEAN (1- ORDER) DX1 DY1 DX2 DY2)
		       (DECF X DX2) (DECF Y DY2)
		       (PEAN (1- ORDER) (- DX2) (- DY2) (- DX1) (- DY1))))))
    (PEAN ORDER 1 0 0 1)))

(DEFUN SETUP-PEANO-COLOR-MAP ()
  (SEND COLOR:COLOR-SCREEN :SET-MAP-MODE :24-BIT 0)
  (LOOP FOR RED FROM 0 BY (FLOOR 1023 255)
	FOR GREEN FROM 0 BY (FLOOR 1023 255)
	FOR BLUE FROM 0 BY (FLOOR 1023 255)
	FOR COLOR TO 255 DO
    (SEND COLOR:COLOR-SCREEN :WRITE-COLOR-MAP-IMAGE COLOR
	  (LOGAND RED 1023) (LOGAND GREEN 1023) (LOGAND BLUE 1023)))
  (SEND COLOR:COLOR-SCREEN :NEW-COLOR-MAP NIL 0))

(DEFUN mod-2-COLOR-MAP ()
  ;(SEND COLOR:COLOR-SCREEN :SET-MAP-MODE :8-BIT 1)
  (LOOP FOR COLOR TO 255
	for m = (- 1024 (LOGAND color 1)) DO
    (SEND COLOR:COLOR-SCREEN :WRITE-COLOR-MAP COLOR	;-IMAGE
	  m m m 1))
  #|(SEND COLOR:COLOR-SCREEN :NEW-COLOR-MAP NIL 0)|#)

(defun flip-color-maps ()
  (loop do (mod-2-color-map)
	   (read-char)
	   (step-color-map 0 'nil 'nil 'nil)))

(defvar polybaz-array (make-array 22 :fill-pointer 0))

(defvar vdx0101 '((0 . 22) (1 . 20) (0 . 21) (-1 . 23)))

(defvar vdy0101 '((0 . 21) (1 . 23) (0 . 22) (-1 . 20)))

(defvar vd1210 '((1 . 22) (0 . 22) ( -1 . 22) (-2 . 11)
		   (-1 . 22) (0 . 22) (1 . 22) (2 . 11))) 

(defvar big-corn  '((1 . 22) (0 . 22) ( -1 . 22) (-2 . 2)
		   (-1 . 2) (0 . 22) (1 . 2) (2 . 3)))

(defvar vd1111 '((0 . 22) (1 . 22) (0 . 22) (-1 . 22)
		   (0 . 22) (-1 . 22) (0 . 22) (1 . 22)))

(defvar 3slope '((2 . 22) (0 . 22) (-1 . 9) ( -2 . 22) (-3 . 11)
		   (-2 . 22) (0 . 22) (1 . 9) (2 . 22) (3 . 11)))

(defun wave-fixup (vel-durs amp)
  (multiple-value-bind (pos neg) (pos-neg vel-durs)
    (multiple-value-bind (min max)
	(min-max (setq vel-durs (pn-scale vel-durs (- neg) pos (min (- neg) pos))))
      (multiple-value-SETQ (vel-durs min) (pn-scale vel-durs amp amp (- max min)))
      (dc-tweak vel-durs min))))

(defun pos-neg (vel-durs &aux (pos 0) (neg 0))
  (dolist (vel-dur vel-durs) (if (plusp (car vel-dur))
				 (setq pos (+ pos (* (car vel-dur) (cdr vel-dur))))
			      (setq neg (+ neg (* (car vel-dur) (cdr vel-dur))))))
  (values pos neg))

(defun min-max (vel-durs &aux (min 0) (max 0) (loc 0))
  (dolist (vel-dur vel-durs)
    (setq min (min min (setq loc (+ loc (* (car vel-dur) (cdr vel-dur))))))
    (setq max (max max loc)))
  (values min max))

(defun pn-scale (vel-durs pos neg den)
  (loop for vel-dur in vel-durs
	for dur = (FLOOR (* (cdr vel-dur) (if (plusp (car vel-dur)) pos neg))
		      den)
	collect (cons (car vel-dur) dur) into vel-durs
	sum (* (car vel-dur) dur) into loc
	finally (RETURN (VALUES VEL-DURS LOC))))

(defun dc-tweak (vel-durs gap &aux time)
  (loop for vel-dur in vel-durs
	for dur = (if (plusp (* (car vel-dur) gap))
		      (cond ((> (cdr vel-dur) (setq time (FLOOR gap (car vel-dur))))
			     (setq gap (- gap (* time (car vel-dur))))
			     (- (cdr vel-dur) time))
			    ((setq gap (- gap (* (car vel-dur) (cdr vel-dur))))
			     0))
		      (cdr vel-dur))
	sum dur into period
	collect (cons (car vel-dur) dur) into vel-durs
	sum (* (car vel-dur) dur) into loc
	minimize loc into min
	finally (RETURN (VALUES VEL-DURS (- MIN) PERIOD))))

(defun polyKEY (&KEY (xvd vd1210) (yvd vd1210)
		     (w (get-window)) (BW-TO-COLOR 'NIL)
		     (x-amp (tv:sheet-inside-width w)) (y-amp (tv:sheet-inside-height w))
		&aux (goodies (list xvd yvd x-amp y-amp))
		     (pointer (fill-pointer polybaz-array))
		     (*print-pretty* 'nil))
  (SEND W :EXPOSE)
  (send w :set-label `(:style ,(fifth (send w :label))
		       :string , (format 'nil "Polybaz ~D ~D"
				        (if (AND (!= pointer 0)
						 (equal (aref polybaz-array (1- pointer))
							     goodies))
					    pointer
					  (ASET goodies polybaz-array pointer)
					  (SETF pointer (INCF (FILL-POINTER polybaz-array))))
					goodies)))
  (MULTIPLE-VALUE-CALL w ':polybaz (wave-fixup xvd x-amp) (wave-fixup yvd y-amp) BW-TO-COLOR))

(defun rebaz (index &optional (window (get-window)) (windu 'nil) (dont-clear 'nil)
		    &aux (goodies (aref polybaz-array index)))
  (polybaz :xvd (first goodies) :yvd (second goodies) :window window :windu windu
	   :x-amp (third goodies) :y-amp (fourth goodies) :dont-clear dont-clear))

(defun polybaz (&key (xvd vd1210)
			  (yvd vd1210)
			  (window (get-window))
			  (x-amp (tv:sheet-inside-width window))
			  (y-amp (tv:sheet-inside-height window))
			  (windu 'nil) (dont-clear 'nil)
		&aux (goodies (list xvd yvd x-amp y-amp))
		     (pointer (fill-pointer polybaz-array)))
  (SEND Window :EXPOSE)
  (unless dont-clear (send window :clear-window) (when windu (send windu :clear-window)))
  (sys:standard-value-let ((*print-pretty* 'nil))
    (send window :set-label `(:style (:fix :roman :very-small)
		       :string ,(format 'nil "Polybaz ~D ~D"
				        (if (or (= pointer 0)
						(not (equal (aref polybaz-array (1- pointer))
							     goodies)))
					    (PROGN (ASET goodies polybaz-array pointer)
						   (SETF pointer (INCF (FILL-POINTER polybaz-array))))
					  pointer)
					goodies)))
  (MULTIPLE-VALUE-CALL window ':polybaz (wave-fixup xvd x-amp) (wave-fixup yvd y-amp)
		       'nil windu)))

(defmethod (:polybaz graphics-window) (xvd x0 xper yvd y0 yper
				       &OPTIONAL (BW-TO-COLOR 'NIL) (windu 'nil)
					  &aux (xlen (length xvd)) (ylen (length yvd))
					       (stepping 'nil) (lcm (lcm xper yper))
					       input ALU-1 ALU-2
					       (PLANE (LSH 1 8)))
  (BLOCK OUTER
    (WHEN BW-TO-COLOR (SETF ALU-1 (COLOR:SC-FILL-ALU PLANE PLANE))
		      (SETF ALU-2 (COLOR:SC-FILL-ALU 0 PLANE))
		      (SEND COLOR:COLOR-SCREEN :SET-MAP-MODE :8-BIT 1 8 0)
		      (SEND ALU-1 :SET-ARRAY :MASKED-BITS)
		      (SEND ALU-2 :SET-ARRAY :NOT-MASKED-BITS)
		      (ONE-BIT-COLOR-MAP (MOD PLANE 255) 1))
  (send self :set-label
	`(:style ,(fifth (send self :label))
	  :string ,(string-append (sixth (send self :label))
				  (format 'nil ", gcd ~D" (/ (* xper yper) lcm)))))
  (when windu (send windu :set-label (send self :label)))
  (setq input (if (send *standard-input* :exposed-p) *standard-input*
		(send self :select)
		self))
  (do ((back '(#\Triangle #\Circle) forth)
       (forth '(#\Square #\Circle) back)
       (x x0) (y y0) (px 0) (py 0)
       (xvd (copy-TREE xvd)
	    (loop with xv repeat xlen
		  for vd in (cdr xvd)
		  do (push (cons (- (car vd)) (cdr vd)) xv)
		  finally (return xv)))
       (yvd (copy-TREE yvd)
	    (loop with yv repeat ylen
		  for vd in (cdr yvd)
		  do (push (cons (- (car vd)) (cdr vd)) yv)
		  finally (return yv)))
       (time lcm (- lcm time))
       (xd (cdar xvd) (- (cdar xvd) xd))
       (yd (cdar yvd) (- (cdar yvd) yd)))
      ('nil)
    (nconc xvd xvd) (nconc yvd yvd)
    (do (ch)
	('nil)
      (loop with ix and iy and dt
	    for ch = (progn (tv:wait-remote-screen-idle color:color-screen)
			    (send input :tyi-no-hang))
	    until ch
	    for xv = (caar xvd) for yv = (caar yvd)
	    if (> xd yd) do (setq dt yd  xd (- xd yd)  yd (cdar (setq yvd (cdr yvd))))
	    else do (setq dt xd  yd (- yd xd)  xd (cdar (setq xvd (cdr xvd))))
	    do (setq ix (* dt xv) iy (* dt yv))
	       (decf time dt)
	    if (or stepping (!= (* px iy) (* py ix)))
	    do (when windu (send windu ':aos-triangle x0 y0 x y (+ x px) (+ y py)))
	       (send self ':aos-triangle x0 y0 x y (incf x px) (incf y py))
	       (setq px ix py iy)
	    else do (incf px ix) (incf py iy)
	    if ( time 0) do (setq stepping 't)
	    if stepping do (setq time (MOD (+ time lcm) lcm))
			   (return)
	    finally (send input ':untyi ch))
      (WHEN BW-TO-COLOR
	(SEND COLOR:COLOR-SCREEN :CLEAR-WINDOW)
	(SEND COLOR:COLOR-SCREEN :BITBLT
	      ALU-1 TV:WIDTH TV:HEIGHT (TV:SHEET-SCREEN-ARRAY SELF) 0 0 0 0))
      (setq ch (send input :tyi))
      (setq stepping (member ch '(#\Space #\Circle)))
      (WHEN (EQL Ch #\End) (RETURN-FROM OUTER 'T))
      (when (member ch back) (return)))))
  (SEND *STANDARD-INPUT* :EXPOSE))

(DEFUN CAFE-CEILING (&optional (w (get-BW-WINDOW)))
;  (SEND COLOR:COLOR-SCREEN :SET-BOW-MODE 'T)
  (POLYKEY :XVD '((-1 . 55) (0 . 4) (1 . 57) (2 . 1) (1 . 1) (0 . 1))
	   :YVD '((-1 . 55) (0 . 55) (1 . 53) (2 . 1) (1 . 0) (0 . 0))
	   :X-AMP 1088 :Y-AMP 743 :BW-TO-COLOR 'NIL :W w))

(DEFCONSTANT TV:ALU-Y 3 "Should already be.")
(DEFCONSTANT TV:ALU-~Y (LOGXOR #O17 3) "Should already be.")

; superposition of regular hex grid on square grid such that every square contains
; the center of at least one hexagon.  (in fact 1/2 + 1/(sqrt 3) = 1.07735, on avg.)
(DEFUN HEXSQUARE (&OPTIONAL (SQUARE-SIZE 69) (WINDOW *TERMINAL-IO*) (fraction-barred .9)
		  &AUX (width (SEND window :INSIDE-width))
		       (height (SEND window :INSIDE-height))
		       (SIZE (MAX width height))
		       (BLINK (SEND WINDOW :CURSOR-VISIBILITY)))
  (SEND WINDOW :SET-CURSOR-VISIBILITY 'NIL)
  (SEND WINDOW :BITBLT TV:ALU-SETA SIZE SIZE TV:33%-GRAY 0 0 0 0)
  (LOOP WITH DX = (* SQUARE-SIZE #.(- (SQRT 3) 2))
	WITH EDGE-X = (* SQUARE-SIZE #.(- (SQRT 1/3)))
	WITH EDGE-Y = (* DX #.(SQRT 1/3))
	FOR ALUS ON (CIRCULAR-LIST TV:ALU-Y TV:ALU-SETO TV:ALU-SETZ)	;can't #. !
	FOR X FROM 0 BY SQUARE-SIZE BELOW (/ SIZE #.(- (SQRT 3) 1))
	FOR Y FROM 0 BY DX DO
    (LOOP FOR ALU IN ALUS
	  FOR X FROM X BY DX
	  FOR Y FROM Y BY SQUARE-SIZE TO SIZE DO
      (SEND WINDOW :DRAW-REGULAR-POLYGON X Y (+ X EDGE-X) (+ Y EDGE-Y) 6 ALU)
      UNTIL (> Y SIZE)))	
  (LOOP FOR X FROM 0 BY (* 2 SQUARE-SIZE) TO (* fraction-barred width) DO
    (SEND WINDOW :BITBLT TV:ALU-~Y SQUARE-SIZE SIZE TV:33%-GRAY X 0 X 0))
  (LOOP FOR X FROM 0 BY (* 2 SQUARE-SIZE) TO (* fraction-barred height) DO
    (SEND WINDOW :BITBLT TV:ALU-~Y SIZE SQUARE-SIZE TV:33%-GRAY 0 X 0 X))
  (READ-CHAR)
  (SEND WINDOW :SET-CURSOR-VISIBILITY BLINK)) 

(DEFSUBST DOT-SELF (X Y Z) (+ (* X X) (* Y Y) (* Z Z)))

(DEFSUBST DOT-PRODUCT (X1 Y1 Z1 X2 Y2 Z2)
  (+ (* X1 X2) (* Y1 Y2) (* Z1 Z2)))

(DEFSUBST NORM (X Y Z) (SQRT (DOT-SELF X Y Z)))

(DEFUN UNIT-VECTOR (X Y Z &AUX (NORM (NORM X Y Z)))
  (VALUES (/ X NORM) (/ Y NORM) (/ Z NORM)))

(DEFUN CROSS-PRODUCT (X1 Y1 Z1 X2 Y2 Z2)
  (VALUES (- (* Y1 Z2) (* Y2 Z1)) (- (* Z1 X2) (* Z2 X1)) (- (* X1 Y2) (* Y1 X2))))

(DEFUN TUBE (X. Y. Z. X.. Y.. Z..
	     &AUX (X/X (/ (DOT-PRODUCT X. Y. Z. X.. Y.. Z..) (DOT-SELF X. Y. Z.)))
		  (U1 (- X.. (* X. X/X)))
		  (U2 (- Y.. (* Y. X/X)))
		  (U3 (- Z.. (* Z. X/X))))
  (SI:%MULTIPLE-VALUE-CALL-N VALUES
   (UNIT-VECTOR U1 U2 U3) 3
   (SI:%MULTIPLE-VALUE-CALL-N UNIT-VECTOR (CROSS-PRODUCT X. Y. Z. U1 U2 U3) 3) 3))

(DEFUN KNOTTED-TORUS (&KEY (R1 #.(+ 1/2 (SQRT 1/3))) (R2 .5) (R3 .5) (T-STEPS 288)
			   (Z-OFF -1/2) (S-STEPS 69) (WINDOW (get-bw-window)) (TILT 69))
  (LOOP WITH COS-TILT = (COSD TILT)  WITH SIN-TILT = (SIND TILT)
	WITH CENTER-X = (FLOOR (SEND WINDOW :INSIDE-WIDTH) 2)
	WITH CENTER-Y = (FLOOR (SEND WINDOW :INSIDE-HEIGHT) 2)
	WITH SINCOS2T = (MAKE-INSTANCE 'CIRCLE :PERIOD (/ T-STEPS 2))
	WITH SINCOS3T = (MAKE-INSTANCE 'CIRCLE :PERIOD (/ T-STEPS 3) :R R2)
	WITH SINCOS-S = (MAKE-INSTANCE 'CIRCLE :PERIOD S-STEPS :R R3)
	WITH SIN2T  WITH SIN3T
	repeat T-STEPS
	FOR COS2T = (MULTIPLE-VALUE-SETQ (NIL SIN2T) (SEND SINCOS2T :NEXT))
	FOR COS3T = (MULTIPLE-VALUE-SETQ (NIL SIN3T) (SEND SINCOS3T :NEXT))
	FOR OX = (+ R1 R2) THEN X  FOR OY = 0 THEN Y  FOR OZ = Z-OFF THEN Z
	FOR X = (+ (* R1 COS2T) (* COS2T COS3T))
	FOR Y = (+ (* R1 SIN2T) (* SIN2T COS3T))
	FOR Z = (+ SIN3T Z-OFF)
	FOR X. = (+ (* R1 -2 SIN2T) (* (- (* -3 COS2T SIN3T) (* 2 SIN2T COS3T))))
	FOR Y. = (+ (* R1 2 COS2T) (* (- (* 2 COS2T COS3T) (* 3 SIN2T SIN3T))))
	FOR Z. = (* 3 COS3T)
	FOR X.. = (+ (* R1 -4 COS2T) (* (- (* 12 SIN2T SIN3T) (* 13 COS2T COS3T))))
	FOR Y.. = (+ (* R1 -4 SIN2T) (* (- (* -12 COS2T SIN3T) (* 13 SIN2T COS3T))))
	FOR Z.. = (* -9 SIN3T)
	SUM (NORM (- X OX) (- Y OY) (- Z OZ)) DO
    (MULTIPLE-VALUE-BIND (U1 U2 U3 V1 V2 V3) (TUBE X. Y. Z. X.. Y.. Z..)
      (LOOP WITH SIN-S
	    REPEAT S-STEPS
	    FOR COS-S = (MULTIPLE-VALUE-SETQ (NIL SIN-S) (SEND SINCOS-S :NEXT))
	    FOR W1 = (+ X (* COS-S U1) (* SIN-S V1))
	    FOR W2 = (+ Y .5 (* COS-S U2) (* SIN-S V2))
	    FOR W3 = (+ Z (* COS-S U3) (* SIN-S V3)) DO
	(PSETQ W3 (+ (* COS-TILT W3) (* SIN-TILT W2))
	       W2 (/ (+ 3 (- (* COS-TILT W2) (* SIN-TILT W3))) 1.5 CENTER-Y))
	(SEND WINDOW :DRAW-POINT (+ CENTER-X (FLOOR W1 W2)) (- CENTER-Y (FLOOR W3 W2)))))))

(defmacro polyran (&optional (xvd 'vd1210) (yvd 'vd1210) . body)
  `(polybaz :xvd (ranvd ,xvd) :yvd (ranvd ,yvd) . ,body))

(DEFVAR RANVD-SEQUENCE 'NIL)

(defun ranvd (vel-durs)
  (UNLESS RANVD-SEQUENCE (SETF RANVD-SEQUENCE (MAKE-INSTANCE 'RANDOM-SEQUENCE)))
  (loop for vd in vel-durs
	collect (cons (car vd) (MOD (SEND RANVD-SEQUENCE :NEXT) (1+ (cdr vd))))))

(compile-flavor-methods graphics-window color-graphics-window CIRC circle)

(defun bw-to-color (&optional (bw bw-window) (PLANE (LSH 16 8)) ;(COLOR (GET-COLOR-WINDOW))
		    &AUX HEIGHT (WIDTH (MULTIPLE-VALUE-SETQ (NIL HEIGHT)
					 (SEND COLOR:COLOR-SCREEN :VISIBLE-SIZE)))
			 (ALU (COLOR:SC-FILL-ALU PLANE PLANE)))
  (SEND COLOR:COLOR-SCREEN :SET-MAP-MODE :8-BIT 1 8 0)
  (SEND ALU :SET-ARRAY :MASKED-BITS)
  (ONE-BIT-COLOR-MAP (MOD PLANE 255) 1)
  (SEND COLOR:COLOR-SCREEN :BITBLT ALU WIDTH HEIGHT (TV:SHEET-SCREEN-ARRAY BW) 0 0 0 0)
  (SEND ALU :SET-ARRAY :NOT-MASKED-BITS)
  (SEND COLOR:COLOR-SCREEN :BITBLT ALU WIDTH HEIGHT (TV:SHEET-SCREEN-ARRAY BW) 0 0 0 0))

(zl:comment (defvar newton-t0s (make-array 9 :initial-element 0) "Old tan x/3")

;newton: C-(C*(C*(3*CC-C)+3)-CC)/(3*(C*(2*CC-C)+1))
;works fast, but finds undesired solns.
 (defun trivex (x0 y0 x1 y1 x2 y2 &optional (uid 0))
  (decf x1 x0) (decf y1 y0) (decf x2 x0) (decf y2 y0)
  (psetq x2 (+ (* x1 x2) (* y1 y2))  y2 (- (* x1 y2) (* x2 y1)))
  (loop with flip = (> (abs y2) (abs x2))
	with tancot = (if flip (/ x2 y2) (/ y2 x2))
	for app = (aref newton-t0s uid)
		then (- app (/ eps (* 3 (+ 1 (* app (- (* 2 tancot) app))))))
	for eps = (- (* app (+ 3 (* app (- (* 3 tancot) app)))) tancot) 
	when (< (abs eps) .1e-7)
	  return (if flip
		     (values (+ x0 (- (* x1 (setf x2 (+ (sqrt 3)
							(setf (aref newton-t0s uid) app))))
				      (* y1 (setf y2 (- 1 (* (sqrt 3) app))))))
			     (+ y0 (* y1 x2) (* x1 y2)))
		  (values (+ x0 (- x1 (* y1 (setf (aref newton-t0s uid) app))))
			  (+ y0 y1 (* x1 app))))))
)

(defun segseg (x0 y0 x1 y1 x2 y2 x3 y3)
  (decf x1 x0) (decf y1 y0) (decf x2 x0) (decf y2 y0) (decf x3 x0) (decf y3 y0)
  (psetq x2 (+ (* x1 x2) (* y1 y2))  y2 (- (* x1 y2) (* x2 y1)))
  (psetq x3 (+ (* x1 x3) (* y1 y3))  y3 (- (* x1 y3) (* x3 y1)))
  (let* ((dy (- y3 y2))
	 (det (+ (expt x1 2) (expt y1 2)))
	 (x (if (< (min det (abs dy)) .1e-5) (/ (+ det x2 x3) 4)
	     (/ (- (* y3 x2) (* y2 x3)) dy))))
    (values (+ x0 (/ (* x x1) det)) (+ y0 (/ (* x y1) det)))))

(defun trivex (x0 y0 x1 y1 x2 y2 &optional (tan 'nil))
  (decf x1 x0) (decf y1 y0) (decf x2 x0) (decf y2 y0)
  (unless tan (setf tan (tan (/ (atan (- (* x1 y2) (* x2 y1)) (+ (* x1 x2) (* y1 y2))) 3)))) 
  (values (+ x0 (- x1 (* y1 tan))) (+ y0 y1 (* x1 tan)) (- tan)))

(defun trivix (x0 y0 x1 y1 x2 y2 &optional (tan 'nil) (bump 0) &aux det)
  (decf x1 x0) (decf y1 y0) (decf x2 x0) (decf y2 y0)
  (unless tan
    (setf tan
	  (tan (/ (- (atan (setf det (- (* x1 y2) (* x2 y1))) (+ (* x1 x2) (* y1 y2)))
		     (* (signum det) bump))
		  3)))) 
  (values (+ x0 (- x1 (* y1 tan))) (+ y0 y1 (* x1 tan)) (- tan)))

(zl:comment ; Salamin's external variant
 (defun trivex (x0 y0 x1 y1 x2 y2 &optional (tan 'nil))
  (decf x1 x0) (decf y1 y0) (decf x2 x0) (decf y2 y0)
  (unless tan (setf tan (tan (/ (-  (arctn (- (* x1 y2) (* x2 y1)) (+ (* x1 x2) (* y1 y2))) (float pi .0)) 3)))) 
  (values (+ x0 (- x1 (* y1 tan))) (+ y0 y1 (* x1 tan)) (- tan)))

 (defun trivex (x0 y0 x1 y1 x2 y2 &optional (tan 'nil) &aux det)
  (decf x1 x0) (decf y1 y0) (decf x2 x0) (decf y2 y0)
  (unless tan (setf tan (tan (/ (- (atan (setf det (- (* x1 y2) (* x2 y1)))
						(+ (* x1 x2) (* y1 y2)))
					  (* (signum det) (float pi .0)))
				       3))))
  (values (+ x0 (- x1 (* y1 tan))) (+ y0 y1 (* x1 tan)) (- tan)))
)
(defun arctn (y x &aux (a (atan y x)))
  (if (< a 0) (+ a (* 2 (float pi .0))) a))

(defun move-edge (w u1 v1 u2 v2 x1 y1 x2 y2 color2 color1)
  (decf color1 color2)
  (send w :aos-triangle u1 v1 u2 v2 x2 y2 color1)
  (send w :aos-triangle u1 v1 x2 y2 x1 y1 color1))

(defun morely (&key (r1 333) (r2 444) (r3 333) (w (get-window))
		    (dt1 (/ (float pi .0) 384)) (dt2 (/ (sqrt 2) -384)) (dt3 (/ (exp 1) 384))
		    (bump1 0) (bump2 0) (bump3 0))
  (setf bump1 (* bump1 (float pi .0))  bump2 (* bump2 (float pi .0))
	bump3 (* bump3 (float pi .0)))
  (loop with x1 = r1 with y1 = r1 with x2 = (+ r1 r2) with y2 = r2
	with x3 = (/ (+ x1 x2) 2) with y3 = (+ (/ x2 3) r3)
	with c1 = (make-instance 'circle :cx x1 :cy y1 :dt dt1 :r r1)
	with c2 = (make-instance 'circle :cx x2 :cy y2 :dt dt2 :r r2)
	with c3 = (make-instance 'circle :cx x3 :cy y3 :dt dt3 :r r3)
	with u1 with v1 with u2 with v2 with u3 with v3
	with u12 with v12 with u23 with v23 with u31 with v31
	with color1 = 1 with color2 = 3 with color3 = 0	;6
	with pause = 't do
    (multiple-value-setq (x1 y1) (send c1 :step))
    (multiple-value-setq (x2 y2) (send c2 :step))
    (multiple-value-setq (x3 y3) (send c3 :step))
    (multiple-value-bind (x12 y12 tan1) (trivix x1 y1 x2 y2 x3 y3 nil bump1)
      (multiple-value-bind (x21 y21 tan2) (trivix x2 y2 x1 y1 x3 y3 nil bump2)
	(multiple-value-setq (x12 y12) (segseg x1 y1 x12 y12 x2 y2 x21 y21))
	(multiple-value-bind (x23 y23) (trivix x2 y2 x3 y3 x1 y1 tan2)
	  (multiple-value-bind (x32 y32 tan3) (trivix x3 y3 x2 y2 x1 y1 nil bump3)
	    (multiple-value-setq (x23 y23) (segseg x2 y2 x23 y23 x3 y3 x32 y32))
	    (multiple-value-bind (x31 y31) (trivix x3 y3 x1 y1 x2 y2 tan3)
	      (multiple-value-bind (x13 y13) (trivix x1 y1 x3 y3 x2 y2 tan1)
		(multiple-value-setq (x31 y31) (segseg x3 y3 x31 y31 x1 y1 x13 y13)))
	      (cond (u1 (move-edge w u1 v1 u2 v2 x1 y1 x2 y2 color1 0)
			(move-edge w u2 v2 u12 v12 x2 y2 x12 y12 color1 0)
			(move-edge w u12 v12 u1 v1 x12 y12 x1 y1 color1 color3)
			(move-edge w u2 v2 u3 v3 x2 y2 x3 y3 color1 0)
			(move-edge w u3 v3 u23 v23 x3 y3 x23 y23 color1 0)
			(move-edge w u23 v23 u2 v2 x23 y23 x2 y2 color1 0)
			(move-edge w u3 v3 u1 v1 x3 y3 x1 y1 color1 0)
			(move-edge w u1 v1 u31 v31 x1 y1 x31 y31 color1 color3)
			(move-edge w u31 v31 u3 v3 x31 y31 x3 y3 color1 0)
			(move-edge w u12 v12 u23 v23 x12 y12 x23 y23 color2 0)
			(move-edge w u23 v23 u31 v31 x23 y23 x31 y31 color2 0)
			(move-edge w u31 v31 u12 v12 x31 y31 x12 y12 color2 color3))
		    ('t (send w :aos-triangle x1 y1 x2 y2 x12 y12 color1)
			(send w :aos-triangle x2 y2 x3 y3 x23 y23 color1)
			(send w :aos-triangle x3 y3 x1 y1 x31 y31 color1)
			;(send w :aos-triangle x1 y1 x12 y12 x31 y31 color3)
			(send w :aos-triangle x12 y12 x23 y23 x31 y31 color2)))
	      (setf u1 x1  v1 y1  u2 x2  v2 y2  u3 x3  v3 y3
		    u12 x12  v12 y12  u23 x23  v23 y23  u31 x31  v31 y31))))))
    (setf pause (eql #\Space (send *standard-input* (if pause :tyi :tyi-no-hang))))))

(defun morley (&key (r1 333) (r2 444) (r3 333) (w (get-window))
		    (dt1 (/ (float pi .0) 384)) (dt2 (/ (sqrt 2) -384)) (dt3 (/ (exp 1) 384)))
  (loop with x1 = r1 with y1 = r1 with x2 = (+ r1 r2) with y2 = r2
	with x3 = (/ (+ x1 x2) 2) with y3 = (+ (/ x2 3) r3)
	with c1 = (make-instance 'circle :cx x1 :cy y1 :dt dt1 :r r1)
	with c2 = (make-instance 'circle :cx x2 :cy y2 :dt dt2 :r r2)
	with c3 = (make-instance 'circle :cx x3 :cy y3 :dt dt3 :r r3)
	with u1 with v1 with u2 with v2 with u3 with v3
	with u12 with v12 with u23 with v23 with u31 with v31
	with color1 = 1 with color2 = 3 with color3 = 0	;6
	with pause = 't do
    (multiple-value-setq (x1 y1) (send c1 :step))
    (multiple-value-setq (x2 y2) (send c2 :step))
    (multiple-value-setq (x3 y3) (send c3 :step))
    (multiple-value-bind (x12 y12 tan1) (trivex x1 y1 x2 y2 x3 y3)
      (multiple-value-bind (x21 y21 tan2) (trivex x2 y2 x1 y1 x3 y3)
	(multiple-value-setq (x12 y12) (segseg x1 y1 x12 y12 x2 y2 x21 y21))
	(multiple-value-bind (x23 y23) (trivex x2 y2 x3 y3 x1 y1 tan2)
	  (multiple-value-bind (x32 y32 tan3) (trivex x3 y3 x2 y2 x1 y1)
	    (multiple-value-setq (x23 y23) (segseg x2 y2 x23 y23 x3 y3 x32 y32))
	    (multiple-value-bind (x31 y31) (trivex x3 y3 x1 y1 x2 y2 tan3)
	      (multiple-value-bind (x13 y13) (trivex x1 y1 x3 y3 x2 y2 tan1)
		(multiple-value-setq (x31 y31) (segseg x3 y3 x31 y31 x1 y1 x13 y13)))
	      (cond (u1 (move-edge w u1 v1 u2 v2 x1 y1 x2 y2 color1 0)
			(move-edge w u2 v2 u12 v12 x2 y2 x12 y12 color1 0)
			(move-edge w u12 v12 u1 v1 x12 y12 x1 y1 color1 color3)
			(move-edge w u2 v2 u3 v3 x2 y2 x3 y3 color1 0)
			(move-edge w u3 v3 u23 v23 x3 y3 x23 y23 color1 0)
			(move-edge w u23 v23 u2 v2 x23 y23 x2 y2 color1 0)
			(move-edge w u3 v3 u1 v1 x3 y3 x1 y1 color1 0)
			(move-edge w u1 v1 u31 v31 x1 y1 x31 y31 color1 color3)
			(move-edge w u31 v31 u3 v3 x31 y31 x3 y3 color1 0)
			(move-edge w u12 v12 u23 v23 x12 y12 x23 y23 color2 0)
			(move-edge w u23 v23 u31 v31 x23 y23 x31 y31 color2 0)
			(move-edge w u31 v31 u12 v12 x31 y31 x12 y12 color2 color3))
		    ('t (send w :aos-triangle x1 y1 x2 y2 x12 y12 color1)
			(send w :aos-triangle x2 y2 x3 y3 x23 y23 color1)
			(send w :aos-triangle x3 y3 x1 y1 x31 y31 color1)
			;(send w :aos-triangle x1 y1 x12 y12 x31 y31 color3)
			(send w :aos-triangle x12 y12 x23 y23 x31 y31 color2)))
	      (setf u1 x1  v1 y1  u2 x2  v2 y2  u3 x3  v3 y3
		    u12 x12  v12 y12  u23 x23  v23 y23  u31 x31  v31 y31))))))
    (setf pause (eql #\Space (send *standard-input* (if pause :tyi :tyi-no-hang))))))

;image of z under [0,1] -> [z0,z1]: [-Y*(Y1-Y0)+X*(X1-X0)+X0,X*(Y1-Y0)+Y0+(X1-X0)*Y]
(defun segmap (x0 y0 x1 y1 x y)
  (decf x1 x0) (decf y1 y0)
  (values (+ x0 (- (* x x1) (* y y1))) (+ y0 (* x y1) (* y x1))))

(defun napoleon (&key (r1 222) (r2 333) (r3 222) (w (get-window)) (color1 -5) (color2 -8)
	         (dt1 (/ (float pi .0) 384)) (dt2 (/ (sqrt 2) -384)) (dt3 (/ (exp 1) 384)))
  (loop with x1 = (* 2 r1) with y1 = (* 2 r1) with x2 = (+ x1 r2) with y2 = (* 2 r2)
	with x3 = (/ (+ x1 x2) 2) with y3 = (+ (/ x2 3) r3)
	with c1 = (make-instance 'circle :cx x1 :cy y1 :dt dt1 :r r1)
	with c2 = (make-instance 'circle :cx x2 :cy y2 :dt dt2 :r r2)
	with c3 = (make-instance 'circle :cx x3 :cy y3 :dt dt3 :r r3)
	with u1 with v1 with u2 with v2 with u3 with v3
	with u12 with v12 with u23 with v23 with u31 with v31
	with u21 with v21 with u32 with v32 with u13 with v13
	;with color1 = 1 with color2 = 3 ;with color3 = 6
	with pause = 't do
    (multiple-value-setq (x1 y1) (send c1 :step))
    (multiple-value-setq (x2 y2) (send c2 :step))
    (multiple-value-setq (x3 y3) (send c3 :step))
    (multiple-value-bind (x12 y12) (segmap x1 y1 x2 y2 .5 (sqrt 3/4))
      (multiple-value-bind (x23 y23) (segmap x2 y2 x3 y3 .5 (sqrt 3/4))
	(multiple-value-bind (x31 y31) (segmap x3 y3 x1 y1 .5 (sqrt 3/4))
	  (let* ((x21 (/ (+ x1 x2 x12) 3)) (y21 (/ (+ y1 y2 y12) 3))
		 (x32 (/ (+ x2 x3 x23) 3)) (y32 (/ (+ y2 y3 y23) 3))
		 (x13 (/ (+ x3 x1 x31) 3)) (y13 (/ (+ y3 y1 y31) 3)))
	    (cond (u1 (move-edge w u1 v1 u12 v12 x1 y1 x12 y12 color1 0)
		      (move-edge w u12 v12 u2 v2 x12 y12 x2 y2 color1 0)
		      (move-edge w u2 v2 u1 v1 x2 y2 x1 y1 color1 0)
		      (move-edge w u2 v2 u23 v23 x2 y2 x23 y23 color1 0)
		      (move-edge w u23 v23 u3 v3 x23 y23 x3 y3 color1 0)
		      (move-edge w u3 v3 u2 v2 x3 y3 x2 y2 color1 0)
		      (move-edge w u3 v3 u31 v31 x3 y3 x31 y31 color1 0)
		      (move-edge w u31 v31 u1 v1 x31 y31 x1 y1 color1 0)
		      (move-edge w u1 v1 u3 v3 x1 y1 x3 y3 color1 0)
		      (move-edge w u21 v21 u32 v32 x21 y21 x32 y32 color2 0)
		      (move-edge w u32 v32 u13 v13 x32 y32 x13 y13 color2 0)
		      (move-edge w u13 v13 u21 v21 x13 y13 x21 y21 color2 0))
		  ('t (send w :aos-triangle x1 y1 x12 y12 x2 y2 color1)
		      (send w :aos-triangle x2 y2 x23 y23 x3 y3 color1)
		      (send w :aos-triangle x3 y3 x31 y31 x1 y1 color1)
		      (send w :aos-triangle x21 y21 x32 y32 x13 y13 color2)))
	    (setf u1 x1  v1 y1  u2 x2  v2 y2  u3 x3  v3 y3
		    u12 x12  v12 y12  u23 x23  v23 y23  u31 x31  v31 y31
		    u21 x21  v21 y21  u32 x32  v32 y32  u13 x13  v13 y13)))))
    (setf pause (eql #\Space (send *standard-input* (if pause :tyi :tyi-no-hang))))))

(defun trisex (ignore ignore x2 y2 x3 y3 &optional ignore)
  (values (+ (* (float 2/3) x2) (* (float 1/3) x3))
	  (+ (* (float 2/3) y2) (* (float 1/3) y3))))

(defun garfunkle (&key (r1 222) (r2 333) (r3 222) (w (get-window))
	         (dt1 (/ (float pi .0) 384)) (dt2 (/ (sqrt 2) -384)) (dt3 (/ (exp 1) 384)))
 (stack-let (;(in-x (make-array 6)) (in-y (make-array 6))
	     (u-out (make-array 12)) (v-out (make-array 12))
	     (x-out (make-array 12)) (y-out (make-array 12)))
  (loop with x1 = (* 2 r1) with y1 = (* 2 r1) with x2 = (+ x1 r2) with y2 = (* 2 r2)
	with x3 = (/ (+ x1 x2) 2) with y3 = (+ (/ x2 3) r3)
	with c1 = (make-instance 'circle :cx x1 :cy y1 :dt dt1 :r r1)
	with c2 = (make-instance 'circle :cx x2 :cy y2 :dt dt2 :r r2)
	with c3 = (make-instance 'circle :cx x3 :cy y3 :dt dt3 :r r3)
	with u1 with v1 with u2 with v2 with u3 with v3
	with u12 with v12 with u23 with v23 with u31 with v31
	with u21 with v21 with u32 with v32 with u13 with v13
	with color1 = 1 with color2 = 4 with color3 = 16
	with pause = 't do
    (multiple-value-setq (x1 y1) (send c1 :step))
    (multiple-value-setq (x2 y2) (send c2 :step))
    (multiple-value-setq (x3 y3) (send c3 :step))
    (multiple-value-bind (x23 y23 tan1) (trisex x1 y1 x2 y2 x3 y3)
      (zl:comment (x23 y23) (segseg x1 y1 x23 y23 x2 y2 x3 y3))
      (multiple-value-bind (x32 y32) (trisex x1 y1 x3 y3 x2 y2 tan1)
	(zl:comment (x32 y32) (segseg x1 y1 x32 y32 x2 y2 x3 y3))
	(multiple-value-bind (x31 y31 tan2) (trisex x2 y2 x3 y3 x1 y1)
	  (zl:comment (x31 y31) (segseg x2 y2 x31 y31 x3 y3 x1 y1))
	  (multiple-value-bind (x13 y13) (trisex x2 y2 x1 y1 x3 y3 tan2)
	    (zl:comment (x13 y13) (segseg x2 y2 x13 y13 x3 y3 x1 y1))
	    (multiple-value-bind (x12 y12 tan3) (trisex x3 y3 x1 y1 x2 y2)
	      (zl:comment (x12 y12) (segseg x3 y3 x12 y12 x1 y1 x2 y2))
	      (multiple-value-bind (x21 y21) (trisex x3 y3 x2 y2 x1 y1 tan3)
		(zl:comment (x21 y21) (segseg x3 y3 x21 y21 x1 y1 x2 y2))
		(macrolet ((conc (x n) (find-symbol (format 'nil "~a~d" x n)))
			   (salt (n i j)
			     `(multiple-value-bind (x y)
				  (segmap (conc x ,i) (conc y ,i) (conc x ,j) (conc y ,j)
					  .5 (sqrt 3/4))
				(setf (aref x-out ,n) x  (aref y-out ,n) y)))
			   (m-e (n1 n2 &rest rest)
			     `(move-edge w (conc u ,n1) (conc v ,n1) (conc u ,n2) (conc v ,n2)
					   (conc x ,n1) (conc y ,n1) (conc x ,n2) (conc y ,n2)
					   .,rest)))
		  (salt 0 32 23) (salt 1 31 32) (salt 2 13 31)
		  (salt 3 12 13) (salt 4 21 12) (salt 5 23 21)
		  (salt 6 23 32) (salt 7 32 31) (salt 8 31 13)
		  (salt 9 13 12) (salt 10 12 21) (salt 11 21 23)
		  (cond (u1 (m-e 1 12 color1 0) (m-e 12 13 color1 0) (m-e 13 1 color1 0)
			    (m-e 2 23 color1 0) (m-e 23 21 color1 0) (m-e 21 2 color1 0)
			    (m-e 3 31 color1 0) (m-e 31 32 color1 0) (m-e 32 3 color1 0)
			    ;(m-e 32 23 color2 0)
			    (move-edge w u23 v23 (aref u-out 0) (aref v-out 0)
				         x23 y23 (aref x-out 0) (aref y-out 0) color2 0)
			    (move-edge w (aref u-out 0) (aref v-out 0) u32 v32
				         (aref x-out 0) (aref y-out 0) x32 y32 color2 0)
			    (move-edge w u32 v32 (aref u-out 6) (aref v-out 6)
				         x32 y32 (aref x-out 6) (aref y-out 6) color2 0)
			    (move-edge w (aref u-out 6) (aref v-out 6) u23 v23
				         (aref x-out 6) (aref y-out 6) x23 y23 color2 0)
			    ;(m-e 31 32 color2 0)
			    (move-edge w u32 v32 (aref u-out 1) (aref v-out 1)
				         x32 y32 (aref x-out 1) (aref y-out 1) color2 0)
			    (move-edge w (aref u-out 1) (aref v-out 1) u31 v31
				         (aref x-out 1) (aref y-out 1) x31 y31 color2 0)
			    (move-edge w u31 v31 (aref u-out 7) (aref v-out 7)
				         x31 y31 (aref x-out 7) (aref y-out 7) color2 0)
			    (move-edge w (aref u-out 7) (aref v-out 7) u32 v32
				         (aref x-out 7) (aref y-out 7) x32 y32 color2 0)
			    ;(m-e 13 31 color2 0)
			    (move-edge w u31 v31 (aref u-out 2) (aref v-out 2)
				         x31 y31 (aref x-out 2) (aref y-out 2) color2 0)
			    (move-edge w (aref u-out 2) (aref v-out 2) u13 v13
				         (aref x-out 2) (aref y-out 2) x13 y13 color2 0)
			    (move-edge w u13 v13 (aref u-out 8) (aref v-out 8)
				         x13 y13 (aref x-out 8) (aref y-out 8) color2 0)
			    (move-edge w (aref u-out 8) (aref v-out 8) u31 v31
				         (aref x-out 8) (aref y-out 8) x31 y31 color2 0)
			    ;(m-e 12 13 color2 0)
			    (move-edge w u13 v13 (aref u-out 3) (aref v-out 3)
				         x13 y13 (aref x-out 3) (aref y-out 3) color2 0)
			    (move-edge w (aref u-out 3) (aref v-out 3) u12 v12
				         (aref x-out 3) (aref y-out 3) x12 y12 color2 0)
			    (move-edge w u12 v12 (aref u-out 9) (aref v-out 9)
				         x12 y12 (aref x-out 9) (aref y-out 9) color2 0)
			    (move-edge w (aref u-out 9) (aref v-out 9) u13 v13
				         (aref x-out 9) (aref y-out 9) x13 y13 color2 0)
			    ;(m-e 21 12 color2 0)
			    (move-edge w u12 v12 (aref u-out 4) (aref v-out 4)
				         x12 y12 (aref x-out 4) (aref y-out 4) color2 0)
			    (move-edge w (aref u-out 4) (aref v-out 4) u21 v21
				         (aref x-out 4) (aref y-out 4) x21 y21 color2 0)
			    (move-edge w u21 v21 (aref u-out 10) (aref v-out 10)
				         x21 y21 (aref x-out 10) (aref y-out 10) color2 0)
			    (move-edge w (aref u-out 10) (aref v-out 10) u12 v12
				         (aref x-out 10) (aref y-out 10) x12 y12 color2 0)
			    ;(m-e 23 21 color2 0)
			    (move-edge w u21 v21 (aref u-out 5) (aref v-out 5)
				         x21 y21 (aref x-out 5) (aref y-out 5) color2 0)
			    (move-edge w (aref u-out 5) (aref v-out 5) u23 v23
				         (aref x-out 5) (aref y-out 5) x23 y23 color2 0)
			    (move-edge w u23 v23 (aref u-out 11) (aref v-out 11)
				         x23 y23 (aref x-out 11) (aref y-out 11) color2 0)
			    (move-edge w (aref u-out 11) (aref v-out 11) u21 v21
				         (aref x-out 11) (aref y-out 11) x21 y21 color2 0)
			    (loop for i below 6 for j = (if (= i 5) 0 (+ 1 i)) do
			      (move-edge w (aref u-out i) (aref v-out i)
					   (aref u-out j) (aref v-out j)
					   (aref x-out i) (aref y-out i)
					   (aref x-out j) (aref y-out j) color3 0)
			      (let ((i (+ i 6)) (j (+ j 6)) (color3 (- color3)))
				(move-edge w (aref u-out i) (aref v-out i)
					     (aref u-out j) (aref v-out j)
					     (aref x-out i) (aref y-out i)
					     (aref x-out j) (aref y-out j) color3 0))))
			('t (send w :aos-triangle x1 y1 x12 y12 x13 y13 color1)
			    ;(send w :aos-triangle x1 y1 x32 y32 x3 y3 color1)
			    (send w :aos-triangle x2 y2 x23 y23 x21 y21 color1)
			    ;(send w :aos-triangle x2 y2 x13 y13 x1 y1 color1)
			    (send w :aos-triangle x3 y3 x31 y31 x32 y32 color1)
			    ;(send w :aos-triangle x3 y3 x21 y21 x2 y2 color1)
			    (send w :aos-triangle x32 y32 x23 y23
				    (aref x-out 0) (aref y-out 0) color2)
			    (send w :aos-triangle x23 y23 x32 y32
				    (aref x-out 6) (aref y-out 6) color2)
			    (send w :aos-triangle x31 y31 x32 y32
				  (aref x-out 1) (aref y-out 1) color2)
			    (send w :aos-triangle x32 y32 x31 y31
				  (aref x-out 7) (aref y-out 7) color2)
			    (send w :aos-triangle x13 y13 x31 y31
				  (aref x-out 2) (aref y-out 2) color2)
			    (send w :aos-triangle x31 y31 x13 y13
				  (aref x-out 8) (aref y-out 8) color2)
			    (send w :aos-triangle x12 y12 x13 y13
				  (aref x-out 3) (aref y-out 3) color2)
			    (send w :aos-triangle x13 y13 x12 y12
				  (aref x-out 9) (aref y-out 9) color2)
			    (send w :aos-triangle x21 y21 x12 y12
				  (aref x-out 4) (aref y-out 4) color2)
			    (send w :aos-triangle x12 y12 x21 y21
				  (aref x-out 10) (aref y-out 10) color2)
			    (send w :aos-triangle x23 y23 x21 y21
				  (aref x-out 5) (aref y-out 5) color2)
			    (send w :aos-triangle x21 y21 x23 y23
				  (aref x-out 11) (aref y-out 11) color2)
			    (loop for i from 1 to 4  for j = (+ i 1) do
			      (send w :aos-triangle (aref x-out 0) (aref y-out 0)
				    (aref x-out i) (aref y-out i)
				    (aref x-out j) (aref y-out j) color3))
			    (loop for i from 7 to 10  for j = (+ i 1) do
			      (send w :aos-triangle (aref x-out 6) (aref y-out 6)
				    (aref x-out i) (aref y-out i)
				    (aref x-out j) (aref y-out j) (- color3))))))
		(setf u1 x1  v1 y1  u2 x2  v2 y2  u3 x3  v3 y3
		    u12 x12  v12 y12  u23 x23  v23 y23  u31 x31  v31 y31
		    u21 x21  v21 y21  u32 x32  v32 y32  u13 x13  v13 y13)
		(rotatef u-out x-out) (rotatef v-out y-out)))))))
    (setf pause (eql #\Space (send *standard-input* (if pause :tyi :tyi-no-hang)))))))

(defun graham (n &optional (window (get-window))
	         &aux (width (send window :inside-width))
		      (height (send window :inside-height))
		      (r (/ width (* 2.0 n)))
		      (r-1 (- r 0))
		      (straight-y (- (/ height 2) (* r 13/3)))
		      (wiggle-y (+ (/ height 2) (* r 1/3)))
		      (2r (+ r r))
		      (wid (* n 2r)))
  (graphics:draw-rectangle 0 straight-y wid (+ straight-y 2r 2r) :stream window :alu :flip)
  (graphics:draw-rectangle 0 wiggle-y wid (+ wiggle-y 2r 2r) :stream window :alu :flip)
  (loop for x from r by 2r until ( x wid) do
    (graphics:draw-circle x (+ straight-y r) r-1 :stream window :alu :flip)
    (graphics:draw-circle x (+ straight-y 2r r) r-1 :stream window :alu :flip))
  (loop with 2rsin = (* 2r (/ (- 3 (sqrt 6)) 6))
	with 2rcos = (* 2r (/ (+ (sqrt 18) (sqrt 3)) 6))
	with 2rsin =  (* 2r (/ (+ 3 (sqrt 6)) 6))	;=+pi/3
	with 2rcos =  (* 2r (/ (- (sqrt 18) (sqrt 3)) 6))
	with wyggle = (+ (* 2 wiggle-y) 2r 2rsin)
	for x from r by 2rcos until ( x wid)
	for y = (+ r wiggle-y) then (- wyggle y) do
    (graphics:draw-circle x y r-1 :stream window :alu :flip)
    (graphics:draw-circle (+ 2rcos x) (+ 2rsin y) r-1 :stream window :alu :flip)))

(defun zcoz (&key (window (get-bw-window)) (x-span 8.2d0) (x0 0) (ox 1/5) (oy 1/2)
	     (dx 1) (dy 1) center)
  (when center (multiple-value-setq (ox oy)
		 (oxy-from-center (realpart center) (imagpart center) x-span window)))
  (loop with c1 = (+ pi 3.69672)  with c0 = (* pi 3.69672)	; cos  = -pi
	with pi/2 = (float (/ pi 2) x-span)
	with width  = (send window :inside-width)
	with height = (send window :inside-height)
	with gain = (/ x-span width)
	with xo = (floor (* width ox))  with yo = (floor (* height (- 1 oy)))
	with xmax = (- width xo)
	with oflo = (typecase x-span (double-float 703) (t 84))
	for x from (max x0 (- xo)) to xmax by dx do
    (loop for y from (max 0 (floor (* height oy -1))) to yo by dy
	  for z0 = (* (complex x y) gain)
	  for alu = (if (and (or (> x 0) (= y 0))
			     (loop repeat 333
				   for z = z0 then (* z (cos z))
				   for rez = (abs (float (realpart z) .0))
				   for imz = (abs (float (imagpart z) .0))
				   for zz = (+ (expt rez 2) (expt imz 2))
				   thereis ( zz (* pi/2 rez))	;|z-pi/4|pi/4
				   when (or (> zz oflo)
					    ( zz (- (* c1 rez) c0))
					    (capturep rez imz))
				     return 'nil))
			:draw :erase) do
      (send window :draw-point (+ xo x) (+ yo y) alu)
      (send window :draw-point (- xo x) (+ yo y) alu)
      (send window :draw-point (+ xo x) (- yo y) alu)
      (send window :draw-point (- xo x) (- yo y) alu))))

(defvar capvec (make-array 1 :fill-pointer 1))

(defun capturep (re im)
  (and ( re pi) ( (* im re) 1)
       (loop for n from (* pi (fill-pointer capvec)) by pi to re do
	 (vector-push-extend (/ (nutsol n (/ 1 (+ (/ n 2) (/ 2 pi)))) 2) capvec)
	finally (multiple-value-bind (m f) (floor re pi)
		  (return (< (+ (expt f 2) (expt im 2)) (* f 2 (aref capvec m))))))))

(defun nutsol (n g)
  (let* ((n+g (+ n g))
	 (cos  (cos g))
	 (dg (/ (- (* n+g cos) n) (- (* n+g (sin g)) cos))))
    (if (< (abs dg) .00001) g
     (nutsol n (+ g dg)))))

(defun center-from-oxy (ox oy x-span &optional (window (get-bw-window)))
  (values (* x-span (- 1/2 ox))
	  (* x-span (- 1/2 oy) (/ (send window :inside-height) (send window :inside-width)))))

(defun oxy-from-center (x y x-span &optional (window (get-bw-window)))
  (values (- 1/2 (/ x x-span))
	  (- 1/2 (/ (* (send window :inside-width) y) (send window :inside-height) x-span))))

(defun szcosfp (z &optional (x0 z) (eps (* 22 (typecase z
					     (double-float double-float-epsilon)
					     (t single-float-epsilon)))))
  (loop for x = x0 then (* x (+ 1 delta))
	for zsinx = (* z (sin x))
	for delta = (/ (- (/ (* z (cos x)) x) 1) (+ zsinx 1))
	when (< (abs delta) eps) return zsinx))

(defun zcos (&key (window (get-bw-window)) (x-span 8.2d0) (x0 0) (ox 1/5) (oy 1/2))
  (loop with width  = (send window :inside-width)
	with height = (send window :inside-height)
	with gain = (/ x-span width)
	with xo = (floor (* width ox))  with yo = (floor (* height (- 1 oy)))
	with xmax = (- width xo)
	for x from (max x0 (- xo)) to xmax by 5 do
    (loop for y from 0 to yo by 5
	  for z = (* (complex x y) gain)
	  for alu = (if (< (abs (szcosfp z)) 1) :draw :erase) do
      (send window :draw-point (+ xo x) (+ yo y) alu)
      (send window :draw-point (+ xo x) (- yo y) alu)
      (send window :draw-point (- xo x) (+ yo y) alu)
      (send window :draw-point (- xo x) (- yo y) alu))))

(defun penrose (scale &optional (window (get-bw-window)) (slip -3/5)
		      &aux (yscale (* scale 4/4))
			   (width (ceiling (send window :inside-width) scale))
			   (height (ceiling (send window :inside-height) yscale)))
  (flet ((fill (&optional (xmax width) (y0 -1))
	  (graphics:with-graphics-scale (window scale yscale)
	    (loop for x from 0 to xmax do
	      (loop for y from y0 by 2 to height
		    for y+ = (+ y (mod (* 5/5 x) 2)) do
		(graphics:draw-rectangle x y+ (+ 1 x) (+ 1 y+) :stream window :alu :flip))))))
    (fill)
    (graphics:with-graphics-rotation (window (atan 5/12))
      ;(graphics:with-graphics-translation (window (/ width 3) (/ height 3))
      (fill (* width 6/5) (* height slip))
      (graphics:with-graphics-rotation (window (atan 3/4))
	;(fill width (* height slip))
))))

(DEFSUBST ROTTEN (X) (IF (ZL:BIGP X) (ASH X -11) (ROT X 11)))
  
(DEFUN RATHASH (X)
  (SETF X (RATIONAL X))
  (SI:NON-NEGATIVE-FIXNUM (LOGXOR (ROTTEN (SI:NON-NEGATIVE-FIXNUM (NUMERATOR X)))
				  (SI:NON-NEGATIVE-FIXNUM (DENOMINATOR X)))))

(DEFFLAVOR MEMOIZE-MIXIN ((COPY-HASHES 0)) ()
  (:REQUIRED-INSTANCE-VARIABLES SI:HASH-TABLE-LOCK Si:SIZE SI:HASH-ARRAY SI:FILLED-ELEMENTS
				REHASH-SIZE))

(defflavor RATTY-hash-table (REHASH-SIZE) (si:equal-hash-table))
(si:define-equal-hash-table-flavor RATTY-HASH-table eql RAThash)

(DEFFLAVOR MEMOIZE-TABLE () (MEMOIZE-MIXIN RATTY-HASH-TABLE))

(DEFMETHOD (:COPY-HASH MEMOIZE-MIXIN :AFTER) (&REST IGNORE)
  (INCF COPY-HASHES))

(DEFMETHOD (:INIT RATTY-HASH-TABLE :AFTER) (&REST IGNORE)
  (SETF REHASH-SIZE (* SI:SIZE SI:REHASH-THRESHOLD)))

(DEFMETHOD (:SET-REHASH-THRESHOLD RATTY-HASH-TABLE :AFTER) (&REST IGNORE)
  (SETF REHASH-SIZE (* SI:SIZE SI:REHASH-THRESHOLD)))

(DEFMETHOD (:GROW RATTY-hash-table :AFTER) (&REST IGNORE)
  (SETF REHASH-SIZE (* SI:SIZE SI:REHASH-THRESHOLD)))

(DEFMETHOD (:PAGE-IN-ARRAY SI:BASIC-HASH-TABLE) (&REST OPTIONS)
  (APPLY #'SYS:PAGE-IN-ARRAY SI:HASH-ARRAY 'NIL 'NIL OPTIONS))

(DEFMETHOD (:MEMOIZE MEMOIZE-MIXIN) (KEY FUNCTION)
  (DECLARE (SYS:DOWNWARD-FUNARG FUNCTION))	;useless?
  (SI:LOCK-HASH-TABLE
    (LET* (GC-UNSTABLE HASH-CODE
	   (HASH (MULTIPLE-VALUE-SETQ (NIL GC-UNSTABLE) (RATHASH KEY)))
	   (FIRST-FREE 'NIL)
	   (OLD-COPY-HASHES COPY-HASHES)
	   (INDEX (MOD HASH SI:SIZE)))
      (WHEN GC-UNSTABLE (ERROR "MEMOIZE of GC-unstable key"))
      (FLET ((INSERT-HASH (VALUE)
	       (DECLARE (SYS:DOWNWARD-FUNCTION))
	       (IF (!= OLD-COPY-HASHES COPY-HASHES)
		   (SEND SELF :PUT-HASH KEY VALUE)
		   (WITHOUT-INTERRUPTS
		     (WHEN FIRST-FREE (SETF INDEX FIRST-FREE))
		     (SETF (SI:EQUAL-HASH-VALUE INDEX) VALUE
			   (SI:EQUAL-HASH-HASH INDEX) HASH
			   (SI:EQUAL-HASH-KEY INDEX) KEY)
		     (UNLESS HASH-CODE (SETF HASH-CODE (INCF SI:FILLED-ELEMENTS))))
		   (WHEN (> SI:FILLED-ELEMENTS REHASH-SIZE) (SEND SELF :GROW))
		   VALUE)))
	(LOOP REPEAT SI:SIZE
	      DO (COND ((NULL (SETF HASH-CODE (SI:EQUAL-HASH-HASH INDEX)))
			(RETURN (INSERT-HASH (FUNCALL FUNCTION KEY NIL NIL #'INSERT-HASH))))
		       ((AND (< HASH-CODE 0) (NULL FIRST-FREE)) (SETQ FIRST-FREE INDEX))
		       ((AND (= HASH HASH-CODE) (EQL KEY (SI:EQUAL-HASH-KEY INDEX)))
			(RETURN (INSERT-HASH (FUNCALL FUNCTION KEY (SI:EQUAL-HASH-VALUE INDEX)
						      'T #'INSERT-HASH)))))
		 (SETF INDEX (MOD (1+ INDEX) SI:SIZE))
	      FINALLY (SEND SELF ':COPY-HASH SI:HASH-ARRAY SI:SIZE)
		      (RETURN (SEND SELF :MEMOIZE KEY FUNCTION)))))))

(DEFVAR PEANO-HASHES (MAKE-INSTANCE 'MEMOIZE-TABLE :SIZE 15000))

;; unit interval to unit square
(DEFUN HASH-PEANO (REAL &OPTIONAL (A1 1) (A0 0))
  (SEND PEANO-HASHES :MEMOIZE REAL
	  #'(LAMBDA (REAL VALUE FOUNDP STORER)
	      (DECLARE (SYS:DOWNWARD-FUNCTION))
	      (RETURN-FROM HASH-PEANO
		(IF FOUNDP (OR (CAR VALUE)
			       (SETF (CAR VALUE)
				     (DESTRUCTURING-BIND (IGNORE S1 S0) VALUE
							 (/ (- A0 S0) (- S1 A1)))))
		 (CAR (ZL:UNWIND-PROTECT-CASE ()
			  (RPLACA (FUNCALL STORER (LIST 'NIL A1 A0))
				  (SELECTOR (MULTIPLE-VALUE-SETQ (NIL REAL)
					      (FLOOR (* 4 REAL)))
				   = (0 (+ #C(0 1/2)
					   (* #C(0 -1/2)
					      (HASH-PEANO (- 1 REAL) (* A1 #C(0 -1/2))
							  (+ A0 (* A1 #C(0 1/2)))))))
				     (1 (+ #C(0 1/2)
					   (* 1/2 (HASH-PEANO REAL (/ A1 2)
							      (+ A0 (* A1 #C(0 1/2)))))))
				     (2 (+ #C(1/2 1/2)
					   (* 1/2 (HASH-PEANO REAL (/ A1 2)
							      (+ A0 (* A1 #C(1/2 1/2)))))))
				     (3 (+ 1 (* #C(0 1/2) (HASH-PEANO (- 1 REAL)
								      (* A1 #C(0 1/2))
								      (+ A0 A1)))))
				     (4 1)))
			(:ABORT (SEND PEANO-HASHES :REM-HASH REAL)))))))))

(defun pennies (&optional (r 99) (window (get-bw-window))
		&aux (wid (send window :inside-width)) (hei (send window :inside-height))
		     (2r (* 2 r)) (slop (+ 1 (/ (floor hei (- wid hei)))))
		     (dy (/ 2r (sqrt (+ 1 (expt slop 2))))) (dx (* slop dy)))
  (loop for x from dy by (* 2 dx) to (+ wid dy)
	for y from 0 by (* 2 dy) do
    (graphics:draw-polygon (list x y  (+ x dx) (+ y dy)
				 (- (+ x dx) dy) (+ y dy dx)  (- x dy) (+ y dx))
      :alu :flip :stream window)
    (graphics:draw-polygon (list (- (+ x dx) dy) (+ y dy dx)
				 (- (+ x (* 2 dx)) dy) (+ y (* 2 dy) dx)
				 (+ x (* 2 (- dx dy))) (+ y (* 2 (+ dy dx)))
				 (- (+ x dx) dy dy) (+ y dy dx dx))
      :alu :flip :stream window))
  (loop with x- = (sqrt (- (sqrt 3) 3/4))  with xx = (+ 1 x- x-)  with yy = (- 1 (sqrt 3/4))
	for x from (/ (+ dy dx) 2) by (* dx xx) to (+ wid r)
	for y from (/ (+ dy dx) 2) by (* dy xx) do
    (graphics:draw-circle x y r :alu :flip :stream window)
    (graphics:draw-circle (- (+ x (* dx x-)) (* dy yy))
			  (+ y (* dy x-) (* dx yy)) r :alu :flip :stream window)
    (graphics:draw-circle (- (+ x (* dx (+ -1/2 x-))) (* dy (+ yy (sqrt 3/4))))
			  (+ y (* dy (+ -1/2 x-)) (* dx (+ yy (sqrt 3/4))))
			  r :alu :flip :stream window)
    (graphics:draw-circle (- (+ x (* dx (+ 1/2 x-))) (* dy (+ yy (sqrt 3/4))))
			  (+ y (* dy (+ 1/2 x-)) (* dx (+ yy (sqrt 3/4))))
			  r :alu :flip :stream window)
    (graphics:draw-circle (+ x (* dx (- xx 1))) (+ y (* dy (- xx 1)))
			  r :alu :flip :stream window)
    (graphics:draw-circle (- (+ x (* dx (- xx 1/2))) (* dy (sqrt 3/4)))
			  (+ y (* dy (- xx 1/2)) (* dx (sqrt 3/4)))
			  r :alu :flip :stream window)))
